#include "CSVExportOptionDlg.h"


//--------------------------------------------------------------------------------
CSVExportOptionDlg::CSVExportOptionDlg(wxWindow* parent)
  : CSVExportOptionDlgDesign(parent),
    m_proj(NULL)
{
  m_listCtrlCandidate->AppendColumn(_("No."));
  m_listCtrlCandidate->AppendColumn(_("Label"));
  m_listCtrlCandidate->AppendColumn(_("X Data"));
  m_listCtrlCandidate->AppendColumn(_("Points"));
  m_listCtrlCandidate->SetColumnWidth(0, 30);

  m_listCtrlExportedItem->AppendColumn(_("No."));
  m_listCtrlExportedItem->AppendColumn(_("Label"));
  m_listCtrlExportedItem->AppendColumn(_("Header"));
  m_listCtrlExportedItem->SetColumnWidth(0, 30);
}

CSVExportOptionDlg::~CSVExportOptionDlg()
{
}

void CSVExportOptionDlg::SetGraphProject(GraphProject *project)
{
  m_proj = project;
}

void CSVExportOptionDlg::SetExportOption(CSVExportOption const &option)
{
  m_option = option;
  
  UpdateCandidate();
  UpdateExportedItem();
  UpdateExportedItemHeader();
  UpdateOptions();
}

void CSVExportOptionDlg::GetExportOption(CSVExportOption &option) const
{
  option = m_option;
}

void CSVExportOptionDlg::ButtonPush_OnButtonClick(wxCommandEvent &event)
{
  long item = m_listCtrlCandidate->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;

  wxASSERT(m_proj != NULL);
  wxASSERT(item >= 0 && item < m_proj->GetCount());
  if (m_option.exportedData.empty())
  {
    // ߂ăvbVƂ, ɒǉ\菈s킸ɒǉ.
    CSVExportedDataInfo info;
    info.graphDataIndex = item;
    info.headerText = m_proj->GetItem(item)->GetLabel();
    m_option.exportedData.push_back(info);
  }
  else
  {
    // ɒǉĂf[^ Data X ƃf[^vꍇ̂ݒǉ.
    int index0 = m_option.exportedData[0].graphDataIndex;
    GraphDataX dataX0 = m_proj->GetItem(index0)->GetDataX();
    GraphDataX dataX = m_proj->GetItem(item)->GetDataX();
    if (dataX == dataX0)
    {
      CSVExportedDataInfo info;
      info.graphDataIndex = item;
      info.headerText = m_proj->GetItem(item)->GetLabel();
      m_option.exportedData.push_back(info);
    }
    else
    {
      ::wxMessageBox(_("X-data and the number of points, that will be pushed, should be equal to the ones already appended!"),
        _("Warning"), wxOK | wxICON_EXCLAMATION);
    }
  }
  
  UpdateExportedItem();
  UpdateExportedItemHeader();
}

void CSVExportOptionDlg::ButtonPop_OnButtonClick(wxCommandEvent &event)
{
  long item = m_listCtrlExportedItem->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;

  wxASSERT(m_proj != NULL);
  wxASSERT(item >= 0 && item < static_cast<long>(m_option.exportedData.size()));

  m_option.exportedData.erase(m_option.exportedData.begin() + item);

  UpdateExportedItem();
  UpdateExportedItemHeader();
}

void CSVExportOptionDlg::ButtonMoveUp_OnButtonClick(wxCommandEvent &event)
{
  long item = m_listCtrlExportedItem->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;
  
  long itemUp = item - 1;
  if (itemUp < 0) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.exportedData.size()));
  wxASSERT(itemUp >= 0 && itemUp < static_cast<long>(m_option.exportedData.size()));
  std::swap(m_option.exportedData[item], m_option.exportedData[itemUp]);

  m_listCtrlExportedItem->SetItemState(item, 0, wxLIST_STATE_SELECTED);
  m_listCtrlExportedItem->SetItemState(itemUp, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);

  UpdateExportedItem();
  UpdateExportedItemHeader();
}

void CSVExportOptionDlg::ButtonMoveDown_OnButtonClick(wxCommandEvent &event)
{
  long item = m_listCtrlExportedItem->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;
  
  long itemDown = item + 1;
  if (itemDown >= static_cast<long>(m_option.exportedData.size())) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.exportedData.size()));
  wxASSERT(itemDown >= 0 && itemDown < static_cast<long>(m_option.exportedData.size()));
  std::swap(m_option.exportedData[item], m_option.exportedData[itemDown]);

  m_listCtrlExportedItem->SetItemState(item, 0, wxLIST_STATE_SELECTED);
  m_listCtrlExportedItem->SetItemState(itemDown, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);

  UpdateExportedItem();
  UpdateExportedItemHeader();
}

void CSVExportOptionDlg::ListCtrlExportedItem_OnListItemDeselected(wxListEvent &event)
{
  UpdateExportedItemHeader();
}

void CSVExportOptionDlg::ListCtrlExportedItem_OnListItemSelected(wxListEvent &event)
{
  UpdateExportedItemHeader();
}

void CSVExportOptionDlg::TextCtrlHeader_OnText(wxCommandEvent &event)
{
  long item = m_listCtrlExportedItem->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.exportedData.size()));
  m_option.exportedData[item].headerText = m_textCtrlHeader->GetValue();
  UpdateExportedItem();
}

void CSVExportOptionDlg::CheckBoxExportHeader_OnCheckBox(wxCommandEvent &event)
{
  m_option.exportHeader = m_checkBoxExportHeader->GetValue();
}

void CSVExportOptionDlg::CheckBoxExportDataX_OnCheckBox(wxCommandEvent &event)
{
  m_option.exportDataX = m_checkBoxExportDataX->GetValue();
  m_textCtrlDataXHeader->Enable(m_option.exportDataX);
}

void CSVExportOptionDlg::TextCtrlDataXHeader_OnText(wxCommandEvent &event)
{
  m_option.dataXHeaderText = m_textCtrlDataXHeader->GetValue();
}

void CSVExportOptionDlg::UpdateCandidate()
{
  wxASSERT(m_proj != NULL);

  // ACe̐킹
  while (m_listCtrlCandidate->GetItemCount() > static_cast<int>(m_proj->GetCount()))
  {
    m_listCtrlCandidate->DeleteItem(0);
  }
  while (m_listCtrlCandidate->GetItemCount() < static_cast<int>(m_proj->GetCount()))
  {
    m_listCtrlCandidate->InsertItem(0, _T(""));
  }

  // \̓eݒ (ύXӏ̂ݐݒ)
  for (int i = 0; i < m_proj->GetCount(); ++i)
  {
    wxString strNo, label, strDataX, strCount;
    
    strNo.Printf("%d", i + 1);

    GraphDataRefPtr graphData = m_proj->GetItem(i);
    label = graphData->GetLabel();
    
    GraphDataX dataX = graphData->GetDataX();
    strDataX.Printf(_T("S: %f, I: %f"), dataX.begin, dataX.interval);

    int count = graphData->GetCount();
    strCount.Printf(_T("%d"), count);
    
    if (m_listCtrlCandidate->GetItemText(i, 0) != strNo)
    {
      m_listCtrlCandidate->SetItem(i, 0, strNo);
    }
    
    if (m_listCtrlCandidate->GetItemText(i, 1) != label)
    {
      m_listCtrlCandidate->SetItem(i, 1, label);
    }

    if (m_listCtrlCandidate->GetItemText(i, 2) != strDataX)
    {
      m_listCtrlCandidate->SetItem(i, 2, strDataX);
    }

    if (m_listCtrlCandidate->GetItemText(i, 3) != strCount)
    {
      m_listCtrlCandidate->SetItem(i, 3, strCount);
    }
  }
}

void CSVExportOptionDlg::UpdateExportedItem()
{
  wxASSERT(m_proj != NULL);

  // ACe̐킹
  while (m_listCtrlExportedItem->GetItemCount() > static_cast<int>(m_option.exportedData.size()))
  {
    m_listCtrlExportedItem->DeleteItem(0);
  }
  while (m_listCtrlExportedItem->GetItemCount() < static_cast<int>(m_option.exportedData.size()))
  {
    m_listCtrlExportedItem->InsertItem(0, _T(""));
  }

  // \̓eݒ (ύXӏ̂ݐݒ)
  for (int i = 0; i < static_cast<int>(m_option.exportedData.size()); ++i)
  {
    wxString strNo, label, header;

    CSVExportedDataInfo info = m_option.exportedData[i];
    int index = info.graphDataIndex;
    GraphDataRefPtr graphData = m_proj->GetItem(index);

    strNo.Printf(_T("%d"), index + 1);
    label = graphData->GetLabel();
    header = info.headerText;

    if (m_listCtrlExportedItem->GetItemText(i, 0) != strNo)
    {
      m_listCtrlExportedItem->SetItem(i, 0, strNo);
    }

    if (m_listCtrlExportedItem->GetItemText(i, 1) != label)
    {
      m_listCtrlExportedItem->SetItem(i, 1, label);
    }

    if (m_listCtrlExportedItem->GetItemText(i, 2) != header)
    {
      m_listCtrlExportedItem->SetItem(i, 2, header);
    }
  }
}

void CSVExportOptionDlg::UpdateExportedItemHeader()
{
  long item = m_listCtrlExportedItem->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);

  if (-1 == item)
  {
    m_textCtrlHeader->Disable();
  }
  else
  {
    wxASSERT(item >= 0 && item < static_cast<long>(m_option.exportedData.size()));
    m_textCtrlHeader->SetValue(m_option.exportedData[item].headerText);
    m_textCtrlHeader->Enable();
  }
}

void CSVExportOptionDlg::UpdateOptions()
{
  if (m_checkBoxExportHeader->GetValue() != m_option.exportHeader)
  {
    m_checkBoxExportHeader->SetValue(m_option.exportHeader);
  }

  if (m_checkBoxExportDataX->GetValue() != m_option.exportDataX)
  {
    m_checkBoxExportDataX->SetValue(m_option.exportDataX);
  }

  m_textCtrlDataXHeader->SetValue(m_option.dataXHeaderText);
  m_textCtrlDataXHeader->Enable(m_option.exportDataX);
}

//--------------------------------------------------------------------------------
