#include "CSVImportOptionDlg.h"

//------------------------------------------------------------------------
CSVImportOptionDlg::CSVImportOptionDlg(wxWindow* parent)
  : CSVImportOptionDlgDesign(parent)
{
  m_listCtrlDataSeries->AppendColumn(_("Name"));
  m_listCtrlDataSeries->AppendColumn(_("X Data"));
  m_listCtrlDataSeries->AppendColumn(_("Y Data"));
  
  m_spinCtrlXStart->SetRange(-DBL_MAX, DBL_MAX);
  m_spinCtrlXInterval->SetRange(1.0e-10, DBL_MAX);
  m_spinCtrlXStart->SetValue(0.0);
  m_spinCtrlXInterval->SetValue(1.0);

  // wxSpinCtrlDouble  wxFormBuilder ŃT|[gĂȂ̂,
  // ŃCxgڑ.
  m_spinCtrlXStart->Connect(::wxEVT_COMMAND_SPINCTRLDOUBLE_UPDATED,
    wxSpinDoubleEventHandler(CSVImportOptionDlg::SpinCtrlXStart_OnSpinCtrlDouble), NULL, this);
  m_spinCtrlXInterval->Connect(::wxEVT_COMMAND_SPINCTRLDOUBLE_UPDATED,
    wxSpinDoubleEventHandler(CSVImportOptionDlg::SpinCtrlXInterval_OnSpinCtrlDouble), NULL, this);
}

CSVImportOptionDlg::~CSVImportOptionDlg()
{
  m_spinCtrlXStart->Disconnect(::wxEVT_COMMAND_SPINCTRLDOUBLE_UPDATED,
    wxSpinDoubleEventHandler(CSVImportOptionDlg::SpinCtrlXStart_OnSpinCtrlDouble), NULL, this);
  m_spinCtrlXInterval->Disconnect(::wxEVT_COMMAND_SPINCTRLDOUBLE_UPDATED,
    wxSpinDoubleEventHandler(CSVImportOptionDlg::SpinCtrlXInterval_OnSpinCtrlDouble), NULL, this);
}

void CSVImportOptionDlg::SetFileInfo(CSVFileInfo const &info)
{
  wxString strRows, strCols;
  strRows.Printf(_T("%d"), info.rowCount);
  strCols.Printf(_T("%d"), info.columnCount);
  
  // x̐ݒ
	m_staticTextFileName->SetLabel(info.fileName);
	m_staticTextRows->SetLabel(strRows);
	m_staticTextColumns->SetLabel(strCols);

  // XsRg[̃Wݒ
  m_spinCtrlHeaderRows->SetRange(0, info.rowCount);
  //m_spinCtrlHeaderRows->SetValue(0);
  m_spinCtrlDataRows->SetRange(1, info.rowCount);
  //m_spinCtrlDataRows->SetValue(info.rowCount);

  m_spinCtrlXDataColumn->SetRange(1, info.columnCount);
  m_spinCtrlYDataColumn->SetRange(1, info.columnCount);

  m_info = info;
}

void CSVImportOptionDlg::SetImportOption(CSVImportOption const &option)
{
  m_option = option;
  
  UpdateRows();
  UpdateDataSeries();
  UpdateDataItem();
}

void CSVImportOptionDlg::GetImportOption(CSVImportOption &option) const
{
  option = m_option;
}

void CSVImportOptionDlg::SpinCtrlHeaderRows_OnSpinCtrl(wxSpinEvent &event)
{
  m_option.headerCount = m_spinCtrlHeaderRows->GetValue();
  if (m_option.headerCount + m_option.rowCount >  m_info.rowCount)
  {
    m_option.rowCount = m_info.rowCount - m_option.headerCount;
    UpdateRows();
  }
}

void CSVImportOptionDlg::SpinCtrlDataRows_OnSpinCtrl(wxSpinEvent &event)
{
  m_option.rowCount = m_spinCtrlDataRows->GetValue();
  if (m_option.headerCount + m_option.rowCount >  m_info.rowCount)
  {
    m_option.headerCount = m_info.rowCount - m_option.rowCount;
    UpdateRows();
  }
}

void CSVImportOptionDlg::ListCtrlDataSeries_OnListItemDeselected(wxListEvent &event)
{
  UpdateDataItem();
}

void CSVImportOptionDlg::ListCtrlDataSeries_OnListItemSelected(wxListEvent &event)
{
  UpdateDataItem();
}

void CSVImportOptionDlg::ButtonAdd_OnButtonClick(wxCommandEvent &event)
{
  CSVDataSourceInfo newItem;
  newItem.label = _("Untitled");
  newItem.columnX = 0;
  newItem.columnY = 1;
  newItem.useDataX = false;

  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item)
  {
    // IĂȂƂ, ԍŌɒǉ
    m_option.dataSources.push_back(newItem);
  }
  else
  {
    // IĂƂ, IACěɒǉ
    wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));

    m_option.dataSources.insert(m_option.dataSources.begin() + item + 1, newItem);
  }

  UpdateDataSeries();
  UpdateDataItem();
}

void CSVImportOptionDlg::ButtonRemove_OnButtonClick(wxCommandEvent &event)
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item)
  {
    wxMessageBox(_("Please select a item"), _("Warning"));
  }
  else
  {
    wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));
    m_option.dataSources.erase(m_option.dataSources.begin() + item);
  }

  UpdateDataSeries();
  UpdateDataItem();
}

void CSVImportOptionDlg::ButtonMoveUp_OnButtonClick(wxCommandEvent &event)
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;
  
  long itemUp = item - 1;
  if (itemUp < 0) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));
  wxASSERT(itemUp >= 0 && itemUp < static_cast<long>(m_option.dataSources.size()));
  std::swap(m_option.dataSources[item], m_option.dataSources[itemUp]);

  m_listCtrlDataSeries->SetItemState(item, 0, wxLIST_STATE_SELECTED);
  m_listCtrlDataSeries->SetItemState(itemUp, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);

  UpdateDataSeries();
  UpdateDataItem();
}

void CSVImportOptionDlg::ButtonMoveDown_OnButtonClick(wxCommandEvent &event)
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;
  
  long itemDown = item + 1;
  if (itemDown >= static_cast<long>(m_option.dataSources.size())) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));
  wxASSERT(itemDown >= 0 && itemDown < static_cast<long>(m_option.dataSources.size()));
  std::swap(m_option.dataSources[item], m_option.dataSources[itemDown]);

  m_listCtrlDataSeries->SetItemState(item, 0, wxLIST_STATE_SELECTED);
  m_listCtrlDataSeries->SetItemState(itemDown, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);

  UpdateDataSeries();
  UpdateDataItem();
}

void CSVImportOptionDlg::TextCtrlLabel_OnText(wxCommandEvent &event)
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));
  m_option.dataSources[item].label = m_textCtrlLabel->GetValue();
  UpdateDataSeries();
}

void CSVImportOptionDlg::SpinCtrlXDataColumn_OnSpinCtrl(wxSpinEvent &event)
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));
  m_option.dataSources[item].columnX = m_spinCtrlXDataColumn->GetValue() - 1;
  UpdateDataSeries();
}

void CSVImportOptionDlg::SpinCtrlXStart_OnSpinCtrlDouble(wxSpinDoubleEvent &event)
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));
  m_option.dataSources[item].dataX.begin = m_spinCtrlXStart->GetValue();
  UpdateDataSeries();
}

void CSVImportOptionDlg::SpinCtrlXInterval_OnSpinCtrlDouble(wxSpinDoubleEvent &event)
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));
  m_option.dataSources[item].dataX.interval = m_spinCtrlXInterval->GetValue();
  UpdateDataSeries();
}

void CSVImportOptionDlg::SpinCtrlYDataColumn_OnSpinCtrl(wxSpinEvent &event)
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));
  m_option.dataSources[item].columnY = m_spinCtrlYDataColumn->GetValue() - 1;
  UpdateDataSeries();
}

void CSVImportOptionDlg::CheckBoxUseDirectVal_OnCheckBox(wxCommandEvent &event)
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item) return;

  wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));
  m_option.dataSources[item].useDataX = m_checkBoxUseDirectVal->GetValue();
  UpdateDataSeries();
  UpdateDataItem();
}

void CSVImportOptionDlg::UpdateRows()
{
  if (m_spinCtrlHeaderRows->GetValue() != m_option.headerCount)
  {
    m_spinCtrlHeaderRows->SetValue(m_option.headerCount);
  }

  if (m_spinCtrlDataRows->GetValue() != m_option.rowCount)
  {
    m_spinCtrlDataRows->SetValue(m_option.rowCount);
  }
}

void CSVImportOptionDlg::UpdateDataSeries()
{
  // ACe̐킹
  while (m_listCtrlDataSeries->GetItemCount() > static_cast<int>(m_option.dataSources.size()))
  {
    m_listCtrlDataSeries->DeleteItem(0);
  }
  while (m_listCtrlDataSeries->GetItemCount() < static_cast<int>(m_option.dataSources.size()))
  {
    m_listCtrlDataSeries->InsertItem(0, _T(""));
  }

  // \̓eݒ (ύXӏ̂ݐݒ)
  for (size_t i = 0; i < m_option.dataSources.size(); ++i)
  {
    CSVDataSourceInfo info = m_option.dataSources[i];
    if (m_listCtrlDataSeries->GetItemText(i, 0) != info.label)
    {
      m_listCtrlDataSeries->SetItem(i, 0, info.label);
    }
    
    wxString strXData;
    if (info.useDataX)
    {
      strXData.Printf(_T("S: %f, I: %f"), info.dataX.begin, info.dataX.interval);
    }
    else
    {
      strXData.Printf(_T("Column %d"), info.columnX + 1);
    }
    if (m_listCtrlDataSeries->GetItemText(i, 1) != strXData)
    {
      m_listCtrlDataSeries->SetItem(i, 1, strXData);
    }

    wxString strYData;
    strYData.Printf(_T("Column %d"), info.columnY + 1);
    if (m_listCtrlDataSeries->GetItemText(i, 2) != strYData)
    {
      m_listCtrlDataSeries->SetItem(i, 2, strYData);
    }
  }
}

void CSVImportOptionDlg::UpdateDataItem()
{
  long item = m_listCtrlDataSeries->GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
  if (-1 == item)
  {
    m_textCtrlLabel->Disable();
    m_spinCtrlXDataColumn->Disable();
    m_spinCtrlXStart->Disable();
    m_spinCtrlXInterval->Disable();
    m_spinCtrlYDataColumn->Disable();
    m_checkBoxUseDirectVal->Disable();
  }
  else
  {
    wxASSERT(item >= 0 && item < static_cast<long>(m_option.dataSources.size()));

    // TextCtrlLabel ̐ݒ
    m_textCtrlLabel->Enable();
    if (m_textCtrlLabel->GetValue() != m_option.dataSources[item].label)
    {
      m_textCtrlLabel->SetValue(m_option.dataSources[item].label);
    }

    // SpinCtrlXDataColumn ̐ݒ
    if (m_option.dataSources[item].useDataX)
    {
      m_spinCtrlXDataColumn->Disable();
    }
    else
    {
      m_spinCtrlXDataColumn->Enable();
      int column = m_option.dataSources[item].columnX + 1;
      if (m_spinCtrlXDataColumn->GetValue() != column)
      {
        m_spinCtrlXDataColumn->SetValue(column);
      }
    }

    // SpinCtrlXStart ̐ݒ
    if (m_option.dataSources[item].useDataX)
    {
      m_spinCtrlXStart->Enable();
      if (m_spinCtrlXStart->GetValue() != m_option.dataSources[item].dataX.begin)
      {
        m_spinCtrlXStart->SetValue(m_option.dataSources[item].dataX.begin);
      }
    }
    else
    {
      m_spinCtrlXStart->Disable();
    }

    // SpinCtrlXInterval ̐ݒ
    if (m_option.dataSources[item].useDataX)
    {
      m_spinCtrlXInterval->Enable();
      if (m_spinCtrlXInterval->GetValue() != m_option.dataSources[item].dataX.interval)
      {
        m_spinCtrlXInterval->SetValue(m_option.dataSources[item].dataX.interval);
      }
    }
    else
    {
      m_spinCtrlXInterval->Disable();
    }
    
    // SpinCtrlYDataColumn ̐ݒ
    m_spinCtrlYDataColumn->Enable();
    int column = m_option.dataSources[item].columnY + 1;
    if (m_spinCtrlYDataColumn->GetValue() != column)
    {
      m_spinCtrlYDataColumn->SetValue(column);
    }
    
    // CheckBoxUseDirectVal ̐ݒ
    m_checkBoxUseDirectVal->Enable();
    if (m_checkBoxUseDirectVal->GetValue() != m_option.dataSources[item].useDataX)
    {
      m_checkBoxUseDirectVal->SetValue(m_option.dataSources[item].useDataX);
    }
  }
}

//------------------------------------------------------------------------