#include "CSVImporter.h"
#include <functional>
#include <fstream>
#include <cstdlib>

#define LINE_BUF_SIZE 1024

//--------------------------------------------------------------------------------
// t@C̗eʂmF܂.
static long GetFileSize(FILE *fp)
{
  wxASSERT(fp != NULL);

  long cur = ::ftell(fp);

  ::fseek(fp, 0, SEEK_SET);
  long beg = ::ftell(fp);

  ::fseek(fp, 0, SEEK_END);
  long end = ::ftell(fp);

  ::fseek(fp, cur, SEEK_SET);

  return end - beg;
}

// J}ŋ؂ꂽl\镶ł邩ǂ𔻒f܂.
// Ȃ킿, 󔒂܂̓Zp[^, strtod ŕϊ\ȕł邩ǂmF܂.
// [whitespace] [sign] [digits] [.digits] [ {d | D | e | E}[sign]digits]
inline bool IsCSVChar(char c, char sep)
{
  return (' ' == c || '\t' == c ||
          sep == c ||
          '+' == c || '-' == c ||
          (c >= '0' && c <= '9') ||
          '.' == c ||
          'd' == c || 'D' == c || 'e' == c || 'E' == c);
}

// Zp[^ŋ؂ꂽl\镶ł邩ǂ𔻒f܂.
// strtod ŕϊ\ȕł邩ǂmF܂.
static bool IsCSVString(char const *str, char sep)
{
  char const *p = str;
  while (*p != '\0')
  {
    if (!IsCSVChar(*p, sep))
    {
      return false;
    }
    ++p;
  }
  return true;
}

// ̃Zp[^̐𐔂܂.
static int CountComma(char const *str, char sep)
{
  int count = 0;
  char const *p = str;
  while (*p != '\0')
  {
    if (sep == *p)
    {
      ++count;
    }
    ++p;
  }
  return count;
}

// Zp[^؂̕𕶎̔zɕ܂.
static void SplitLineToStringArray(char const *str, wxArrayString &strings, char sep)
{
  strings.clear();

  char const *p = str;
  char const *beg = str;
  while (*p != '\0')
  {
    if (sep == *p)
    {
      size_t len = p - beg;
      strings.push_back(wxString(beg, len));
      beg = p + 1;
    }
    ++p;
  }
  size_t len = p - beg;
  strings.push_back(wxString(beg, len));
}

// Zp[^؂̕𐔒lɕϊ܂.
static bool ConvLineToValue(char const *line, int row, std::vector<GraphDataYRefPtr> data, char sep)
{
  char const *p = line;
  int column = 0;
  while (*p != '\0')
  {
    char *q;
    double v = strtod(p, &q);
    if (p == q)
    {
      // ϊs
      return false;
    }
    else if (sep == *q)
    {
      // ǂݎ萬
      if (column >= static_cast<int>(data.size()))
      {
        // t@C̗񐔂z (XL) 葽Ƃ
        // (XLt@CƋN)
        return false;
      }

      if (data[column] != NULL)
      {
        // f[^̊i[悪Ƃ
        // (ǂݍ݂KvȂɂĂ͊i[悪mۂĂȂ)

        // s͈͓ǂmF.
        // s͈̔̓`FbN, ĂяoƂōs͂ł̂,
        // ł̓AT[VƂ.
        wxASSERT(row < static_cast<int>(data[column]->raw.size()));
        data[column]->raw[row] = v;
      }

      ++column;
      p = q + 1;
    }
    else if ('\0' == *q)
    {
      // ǂݎ萬
      if (column >= static_cast<int>(data.size()))
      {
        // t@C̗񐔂z (XL) 葽Ƃ
        // (XLt@CƋN)
        return false;
      }

      if (data[column] != NULL)
      {
        // f[^̊i[悪Ƃ
        // (ǂݍ݂KvȂɂĂ͊i[悪mۂĂȂ)

        // s͈͓ǂmF.
        // s͈̔̓`FbN, ĂяoƂōs͂ł̂,
        // ł̓AT[VƂ.
        wxASSERT(row < static_cast<int>(data[column]->raw.size()));
        data[column]->raw[row] = v;
      }

      ++column;
      p = q;
    }
    else
    {
      // G[
      return false;
    }
  }
  return true;
}

// ł邱ƂmF܂.
static bool ConfirmProgression(std::vector<double> const &data, GraphDataX &progr, double epsPercent)
{
  if (data.empty())
  {
    return false;
  }
  else if (1 == data.size())
  {
    progr.begin = data[0];
    progr.interval = 0.0;
    return true;
  }
  else
  {
    progr.begin = data[0];
    progr.interval = data[1] - data[0];
    double eps = progr.interval * epsPercent / 100.0;
    double prev = data[0];
    for (std::vector<double>::const_iterator i = data.begin() + 1; i != data.end(); ++i)
    {
      double diff = (*i - prev) - progr.interval;
      if (diff > eps || diff < -eps)
      {
        return false;
      }
      prev = *i;
    }
    return true;
  }
}

//--------------------------------------------------------------------------------
void CSVImportOption::SetDefaultFromFileInfo(CSVFileInfo const &info)
{
  headerCount = 0;
  rowCount = info.rowCount;

  dataSources.clear();
  for (int i = 1; i < info.columnCount; ++i)
  {
    CSVDataSourceInfo source;
    source.useDataX = false;
    source.columnX = 0;
    source.columnY = i;
    source.label.Printf(_("Column %d"), source.columnY + 1);
    dataSources.push_back(source);
  }
}

void CSVImportOption::SetDefaultFromFileInfo(CSVFileInfo const &info, int headers, wxArrayString const &headerStrings)
{
  SetDefaultFromFileInfo(info);
  headerCount = headers;
  rowCount -= headers;

  for (int i = 0; i < static_cast<int>(dataSources.size()); ++i)
  {
    int labelIndex = i + 1;
    if (labelIndex < static_cast<int>(headerStrings.size()))
    {
      dataSources[i].label = headerStrings[labelIndex];
    }
  }
}

#if !defined(NDEBUG)
void CSVImportOption::LogDebug() const
{
  wxLogDebug("headerCount = %d, rowCount = %d", headerCount, rowCount);
  for (size_t i = 0; i < dataSources.size(); ++i)
  {
    CSVDataSourceInfo info = dataSources[i];
    wxLogDebug("dataSources[%d]: useDataX = %d, dataX = { %f, %f }, columnX = %d, columnY = %d, label = %s",
      i, info.useDataX, info.dataX.begin, info.dataX.interval, info.columnX, info.columnY, info.label);
  }
}
#endif

//--------------------------------------------------------------------------------
CSVImporter::CSVImporter(wxString const &fileName, GraphProject &project)
  : m_project(project),
    m_fileName(fileName),
    m_separator(_T(','))
{
}

wxString CSVImporter::GetFileName() const
{
  return m_fileName;
}

ResultCode CSVImporter::ScanInfo(ProcContext *pPC)
{
  using namespace std::placeholders;

  ResultCode result = RunProcWithContext<ResultCode>
    (std::bind(&CSVImporter::ScanInfoInternal, this, _1), pPC);

  return result;
}

ResultCode CSVImporter::ScanInfoInternal(Progress *pProgress)
{
  char sep;
  bool b = m_separator.GetAsChar(&sep);
  wxASSERT(b);

  // t@CJ.
  std::string fileNameLC = m_fileName.ToStdString();  // current locale

  FILE *file = ::fopen(fileNameLC.c_str(), "r");
  if (NULL == file)
  {
    m_err.Printf(_("Can't open \"%s\"!"), m_fileName);
    return ResultCode::FILE_ERROR;
  }

  // t@C̗eʂmF.
  long fileSize = GetFileSize(file);

  // XL.
  std::string lastHeader;
  int headerCount = 0;
  bool header = true;
  int rowCount = 0;
  int columnCount = 0;
  char line[LINE_BUF_SIZE + 8];
  int const updateInterval = 1000;

  if (pProgress != NULL)
  {
    wxString message;
    message.Printf(_T("Scan \"%s\" ..."), m_fileName);
    pProgress->UpdateMessage(message);
  }
  long posBegin = ::ftell(file);

  while (::fgets(line, LINE_BUF_SIZE, file) != NULL)
  {
    // ǂݍ񂾍s̏I[sł邱ƂmF
    size_t lineSize = ::strlen(line);
    if (line[lineSize - 1] != '\n')
    {
      m_err.Printf(_("The length of the line in CSV file is too much at %s:%d!"), m_fileName, rowCount + 1);
      ::fclose(file);
      return ResultCode::FILE_ERROR;
    }
    else
    {
      // I[̉sĂ.
      line[lineSize - 1] = '\0';
      --lineSize;
    }
    
    // wb_[𔻒肵܂.
    if (header)
    {
      if (IsCSVString(line, sep))
      {
        // wb_[IƔ
        header = false;
      }
      else
      {
        lastHeader = line;
        headerCount++;
      }
    }

    // s, 񐔂JEg܂.
    if (!header)
    {
      int curColumnCount = CountComma(line, sep) + 1;
      if (curColumnCount > columnCount)
      {
        columnCount = curColumnCount;
      }
    }

    if (0 == rowCount % updateInterval && pProgress != NULL)
    {
      // vOXo[XV.
      long posCurrent = ::ftell(file);
      pProgress->UpdateProgress(posCurrent - posBegin, fileSize);

      // fmF
      if (pProgress->IsAbortRequested())
      {
        ::fclose(file);
        return ResultCode::ABORT;
      }
    }

    rowCount++;
  }
  
  ::fclose(file);

  // t@Cݒ肵܂.
  m_info.fileName = m_fileName;
  m_info.rowCount = rowCount;
  m_info.columnCount = columnCount;

  // ftHg̃IvVݒ肵܂.
  if (0 == headerCount)
  {
    m_option.SetDefaultFromFileInfo(m_info);
  }
  else
  {
    wxArrayString headerStrings;
    SplitLineToStringArray(lastHeader.c_str(), headerStrings, sep);
    m_option.SetDefaultFromFileInfo(m_info, headerCount, headerStrings);
  }

  return ResultCode::SUCCESS;
}

void CSVImporter::GetFileInfo(CSVFileInfo &info) const
{
  info = m_info;
}

void CSVImporter::SetOption(CSVImportOption const &option)
{
  m_option = option;
}

void CSVImporter::GetOption(CSVImportOption &option) const
{
  option = m_option;
}

bool CSVImporter::IsOptionValid() const
{
  if (m_option.headerCount + m_option.rowCount > m_info.rowCount)
  {
    return false;
  }
  
  for (int i = 0; i < static_cast<int>(m_option.dataSources.size()); ++i)
  {
    if (!m_option.dataSources[i].useDataX)
    {
      if (m_option.dataSources[i].columnX >= m_info.columnCount)
      {
        return false;
      }
    }
    if (m_option.dataSources[i].columnY >= m_info.columnCount)
    {
      return false;
    }
  }

  return true;
}

void CSVImporter::SetSeparator(wxUniChar separator)
{
  m_separator = separator;
}

wxUniChar CSVImporter::GetSeparator() const
{
  return m_separator;
}

ResultCode CSVImporter::Import(int mipmapResolution, ProcContext *pPC)
{
  using namespace std::placeholders;

  // C|[g. ʂ m_refDataArray Ɋi[.
  ResultCode result = RunProcWithContext<ResultCode>
    (std::bind(&CSVImporter::ImportInternal, this, _1), pPC);
  if (result != ResultCode::SUCCESS)
  {
    m_refDataArray.clear();
    return result;
  }

  // ~bv}bvvZ.
  for (size_t i = 0; i < m_refDataArray.size(); ++i)
  {
    result = m_refDataArray[i]->CreateMipmapForResolution(mipmapResolution, pPC);
    if (result != ResultCode::SUCCESS)
    {
      m_refDataArray.clear();
      m_err = _("Failed to allocate memory!");
      return result;
    }
  }

  // vWFNgɒǉ.
  for (size_t i = 0; i < m_refDataArray.size(); ++i)
  {
    m_project.Append(m_refDataArray[i]);
    GraphDataRefPtr graph = m_project.GetItem(m_project.GetCount() - 1);
    graph->SetLabel(m_option.dataSources[i].label);
  }

  return result;
}

ResultCode CSVImporter::ImportInternal(Progress *pProgress)
{
  try
  {
    char sep;
    bool b = m_separator.GetAsChar(&sep);
    wxASSERT(b);

    // w肳ꂽIvVLǂmF.
    wxASSERT(IsOptionValid());

    // t@CI[v.
    std::string fileNameLC = m_fileName.ToStdString();  // current locale

    FILE *file = ::fopen(fileNameLC.c_str(), "r");
    if (NULL == file)
    {
      m_err.Printf(_("Can't open \"%s\"!"), m_fileName);
      return ResultCode::FILE_ERROR;
    }

    // t@C̗eʂmF.
    long fileSize = GetFileSize(file);

    // KvƂȂ𒲂ׂ.
    std::vector<bool> requiredColumn;
    CheckRequiredColumn(requiredColumn, true, true);

    // f[^̏ݐpӂ.
    std::vector<GraphDataYRefPtr> dataYs;
    dataYs.resize(m_info.columnCount);
    for (size_t i = 0; i < dataYs.size(); ++i)
    {
      if (requiredColumn[i])
      {
        // mۂ, [tB.
        // XLɗ񐔂͍ő (sƂɗ񐔂قȂĂꍇ) 擾邽,
        // 񂪑݂Ȃ̃f[^`ƂȂ\邽, [tB.
        dataYs[i] = GraphDataY::Create();
        dataYs[i]->raw.resize(m_option.rowCount, 0.0);
      }
    }
  
    // Kvłf[^ǂݍ.
    if (pProgress != NULL)
    {
      wxString message;
      message.Printf(_("Import \"%s\" ..."), m_fileName);
      pProgress->UpdateMessage(message);
    }
    long posBegin = ::ftell(file);
    int const updateInterval = 1000;

    int row = 0;
    char line[LINE_BUF_SIZE + 8];

    while (::fgets(line, LINE_BUF_SIZE, file) != NULL)
    {
      // ǂݍ񂾍s̏I[sł邱ƂmF
      size_t lineSize = ::strlen(line);
      if (line[lineSize - 1] != '\n')
      {
        m_err.Printf(_("The length of the line in CSV file is too much at %s:%d!"), m_fileName, row + 1);
        ::fclose(file);
        return ResultCode::FILE_ERROR;
      }
      else
      {
        // I[̉sĂ.
        line[lineSize - 1] = '\0';
        --lineSize;
      }

      if (row < m_option.headerCount)
      {
        // wb_[ǂݔ΂.
      }
      else if (row < m_option.headerCount + m_option.rowCount)
      {
        if (!ConvLineToValue(line, row - m_option.headerCount, dataYs, sep))
        {
          m_err.Printf(_("Conversion from string to numerical value is failed at %s:%d!"), m_fileName, row + 1);
          ::fclose(file);
          return ResultCode::FILE_ERROR; // ϊG[
        }
      }

      if (0 == row % updateInterval && pProgress != NULL)
      {
        // vOXo[XV.
        long posCurrent = ::ftell(file);
        pProgress->UpdateProgress(posCurrent - posBegin, fileSize);

        // fmF
        if (pProgress->IsAbortRequested())
        {
          ::fclose(file);
          return ResultCode::ABORT;
        }
      }

      ++row;
    }
  
    ::fclose(file);

    // ł邩ǂ`FbNKvƂȂ𒲂ׂ.
    std::vector<bool> requiredCheckProg;
    CheckRequiredColumn(requiredCheckProg, true, false);
  
    // ł邩ǂ`FbN.
    std::vector<GraphDataX> dataXs;
    dataXs.resize(m_info.columnCount);
    for (int i = 0; i < m_info.columnCount; ++i)
    {    
      if (requiredCheckProg[i])
      {
        wxASSERT(dataYs[i] != NULL);
        if (!ConfirmProgression(dataYs[i]->raw, dataXs[i], 0.1))
        {
          m_err.Printf(_("Column %d is not arithmetic progression!"), i + 1);
          return ResultCode::FILE_ERROR;
        }
      }
    }

    // m_refDataArray ɒǉ.
    m_refDataArray.clear();
    for (size_t i = 0; i < m_option.dataSources.size(); ++i)
    {
      CSVDataSourceInfo info = m_option.dataSources[i];

      // Otf[^𐶐
      GraphDataX dataX;
      if (info.useDataX)
      {
        dataX = info.dataX;
      }
      else
      {
        wxASSERT(info.columnX < static_cast<int>(dataXs.size()));
        dataX = dataXs[info.columnX];
      }

      wxASSERT(info.columnY < static_cast<int>(dataYs.size()));
      GraphDataYRefPtr dataY = dataYs[info.columnY];
      wxASSERT(dataY.get() != NULL);

      GraphDataRefPtr refData = GraphData::Create(dataX, dataY);
      m_refDataArray.push_back(refData);
    }

    return ResultCode::SUCCESS;
  }
  catch (std::bad_alloc &)
  {
    // s
    m_err = _("Failed to allocate memory in the calculation process!");
    return ResultCode::MEMORY_ERROR;
  }
}

wxString CSVImporter::GetLastErrorMessage() const
{
  return m_err;
}

// KvƂȂ`FbN܂.
void CSVImporter::CheckRequiredColumn(std::vector<bool> &requiredColumn, bool asDataX, bool asDataY)
{
  requiredColumn.resize(m_info.columnCount);
  std::fill(requiredColumn.begin(), requiredColumn.end(), false);
  
  for (size_t i = 0; i < m_option.dataSources.size(); ++i)
  {
    if (!m_option.dataSources[i].useDataX && asDataX)
    {
      int col = m_option.dataSources[i].columnX;
      wxASSERT(col >= 0 && col < m_info.columnCount);
      requiredColumn[col] = true;
    }
    if (asDataY)
    {
      int col = m_option.dataSources[i].columnY;
      wxASSERT(col >= 0 && col < m_info.columnCount);
      requiredColumn[col] = true;
    }
  }
}

//--------------------------------------------------------------------------------
