#pragma once

//--------------------------------------------------------------------------------
#include <vector>
#include <map>
#include <iostream>
#include "FormulaAST.h"
#include "FormulaUtil.h"

/**
 *  LL\͊ƂĎp[T[.
 *  \͂, ۍ\ (abstract syntax tree, AST) 𐶐.
 *  ꂽ AST pčɐ̉Zs.
 *
 *  @tparam T vZƂɗp^.
 *  ݎ̌^z肵Ă.
 *  ׂĂ̐^, ^, std::complex<float>, std::complex<double>.
 *
 *  \note
 *  @`
 *    formula     ::= relation1 ("?" relation1 ":" relation1)*
 *    relation1   ::= relation2 (("&&" | "||") relation2)*
 *    relation2   ::= expression (("<" | ">" | "<=" | ">=" | "==" | "!=") expression)*
 *    expression  ::= term (("+" | "-") term)*
 *    term        ::= sign (("*" | "/" | "%") sign)*
 *    sign        ::= ("+" | "-" | "!"*) factor
 *    factor      ::= NUMBER |
 *                    "(" formula ")" |
 *                    FUNCTION0_ID "()" |
 *                    FUNCTION1_ID "(" formula ")" |
 *                    FUNCTION2_ID "(" formula "," formula ")" |
 *                    FUNCTION3_ID "(" formula "," formula "," formula ")" |
 *                    FUNCTION4_ID "(" formula "," formula "," formula "," formula ")" |
 *                    PARAMETER_ID |
 *                    LITERAL_ID
 *
 *  JnL
 *    formula
 *
 *  Zq̗Dx
 *    ()1:  ()  ֐ĂяoZq
 *        2:  +	  P+Zq
 *        2:  -   P-Zq
 *        2:  !   _ے艉Zq
 *        3:  *   Q*Zq
 *        3:  /   /Zq
 *        3:  %   %Zq
 *        4:  +   Q+Zq
 *        4:  -   Q-Zq
 *        5:  <   <Zq
 *        5:  <=  <=Zq
 *        5:  >   >Zq
 *        5:  >=  >=Zq
 *        5:  ==  ==Zq
 *        5:  !=  !=Zq
 *        6:  &&  _ANDZq
 *        6:  ||  _ORZq
 *    ()7:  ? : Zq
 *
 */
template <typename T>
class FormulaParser /*: private boost::noncopyable*/
{
public:
  typedef typename FormulaAst::Base<T>::RefPtr AstRefPtr;

private:
  typedef std::map<std::wstring, T const *> ParameterMap;
  typedef std::map<std::wstring, T> LiteralMap;
  typedef std::map<std::wstring, typename FormulaAst::Func0CreatorBase<T>::RefPtr> Func0Map;
  typedef std::map<std::wstring, typename FormulaAst::Func1CreatorBase<T>::RefPtr> Func1Map;
  typedef std::map<std::wstring, typename FormulaAst::Func2CreatorBase<T>::RefPtr> Func2Map;
  typedef std::map<std::wstring, typename FormulaAst::Func3CreatorBase<T>::RefPtr> Func3Map;
  typedef std::map<std::wstring, typename FormulaAst::Func4CreatorBase<T>::RefPtr> Func4Map;
  
  struct Token
  {
    std::wstring str; //  g[N
    int line;         //  g[N̂s (0 ͂߂)
    int pos;          //  g[N̂s̈ʒu (0 ͂߂)
    
    Token(std::wstring const &s, int l, int p) : str(s), line(l), pos(p) { }
  };
  
  typedef std::vector<Token> TokenArray;
  
public:
  /**
   *  RXgN^.
   */
  FormulaParser();

  /**
   *  p[X, IuWFNg𐶐.
   *  @param str  .
   *  @return     IuWFNgԂ. sƂ NULL ƂȂ.
   */
  AstRefPtr parse(std::wstring const &str);
  
  /**
   *  p[XI̍s擾.
   *  @return p[X (܂͒f) ̍s.
   */
  int getLine() const;
  
  /**
   *  p[XI̍sʒu擾.
   *  @return p[X (܂͒f) ̍sʒu.
   */
  int getLinePosition() const;
  
  /**
   *  Ŏgp\ȕϐݒ肷.
   *  p[Xɂ萶 AST ŐvZƂ, \a pParameter ŎQƂlp.
   *  @param name       ϐ.
   *  @param pParameter ̌vZŗpϐ̃|C^.
   */
  bool setParameter(std::wstring const &name, T const *pParameter);
  
  /**
   *  Ŏgp\Ȓ萔ݒ肷.
   *  @param name   萔.
   *  @param value  萔l.
   */
  bool setLiteral(std::wstring const &name, T const &value);
  
  /**
   *  Ŏgp\ 0 ֐ݒ肷.
   *  @param name             ֐.
   *  @param func             ֐IuWFNg.
   *  @param refTransparency  ֐ɎQƓƂ true w肷.
   */
  template <typename Functor>
  bool setFunction0(std::wstring const &name, Functor func, bool refTransparency);

  /**
   *  Ŏgp\ 1 ֐ݒ肷.
   *  @param name             ֐.
   *  @param func             ֐IuWFNg.
   *  @param refTransparency  ֐ɎQƓƂ true w肷.
   */
  template <typename Functor>
  bool setFunction1(std::wstring const &name, Functor func, bool refTransparency);

  /**
   *  Ŏgp\ 2 ֐ݒ肷.
   *  @param name             ֐.
   *  @param func             ֐IuWFNg.
   *  @param refTransparency  ֐ɎQƓƂ true w肷.
   */
  template <typename Functor>
  bool setFunction2(std::wstring const &name, Functor func, bool refTransparency);

  /**
   *  Ŏgp\ 3 ֐ݒ肷.
   *  @param name             ֐.
   *  @param func             ֐IuWFNg.
   *  @param refTransparency  ֐ɎQƓƂ true w肷.
   */
  template <typename Functor>
  bool setFunction3(std::wstring const &name, Functor func, bool refTransparency);

  /**
   *  Ŏgp\ 4 ֐ݒ肷.
   *  @param name             ֐.
   *  @param func             ֐IuWFNg.
   *  @param refTransparency  ֐ɎQƓƂ true w肷.
   */
  template <typename Functor>
  bool setFunction4(std::wstring const &name, Functor func, bool refTransparency);
  
  /**
   *  ꂽg[N󎚂.
   */
  void printTokens() const;
  
private:
  //  g[N, oϐ tokens_ ɕg[Ni[.
  void tokenize(std::wstring const &str);
  
  //  g[N̓ǂݍ݂Jn.
  void init() { iter_ = tokens_.begin(); }
  
  //  g[N̓ǂݍ݂I[ɒBǂׂ.
  bool isTerm() { return iter_ == tokens_.end(); }
  
  //  ̃g[Nǂݍ, ǂݍ݈ʒui߂. ǂݍ߂ȂƂ͋̕Ԃ.
  std::wstring next() { return isTerm() ? L"" : (iter_++)->str; }

  //  ǂݍ݈ʒu߂.
  void prev() { if (iter_ != tokens_.begin()) --iter_; }
  
  //  p[X̂߂̃o֐.
  AstRefPtr formula();
  AstRefPtr relation1();
  AstRefPtr relation2();
  AstRefPtr expression();
  AstRefPtr term();
  AstRefPtr sign();
  AstRefPtr factor();
  
private:
  //  ŗp\ȕϐ, 萔, ֐i[Rei.
  ParameterMap parameters_;
  LiteralMap literals_;
  Func0Map funcs0_;
  Func1Map funcs1_;
  Func2Map funcs2_;
  Func3Map funcs3_;
  Func4Map funcs4_;
  
  //  g[Nǂݍ݂Ɏgpϐ.
  TokenArray tokens_;
  typename TokenArray::const_iterator iter_;
};

template <typename T>
FormulaParser<T>::FormulaParser()
  : iter_(tokens_.end())
{
}

template <typename T>
typename FormulaParser<T>::AstRefPtr FormulaParser<T>::parse(std::wstring const &str)
{
  //  g[N.
  tokenize(str);
  
  //  p[X, AST 𐶐.
  init();
  AstRefPtr refTree = formula();
  if (!refTree) return AstRefPtr(); //  G[
  
  //  I[܂Ńp[XƂmF.
  if (!isTerm()) return AstRefPtr();  //  G[
  
  //  p[Xɐ.
  return refTree;
}

template <typename T>
int FormulaParser<T>::getLine() const
{
  if (tokens_.empty()) return 0;
  
  if (iter_ == tokens_.begin()) return iter_->line;
  else                          return (iter_ - 1)->line;
}

template <typename T>
int FormulaParser<T>::getLinePosition() const
{
  if (tokens_.empty()) return 0;
  
  if (iter_ == tokens_.begin()) return iter_->pos;
  else                          return (iter_ - 1)->pos;
}

template <typename T>
bool FormulaParser<T>::setParameter(std::wstring const &name, T const *pParameter)
{
  if (parameters_.find(name) == parameters_.end() && literals_.find(name) == literals_.end())
  {
    //  w肳ꂽO, ϐ, 萔ƂĂ܂gpĂȂ.
    parameters_[name] = pParameter;
    return true;
  }
  else
  {
    return false;
  }
}

template <typename T>
bool FormulaParser<T>::setLiteral(std::wstring const &name, T const &value)
{
  if (parameters_.find(name) == parameters_.end() && literals_.find(name) == literals_.end())
  {
    //  w肳ꂽO, ϐ, 萔ƂĂ܂gpĂȂ.
    literals_[name] = value;
    return true;
  }
  else
  {
    return false;
  }
}

template <typename T>
template <typename Functor>
bool FormulaParser<T>::setFunction0(std::wstring const &name, Functor func, bool refTransparency)
{
  using namespace FormulaAst;

  if (funcs0_.find(name) == funcs0_.end())
  {
    //  w肳ꂽO, 0 ֐ƂĂ܂gpĂȂ.
    funcs0_[name] = Func0Creator<T, Functor>::create(name, func, refTransparency);
    return true;
  }
  else
  {
    return false;
  }
}

template <typename T>
template <typename Functor>
bool FormulaParser<T>::setFunction1(std::wstring const &name, Functor func, bool refTransparency)
{
  using namespace FormulaAst;

  if (funcs1_.find(name) == funcs1_.end())
  {
    //  w肳ꂽO, 1 ֐ƂĂ܂gpĂȂ.
    funcs1_[name] = Func1Creator<T, Functor>::create(name, func, refTransparency);
    return true;
  }
  else
  {
    return false;
  }
}

template <typename T>
template <typename Functor>
bool FormulaParser<T>::setFunction2(std::wstring const &name, Functor func, bool refTransparency)
{
  using namespace FormulaAst;

  if (funcs2_.find(name) == funcs2_.end())
  {
    //  w肳ꂽO, 2 ֐ƂĂ܂gpĂȂ.
    funcs2_[name] = Func2Creator<T, Functor>::create(name, func, refTransparency);
    return true;
  }
  else
  {
    return false;
  }
}

template <typename T>
template <typename Functor>
bool FormulaParser<T>::setFunction3(std::wstring const &name, Functor func, bool refTransparency)
{
  using namespace FormulaAst;

  if (funcs3_.find(name) == funcs3_.end())
  {
    //  w肳ꂽO, 3 ֐ƂĂ܂gpĂȂ.
    funcs3_[name] = Func3Creator<T, Functor>::create(name, func, refTransparency);
    return true;
  }
  else
  {
    return false;
  }
}

template <typename T>
template <typename Functor>
bool FormulaParser<T>::setFunction4(std::wstring const &name, Functor func, bool refTransparency)
{
  using namespace FormulaAst;

  if (funcs4_.find(name) == funcs4_.end())
  {
    //  w肳ꂽO, 4 ֐ƂĂ܂gpĂȂ.
    funcs4_[name] = Func4Creator<T, Functor>::create(name, func, refTransparency);
    return true;
  }
  else
  {
    return false;
  }
}

template <typename T>
void FormulaParser<T>::printTokens() const
{
  for (TokenArray::const_iterator i = tokens_.begin(); i != tokens_.end(); ++i)
  {
    std::wcout << i->str << std::endl;
  }
}

template <typename T>
void FormulaParser<T>::tokenize(std::wstring const &str)
{
  int line = 0, start_pos = 0, pos = 0;

  tokens_.clear();
  std::wstring::const_iterator start = str.begin(); //  g[N̎n܂.
    
  std::wstring::const_iterator i = str.begin();
  while (i != str.end())
  {
    //  2 ̋؂蕶ɂĊmF.
    if (str.end() - i >= 2)
    {
      std::wstring tmp(i, i + 2);
      if (tmp == L"&&" || tmp == L"||" || tmp == L"<=" || tmp == L">=" ||
          tmp == L"==" || tmp == L"!=")
      {
        //  g[NƂėp؂蕶Ƃ
        if (i != start) tokens_.push_back(Token(std::wstring(start, i), line, start_pos));
        tokens_.push_back(Token(tmp, line, pos));
        
        start = i + 2;
        start_pos = pos + 2;
        
        i += 2;
        pos += 2;
        continue;
      }
      else if (tmp == L"\r\n")
      {
        //  g[NƂėp, j؂蕶Ƃ
        if (i != start) tokens_.push_back(Token(std::wstring(start, i), line, start_pos));
        
        start = i + 2;
        start_pos = 0;
        
        i += 2;
        pos = 0;
        
        ++line;
        continue;
      }
    }

    //  1 ̋؂蕶ɂĊmF.
    if (*i == L'<' || *i == L'>' || *i == L'+' || *i == L'-' || *i == L'!' ||
        *i == L'*' || *i == L'/' || *i == L'%' || *i == L'(' || *i == L')' ||
        *i == L',' || *i == L'?' || *i == L':')
    {
      //  g[NƂėp؂蕶Ƃ
      if (i != start) tokens_.push_back(Token(std::wstring(start, i), line, start_pos));
      tokens_.push_back(Token(std::wstring(1, *i), line, pos));
      
      start = i + 1;
      start_pos = pos + 1;

      ++i;
      ++pos;
    }
    else if (*i == L' ' || *i == L'\t' || *i == L'\n')
    {
      //  g[NƂėp, j؂蕶Ƃ
      if (i != start) tokens_.push_back(Token(std::wstring(start, i), line, start_pos));
      
      start = i + 1;
      start_pos = (*i == L'\n') ? 0 : pos + 1;
      
      ++i;
      pos = (*i == L'\n') ? 0 : pos + 1;  // oOۂ
      
      if (*i == L'\n') ++line;
    }
    else
    {
      ++i;
      ++pos;
    }
  }
  if (i != start) tokens_.push_back(Token(std::wstring(start, i), line, pos));
}

template <typename T>
typename FormulaParser<T>::AstRefPtr FormulaParser<T>::formula()
{
  //  @`
  //    formula     ::= relation1 ("?" relation1 ":" relation1)*
  //  ̎.
  using namespace FormulaAst;
  
  AstRefPtr refTmp = relation1();
  if (!refTmp) return AstRefPtr(); //  G[
  
  //  ܂C relations ɏԂɊi[.
  std::vector<AstRefPtr> relations;
  relations.push_back(refTmp);
  
  for ( ; ; )
  {
    std::wstring token = next();
    if (token.empty())
    {
      //  Ō܂Ńg[Nǂݍ񂾂Ƃ
      break;
    }
    else if (token == L"?")
    {
      refTmp = relation1();
      if (!refTmp) return AstRefPtr(); //  G[
      relations.push_back(refTmp);
      
      token = next();
      if (token != L":") return AstRefPtr();  //  G[
      
      refTmp = relation1();
      if (!refTmp) return AstRefPtr();  //  G[
      relations.push_back(refTmp);
    }
    else
    {
      //  ͕s\̃g[Nǂݍ񂾂Ƃ
      prev();
      break;
    }
  }
  
  //  relations 琔؂\zD
  refTmp = relations.back();
  relations.pop_back();
  while (!relations.empty())
  {
    AstRefPtr refCenter = relations.back();
    relations.pop_back();
    AstRefPtr refLeft = relations.back();
    relations.pop_back();
    refTmp = OpConditional<T>::create(refLeft, refCenter, refTmp);
  }

  return refTmp;
}

template <typename T>
typename FormulaParser<T>::AstRefPtr FormulaParser<T>::relation1()
{
  //  @`
  //    relation1   ::= relation2 (("&&" | "||") relation2)*
  //  ̎.
  using namespace FormulaAst;

  AstRefPtr refLeft = relation2();
  if (!refLeft) return AstRefPtr(); //  G[
  for ( ; ; )
  {
    std::wstring token = next();
    if (token.empty())
    {
      //  Ō܂Ńg[Nǂݍ񂾂Ƃ
      break;
    }
    else if (token == L"&&")
    {
      AstRefPtr refRight = relation2();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpAnd<T>::create(refLeft, refRight);
    }
    else if (token == L"||")
    {
      AstRefPtr refRight = relation2();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpOr<T>::create(refLeft, refRight);
    }
    else
    {
      //  ͕s\̃g[Nǂݍ񂾂Ƃ
      prev();
      break;
    }
  }
  return refLeft;
}

template <typename T>
typename FormulaParser<T>::AstRefPtr FormulaParser<T>::relation2()
{
  //  @`
  //    relation ::= expression (("<" | ">" | "<=" | ">=" | "==" | "!=") expression)*
  //  ̎.
  using namespace FormulaAst;

  AstRefPtr refLeft = expression();
  if (!refLeft) return AstRefPtr(); //  G[
  for ( ; ; )
  {
    std::wstring token = next();
    if (token.empty())
    {
      //  Ō܂Ńg[Nǂݍ񂾂Ƃ
      break;
    }
    else if (token == L"<")
    {
      AstRefPtr refRight = expression();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpLss<T>::create(refLeft, refRight);
    }
    else if (token == L">")
    {
      AstRefPtr refRight = expression();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpGtr<T>::create(refLeft, refRight);
    }
    else if (token == L"<=")
    {
      AstRefPtr refRight = expression();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpLeq<T>::create(refLeft, refRight);
    }
    else if (token == L">=")
    {
      AstRefPtr refRight = expression();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpGeq<T>::create(refLeft, refRight);
    }
    else if (token == L"==")
    {
      AstRefPtr refRight = expression();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpEqu<T>::create(refLeft, refRight);
    }
    else if (token == L"!=")
    {
      AstRefPtr refRight = expression();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpNeq<T>::create(refLeft, refRight);
    }
    else
    {
      //  ͕s\̃g[Nǂݍ񂾂Ƃ
      prev();
      break;
    }
  }
  return refLeft;
}

template <typename T>
typename FormulaParser<T>::AstRefPtr FormulaParser<T>::expression()
{
  //  @`
  //    expression ::= term (("+" | "-") term)*
  //  ̎.
  using namespace FormulaAst;
    
  AstRefPtr refLeft = term();
  if (!refLeft) return AstRefPtr(); //  G[
    
  for ( ; ; )
  {
    std::wstring token = next();
    if (token.empty())
    {
      //  Ō܂Ńg[Nǂݍ񂾂Ƃ
      break;
    }
    else if (token == L"+")
    {
      AstRefPtr refRight = term();
      if (!refRight) return AstRefPtr();  //  EӓŃG[
      refLeft = OpAdd<T>::create(refLeft, refRight);
    }
    else if (token == L"-")
    {
      AstRefPtr refRight = term();
      if (!refRight) return AstRefPtr();  //  EӓŃG[
      refLeft = OpSub<T>::create(refLeft, refRight);
    }
    else
    {
      //  ͕s\̃g[Nǂݍ񂾂Ƃ
      prev();
      break;
    }
  }
  return refLeft;
}

template <typename T>
typename FormulaParser<T>::AstRefPtr FormulaParser<T>::term()
{
  //  @`
  //    term ::= sign (("*" | "/" | "%") sign)*
  //  ̎.
  using namespace FormulaAst;

  AstRefPtr refLeft = sign();
  if (!refLeft) return AstRefPtr(); //  ӓŃG[
  for ( ; ; )
  {
    std::wstring token = next();
    if (token.empty())
    {
      //  Ō܂Ńg[Nǂݍ񂾂Ƃ
      break;
    }
    else if (token == L"*")
    {
      AstRefPtr refRight = sign();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpMul<T>::create(refLeft, refRight);
    }
    else if (token == L"/")
    {
      AstRefPtr refRight = sign();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpDiv<T>::create(refLeft, refRight);
    }
    else if (token == L"%")
    {
      AstRefPtr refRight = sign();
      if (!refRight) return AstRefPtr();  //  G[
      refLeft = OpMod<T>::create(refLeft, refRight);
    }
    else
    {
      //  ͕s\̃g[Nǂݍ񂾂Ƃ
      prev();
      break;
    }
  }
  return refLeft;
}


template <typename T>
typename FormulaParser<T>::AstRefPtr FormulaParser<T>::sign()
{
  //  @`
  //    sign ::= ("+" | "-" | "!"*) factor
  //  ̎.
  using namespace FormulaAst;

  std::wstring leftOp = next();
  int leftOpCount = 0;
  if (leftOp.empty())
  {
    //  Ō܂Ńg[Nǂݍ񂾂Ƃ
    return AstRefPtr(); //  G[
  }
  else if (leftOp == L"+" || leftOp == L"-")
  {
    leftOpCount = 1;
  }
  else if (leftOp == L"!")
  {
    leftOpCount = 1;
    for ( ; ; )
    {
      std::wstring token = next();
      if (token.empty())
      {
        //  Ō܂Ńg[Nǂݍ񂾂Ƃ
        return AstRefPtr(); //  G[
      }
      else if (token == L"!")
      {
        ++leftOpCount;
      }
      else
      {
        prev();
        break;
      }
    }
  }
  else
  {
    prev();
  }
  
  AstRefPtr refChild = factor();
  if (!refChild) return AstRefPtr();  //  G[

  for (int i = 0; i < leftOpCount; ++i)
  {
    if      (leftOp == L"+")    refChild = OpPlus<T>::create(refChild);
    else if (leftOp == L"-")    refChild = OpMinus<T>::create(refChild);
    else /* (leftOp == L"!") */ refChild = OpNot<T>::create(refChild);
  }
  
  return refChild;
}

template <typename T>
typename FormulaParser<T>::AstRefPtr FormulaParser<T>::factor()
{
  //  @`
  //    factor      ::= NUMBER |
  //                    "(" formula ")" |
  //                    FUNCTION0_ID "()" |
  //                    FUNCTION1_ID "(" formula ")" |
  //                    FUNCTION2_ID "(" formula "," formula ")" |
  //                    FUNCTION3_ID "(" formula "," formula "," formula ")" |
  //                    FUNCTION4_ID "(" formula "," formula "," formula "," formula ")" |
  //                    PARAMETER_ID |
  //                    LITERAL_ID
  //  ̎.
  using namespace FormulaAst;

  std::wstring token = next();
  if (token.empty())  return AstRefPtr(); //  G[
  
  T value;
  if (FormulaUtil::convTokenToValue<T>(token, value))
  {
    //  NUMBER ɈvƂ
    return Literal<T>::create(value);
  }
  else if (token == L"(")
  {
    //  "(" formula ")" Ɉv\Ƃ
    AstRefPtr refChild = formula();
    if (!refChild) return AstRefPtr();  //  G[
    token = next();
    if (token != L")") return AstRefPtr();  //  ʂĂȂ, G[
    return refChild;
  }
  else
  {
    //  ֐Ăяo, ϐ, 萔Ɉv\Ƃ
    std::wstring name = token;
    token = next();
    if (token == L"(")
    {
      //  ֐ĂяoɈv\Ƃ
      token = next();
      if (token.empty())
      {
        return AstRefPtr(); //  G[
      }
      else if (token == L")")
      {
        //  0 ֐ĂяoɈv\Ƃ
        Func0Map::iterator iter = funcs0_.find(name);
        if (iter != funcs0_.end())  return iter->second->createFunction();
        else                        return AstRefPtr(); //  G[
      }
      
      prev();
      AstRefPtr refArg0 = formula();
      if (!refArg0) return AstRefPtr(); //  G[
      
      token = next();
      if (token.empty())
      {
        return AstRefPtr(); //  G[
      }
      else if (token == L")")
      {
        //  1 ֐ĂяoɈv\Ƃ
        Func1Map::iterator iter = funcs1_.find(name);
        if (iter != funcs1_.end())  return iter->second->createFunction(refArg0);
        else                        return AstRefPtr(); //  G[
      }
      
      if (token != L",") return AstRefPtr();  //  G[
      AstRefPtr refArg1 = formula();
      if (!refArg1) return AstRefPtr(); //  G[
      
      token = next();
      if (token.empty())
      {
        return AstRefPtr(); //  G[
      }
      else if (token == L")")
      {
        //  2 ֐ĂяoɈv\Ƃ
        Func2Map::iterator iter = funcs2_.find(name);
        if (iter != funcs2_.end())  return iter->second->createFunction(refArg0, refArg1);
        else                        return AstRefPtr(); //  G[
      }
      
      if (token != L",") return AstRefPtr();  //  G[
      AstRefPtr refArg2 = formula();
      if (!refArg2) return AstRefPtr(); //  G[
      
      token = next();
      if (token.empty())
      {
        return AstRefPtr(); //  G[
      }
      else if (token == L")")
      {
        //  3 ֐ĂяoɈv\Ƃ
        Func3Map::iterator iter = funcs3_.find(name);
        if (iter != funcs3_.end())  return iter->second->createFunction(refArg0, refArg1, refArg2);
        else                        return AstRefPtr(); //  G[
      }
      
      if (token != L",") return AstRefPtr();  //  G[
      AstRefPtr refArg3 = formula();
      if (!refArg3) return AstRefPtr(); //  G[
      
      token = next();
      if (token == L")")
      {
        //  4 ֐ĂяoɈv\Ƃ
        Func4Map::iterator iter = funcs4_.find(name);
        if (iter != funcs4_.end())  return iter->second->createFunction(refArg0, refArg1, refArg2, refArg3);
        else                        return AstRefPtr(); //  G[
      }
      else
      {
        return AstRefPtr(); //  G[
      }
    }
    else
    {
      //  ϐ, 萔Ɉv\Ƃ
      if (!token.empty()) prev();
      
      ParameterMap::iterator param_iter = parameters_.find(name);
      if (param_iter != parameters_.end())
      {
        //  ϐƂ
        return Parameter<T>::create(name, param_iter->second);
      }
      
      LiteralMap::iterator lit_iter = literals_.find(name);
      if (lit_iter != literals_.end())
      {
        //  萔Ƃ
        return NamedLiteral<T>::create(name, lit_iter->second);
      }
      
      //  ϐ, 萔ȂƂ
      return AstRefPtr(); //  G[
    }
  }
}

//--------------------------------------------------------------------------------
