#pragma once

//--------------------------------------------------------------------------------
#include <string>
#include <complex>
#include <exception>
//#include <boost/utility.hpp>
//#include <boost/shared_ptr.hpp>
#include <memory>
#include "FormulaUtil.h"

namespace FormulaAst {
//--------------------------------------------------------------------------------
//  f̉Z (ZłȂƂ͗O𑗏o.)
template <typename T>
bool operator < (std::complex<T> const &lhs, std::complex<T> const &rhs)
{
  if (lhs.imag() != 0 || rhs.imag() != 0) throw std::runtime_error("Operator \'<\' is unavailable for complex number.");
  return lhs.real() < rhs.real();
}

template <typename T>
bool operator <= (std::complex<T> const &lhs, std::complex<T> const &rhs)
{
  if (lhs.imag() != 0 || rhs.imag() != 0) throw std::runtime_error("Operator \'<=\' is unavailable for complex number.");
  return lhs.real() <= rhs.real();
}

template <typename T>
bool operator > (std::complex<T> const &lhs, std::complex<T> const &rhs)
{
  if (lhs.imag() != 0 || rhs.imag() != 0) throw std::runtime_error("Operator \'>\' is unavailable for complex number.");
  return lhs.real() > rhs.real();
}

template <typename T>
bool operator >= (std::complex<T> const &lhs, std::complex<T> const &rhs)
{
  if (lhs.imag() != 0 || rhs.imag() != 0) throw std::runtime_error("Operator \'>=\' is unavailable for complex number.");
  return lhs.real() >= rhs.real();
}

//--------------------------------------------------------------------------------
/**
 *   AST \x[XNX.
 */
template <typename T>
class Base /*: private boost::noncopyable*/
{
public:
  /**
   *  AST ̃X}[g|C^.
   */
  //typedef boost::shared_ptr<Base<T> > RefPtr;
  typedef std::tr1::shared_ptr<Base<T> > RefPtr;
  
public:
  virtual ~Base() { }

  /**
   *  vZ.
   *  @return vZ.
   */
  virtual T execute() = 0;
  
  /**
   *  ̃c[m[heƂĈƂł邩ǂׂ.
   *  ̃m[h Literal ܂ Literal tƂĎ
   *  m[hł邱Ƃ𒲂ׂ.
   *  @return eƂĈ\łƂ true, łȂƂ false.
   */
  virtual bool isLiteral() const = 0;

  /**
   *  Zq̗Dx擾.
   *  @return Dx.
   */
  virtual int getPriority() const = 0;
  
  /**
   *  𕶎ɕϊ.
   *  @return ϊꂽ.
   */
  virtual std::wstring getString() const = 0;

  /**
   *  𕶎ɕϊ. m[h̗D揇ʂ parentPriority 
   *  Ⴂꍇ, IɊʂ.
   */
  virtual std::wstring getStringWithBrackets(int parentPriority)
  {
    if (getPriority() > parentPriority) return L'(' + getString() + L')';
    else                                return getString();
  }  

  /**
   *  svȃem̉Zȗ AST 𐶐.
   *  @return ꂽ AST.
   */
  virtual RefPtr cutDown() = 0;
  
};

/**
 *  em[h.
 */
template <typename T>
class Literal : public Base<T>
{
public:
  static RefPtr create(T const &val)      { return RefPtr(new Literal<T>(val)); }
  virtual T execute()                     { return val_; }
  virtual bool isLiteral() const          { return true; }
  virtual int getPriority() const         { return 0; }
  virtual std::wstring getString() const  { return FormulaUtil::convValueToToken(val_); }
  virtual std::wstring getStringWithBrackets(int parentPriority) { return getString(); }
  virtual RefPtr cutDown()                { return create(val_); }

protected:
  explicit Literal(T const &val) : val_(val) { }

private:
  T val_;
};

//--------------------------------------------------------------------------------
/**
 *  P+Zqm[h.
 */
template <typename T>
class OpPlus : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0)    { return RefPtr(new OpPlus<T>(refArg0)); }
  virtual T execute()                     { return refArg0_->execute(); }
  virtual bool isLiteral() const          { return refArg0_->isLiteral(); }
  virtual int getPriority() const         { return 2; }
  virtual std::wstring getString() const  { return L'+' + refArg0_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown()); }

private:
  OpPlus(RefPtr refArg0) : refArg0_(refArg0) { }

private:
  RefPtr refArg0_;
};

/**
 *  P-Zqm[h.
 */
template <typename T>
class OpMinus : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0)    { return RefPtr(new OpMinus<T>(refArg0)); }
  virtual T execute()                     { return -refArg0_->execute(); }
  virtual bool isLiteral() const          { return refArg0_->isLiteral(); }
  virtual int getPriority() const         { return 2; }
  virtual std::wstring getString() const  { return L'-' + refArg0_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown()); }

private:
  OpMinus(RefPtr refArg0) : refArg0_(refArg0) { }

private:
  RefPtr refArg0_;
};

/**
 *  _ے艉Zqm[h.
 */
template <typename T>
class OpNot : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0)    { return RefPtr(new OpNot<T>(refArg0)); }
  virtual T execute()                     { return refArg0_->execute() != T(0) ? 0 : 1; }
  virtual bool isLiteral() const          { return refArg0_->isLiteral(); }
  virtual int getPriority() const         { return 2; }
  virtual std::wstring getString() const  { return L'!' + refArg0_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown()); }

private:
  OpNot(RefPtr refArg0) : refArg0_(refArg0) { }

private:
  RefPtr refArg0_;
};

/**
 *  Q*Zqm[h.
 */
template <typename T>
class OpMul : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpMul<T>(refArg0, refArg1)); }
  virtual T execute()                     { return refArg0_->execute() * refArg1_->execute(); }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 3; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L'*' +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpMul(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  /Zqm[h.
 */
template <typename T>
class OpDiv : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpDiv<T>(refArg0, refArg1)); }
  virtual T execute()
  {
    T arg0 = refArg0_->execute(), arg1 = refArg1_->execute();
    if (0 == arg1) throw std::runtime_error("Zero division!.");
    return arg0 / arg1;
  }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 3; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L'/' +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpDiv(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  %Zqm[h.
 */
template <typename T>
class OpMod : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpMod<T>(refArg0, refArg1)); }
  virtual T execute()                     { return FormulaUtil::mod(refArg0_->execute(), refArg1_->execute()); }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 3; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L'%' +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpMod(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  Q+Zqm[h.
 */
template <typename T>
class OpAdd : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpAdd<T>(refArg0, refArg1)); }
  virtual T execute()                     { return refArg0_->execute() + refArg1_->execute(); }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 4; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L'+' +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpAdd(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  Q-Zqm[h.
 */
template <typename T>
class OpSub : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpSub<T>(refArg0, refArg1)); }
  virtual T execute()                     { return refArg0_->execute() - refArg1_->execute(); }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 4; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L'-' +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpSub(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  <Zqm[h.
 */
template <typename T>
class OpLss : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpLss<T>(refArg0, refArg1)); }
  virtual T execute()                     { return refArg0_->execute() < refArg1_->execute() ? 1 : 0; }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 5; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L'<' +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpLss(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  <=Zqm[h.
 */
template <typename T>
class OpLeq : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpLeq<T>(refArg0, refArg1)); }
  virtual T execute()                     { return refArg0_->execute() <= refArg1_->execute() ? 1 : 0; }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 5; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L"<=" +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpLeq(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  >Zqm[h.
 */
template <typename T>
class OpGtr : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpGtr<T>(refArg0, refArg1)); }
  virtual T execute()                     { return refArg0_->execute() > refArg1_->execute() ? 1 : 0; }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 5; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L'>' +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpGtr(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  >=Zqm[h.
 */
template <typename T>
class OpGeq : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpGeq<T>(refArg0, refArg1)); }
  virtual T execute()                     { return refArg0_->execute() >= refArg1_->execute() ? 1 : 0; }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 5; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L">=" +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpGeq(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  ==Zqm[h.
 */
template <typename T>
class OpEqu : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpEqu<T>(refArg0, refArg1)); }
  virtual T execute()                     { return refArg0_->execute() == refArg1_->execute() ? 1 : 0; }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 5; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L"==" +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpEqu(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  !=Zqm[h.
 */
template <typename T>
class OpNeq : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpNeq<T>(refArg0, refArg1)); }
  virtual T execute()                     { return refArg0_->execute() != refArg1_->execute() ? 1 : 0; }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 5; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L"!=" +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpNeq(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  _ANDZqm[h.
 */
template <typename T>
class OpAnd : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpAnd<T>(refArg0, refArg1)); }
  virtual T execute()                     { return (refArg0_->execute() != T(0)) && (refArg1_->execute() != T(0)) ? 1 : 0; }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 6; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L"&&" +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpAnd(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/** 
 *  _ORZqm[h.
 */
template <typename T>
class OpOr : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1)  { return RefPtr(new OpOr<T>(refArg0, refArg1)); }
  virtual T execute()                     { return (refArg0_->execute() != T(0)) || (refArg1_->execute() != T(0)) ? 1 : 0; }
  virtual bool isLiteral() const          { return refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 6; }
  virtual std::wstring getString() const  { return refArg0_->getStringWithBrackets(getPriority()) + L"||" +
                                                   refArg1_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   create(refArg0_->cutDown(), refArg1_->cutDown()); }

private:
  OpOr(RefPtr refArg0, RefPtr refArg1) : refArg0_(refArg0), refArg1_(refArg1) { }

private:
  RefPtr refArg0_, refArg1_;
};

/**
 *  Zqm[h.
 */
template <typename T>
class OpConditional : public Base<T>
{
public:
  static RefPtr create(RefPtr refArg0, RefPtr refArg1, RefPtr refArg2)
    { return RefPtr(new OpConditional<T>(refArg0, refArg1, refArg2)); }
  virtual T execute()
    { return (refArg0_->execute() != T(0)) ? refArg1_->execute() : refArg2_->execute(); }
  virtual bool isLiteral() const
    { return refArg0_->isLiteral() && refArg1_->isLiteral() && refArg2_->isLiteral(); }
  virtual int getPriority() const
    { return 7; }
  virtual std::wstring getString() const
    { return refArg0_->getStringWithBrackets(getPriority()) + L'?' +
             refArg1_->getStringWithBrackets(getPriority()) + L':' + refArg2_->getStringWithBrackets(getPriority()); }
  virtual RefPtr cutDown()
    { return isLiteral() ? Literal<T>::create(execute()) :
             create(refArg0_->cutDown(), refArg1_->cutDown(), refArg2_->cutDown()); }

private:
  OpConditional(RefPtr refArg0, RefPtr refArg1, RefPtr refArg2)
    : refArg0_(refArg0), refArg1_(refArg1), refArg2_(refArg2) { }

private:
  RefPtr refArg0_, refArg1_, refArg2_;
};

//--------------------------------------------------------------------------------
/**
 *  0 ֐ AST m[h𐶐IuWFNg̃x[XNX.
 */
template <typename T>
class Func0CreatorBase /*: private boost::noncopyable*/
{
public:
  //typedef boost::shared_ptr<Func0CreatorBase<T> > RefPtr;
  typedef std::tr1::shared_ptr<Func0CreatorBase<T> > RefPtr;
  typedef typename Base<T>::RefPtr AstRefPtr;

public:
  virtual ~Func0CreatorBase() { }
  virtual AstRefPtr createFunction() = 0;
};

/**
 *  1 ֐ AST m[h𐶐IuWFNg̃x[XNX.
 */
template <typename T>
class Func1CreatorBase /*: private boost::noncopyable*/
{
public:
  //typedef boost::shared_ptr<Func1CreatorBase<T> > RefPtr;
  typedef std::tr1::shared_ptr<Func1CreatorBase<T> > RefPtr;
  typedef typename Base<T>::RefPtr AstRefPtr;

public:
  virtual ~Func1CreatorBase() { }
  virtual AstRefPtr createFunction(AstRefPtr refArg0) = 0;
};

/**
 *  2 ֐ AST m[h𐶐IuWFNg̃x[XNX.
 */
template <typename T>
class Func2CreatorBase /*: private boost::noncopyable*/
{
public:
  //typedef boost::shared_ptr<Func2CreatorBase<T> > RefPtr;
  typedef std::tr1::shared_ptr<Func2CreatorBase<T> > RefPtr;
  typedef typename Base<T>::RefPtr AstRefPtr;

public:
  virtual ~Func2CreatorBase() { }
  virtual AstRefPtr createFunction(AstRefPtr refArg0, AstRefPtr refArg1) = 0;
};

/**
 *  3 ֐ AST m[h𐶐IuWFNg̃x[XNX.
 */
template <typename T>
class Func3CreatorBase /*: private boost::noncopyable*/
{
public:
  //typedef boost::shared_ptr<Func3CreatorBase<T> > RefPtr;
  typedef std::tr1::shared_ptr<Func3CreatorBase<T> > RefPtr;
  typedef typename Base<T>::RefPtr AstRefPtr;

public:
  virtual ~Func3CreatorBase() { }
  virtual AstRefPtr createFunction(AstRefPtr refArg0, AstRefPtr refArg1, AstRefPtr refArg2) = 0;
};

/**
 *  4 ֐ AST m[h𐶐IuWFNg̃x[XNX.
 */
template <typename T>
class Func4CreatorBase /*: private boost::noncopyable*/
{
public:
  //typedef boost::shared_ptr<Func4CreatorBase<T> > RefPtr;
  typedef std::tr1::shared_ptr<Func4CreatorBase<T> > RefPtr;
  typedef typename Base<T>::RefPtr AstRefPtr;

public:
  virtual ~Func4CreatorBase() { }
  virtual AstRefPtr createFunction(AstRefPtr refArg0, AstRefPtr refArg1,
                                   AstRefPtr refArg2, AstRefPtr refArg3) = 0;
};

//--------------------------------------------------------------------------------
/**
 *  0 ֐IuWFNg AST m[h̃NX.
 */
template <typename T, typename Functor>
class Func0 : public Base<T>
{
public:
  Func0(std::wstring const &name, Functor func, bool refTransparency)
    : name_(name), func_(func), refTransparency_(refTransparency) { }
  
  virtual T execute()                     { return func_(); }
  virtual bool isLiteral() const          { return refTransparency_; }
  virtual int getPriority() const         { return 1; }
  virtual std::wstring getString() const  { return name_ + L"()"; }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   RefPtr(new Func0<T, Functor>(name_, func_, refTransparency_)); }

private:
  std::wstring name_;
  Functor func_;
  bool refTransparency_;  //  QƓ (Referential transparency) 邩ǂ.
};

/**
 *  1 ֐IuWFNg AST m[h̃NX.
 */
template <typename T, typename Functor>
class Func1 : public Base<T>
{
public:
  Func1(std::wstring const &name, Functor func, RefPtr refArg0, bool refTransparency)
    : refArg0_(refArg0), name_(name), func_(func), refTransparency_(refTransparency) { }
  
  virtual T execute()                     { return func_(refArg0_->execute()); }
  virtual bool isLiteral() const          { return refTransparency_ && refArg0_->isLiteral(); }
  virtual int getPriority() const         { return 1; }
  virtual std::wstring getString() const  { return name_ + L'(' + refArg0_->getString() + L')'; }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   RefPtr(new Func1<T, Functor>(name_, func_, refArg0_->cutDown(), refTransparency_)); }

private:
  RefPtr refArg0_;
  std::wstring name_;
  Functor func_;
  bool refTransparency_;  //  QƓ (Referential transparency) 邩ǂ.
};

/**
 *  2 ֐IuWFNg AST m[h̃NX.
 */
template <typename T, typename Functor>
class Func2 : public Base<T>
{
public:
  Func2(std::wstring const &name, Functor func, RefPtr refArg0, RefPtr refArg1, bool refTransparency)
    : refArg0_(refArg0), refArg1_(refArg1), name_(name), func_(func), refTransparency_(refTransparency) { }
  
  virtual T execute()                     { return func_(refArg0_->execute(), refArg1_->execute()); }
  virtual bool isLiteral() const          { return refTransparency_ && refArg0_->isLiteral() && refArg1_->isLiteral(); }
  virtual int getPriority() const         { return 1; }
  virtual std::wstring getString() const  { return name_ + L'(' + refArg0_->getString() + L',' + refArg1_->getString() + L')'; }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   RefPtr(new Func2<T, Functor>(name_, func_, refArg0_->cutDown(), refArg1_->cutDown(),
                                                   refTransparency_)); }

private:
  RefPtr refArg0_, refArg1_;
  std::wstring name_;
  Functor func_;
  bool refTransparency_;  //  QƓ (Referential transparency) 邩ǂ.
};

/**
 *  3 ֐IuWFNg AST m[h̃NX.
 */
template <typename T, typename Functor>
class Func3 : public Base<T>
{
public:
  Func3(std::wstring const &name, Functor func, RefPtr refArg0, RefPtr refArg1, RefPtr refArg2, bool refTransparency)
    : refArg0_(refArg0), refArg1_(refArg1), refArg2_(refArg2), name_(name), func_(func), refTransparency_(refTransparency) { }
  
  virtual T execute()                     { return func_(refArg0_->execute(), refArg1_->execute(), refArg2_->execute()); }
  virtual bool isLiteral() const          { return refTransparency_ && refArg0_->isLiteral() && refArg1_->isLiteral() &&
                                                   refArg2_->isLiteral(); }
  virtual int getPriority() const         { return 1; }
  virtual std::wstring getString() const  { return name_ + L'(' + refArg0_->getString() + L',' + refArg1_->getString() + L',' +
                                                   refArg2_->getString() + L')'; }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   RefPtr(new Func3<T, Functor>(name_, func_, refArg0_->cutDown(), refArg1_->cutDown(),
                                                   refArg2_->cutDown(), refTransparency_)); }

private:
  RefPtr refArg0_, refArg1_, refArg2_;
  std::wstring name_;
  Functor func_;
  bool refTransparency_;  //  QƓ (Referential transparency) 邩ǂ.
};

/**
 *  4 ֐IuWFNg AST m[h̃NX.
 */
template <typename T, typename Functor>
class Func4 : public Base<T>
{
public:
  Func4(std::wstring const &name, Functor func, RefPtr refArg0, RefPtr refArg1,
                  RefPtr refArg2, RefPtr refArg3, bool refTransparency)
    : refArg0_(refArg0), refArg1_(refArg1), refArg2_(refArg2), refArg3_(refArg3),
      name_(name), func_(func), refTransparency_(refTransparency) { }
  
  virtual T execute()                     { return func_(refArg0_->execute(), refArg1_->execute(),
                                                         refArg2_->execute(), refArg3_->execute()); }
  virtual bool isLiteral() const          { return refTransparency_ && refArg0_->isLiteral() && refArg1_->isLiteral() &&
                                                   refArg2_->isLiteral() && refArg3_->isLiteral(); }
  virtual int getPriority() const         { return 1; }
  virtual std::wstring getString() const  { return name_ + L'(' + refArg0_->getString() + L',' + refArg1_->getString() + L',' +
                                                   refArg2_->getString() + L',' + refArg3_->getString() + L')'; }
  virtual RefPtr cutDown()                { return isLiteral() ? Literal<T>::create(execute()) :
                                                   RefPtr(new Func4<T, Functor>(name_, func_, refArg0_->cutDown(), refArg1_->cutDown(),
                                                   refArg2_->cutDown(), refArg3_->cutDown(), refTransparency_)); }

private:
  RefPtr refArg0_, refArg1_, refArg2_, refArg3_;
  std::wstring name_;
  Functor func_;
  bool refTransparency_;  //  QƓ (Referential transparency) 邩ǂ.
};

/**
 *  0 ֐IuWFNg AST m[h𐶐NX.
 */
template <typename T, typename Functor>
class Func0Creator : public Func0CreatorBase<T>
{
public:
  static RefPtr create(std::wstring const &name, Functor func, bool refTransparency)
    { return RefPtr(new Func0Creator<T, Functor>(name, func, refTransparency)); }
  virtual AstRefPtr createFunction()
    { return AstRefPtr(new Func0<T, Functor>(name_, func_, refTransparency_)); }

private:
  Func0Creator(std::wstring const &name, Functor func, bool refTransparency)
    : name_(name), func_(func), refTransparency_(refTransparency) { }
  
private:
  std::wstring name_;
  Functor func_;
  bool refTransparency_;  //  QƓ (Referential transparency) 邩ǂ.
};

/**
 *  1 ֐IuWFNg AST m[h𐶐NX.
 */
template <typename T, typename Functor>
class Func1Creator : public Func1CreatorBase<T>
{
public:
  static RefPtr create(std::wstring const &name, Functor func, bool refTransparency)
    { return RefPtr(new Func1Creator<T, Functor>(name, func, refTransparency)); }
  virtual AstRefPtr createFunction(AstRefPtr refArg0)
    { return AstRefPtr(new Func1<T, Functor>(name_, func_, refArg0, refTransparency_)); }

private:
  Func1Creator(std::wstring const &name, Functor func, bool refTransparency)
    : name_(name), func_(func), refTransparency_(refTransparency) { }
  
private:
  std::wstring name_;
  Functor func_;
  bool refTransparency_;
};

/**
 *  2 ֐IuWFNg AST m[h𐶐NX.
 */
template <typename T, typename Functor>
class Func2Creator : public Func2CreatorBase<T>
{
public:
  static RefPtr create(std::wstring const &name, Functor func, bool refTransparency)
    { return RefPtr(new Func2Creator<T, Functor>(name, func, refTransparency)); }
  virtual AstRefPtr createFunction(AstRefPtr refArg0, AstRefPtr refArg1)
    { return AstRefPtr(new Func2<T, Functor>(name_, func_, refArg0, refArg1, refTransparency_)); }

private:
  Func2Creator(std::wstring const &name, Functor func, bool refTransparency)
    : name_(name), func_(func), refTransparency_(refTransparency) { }
  
private:
  std::wstring name_;
  Functor func_;
  bool refTransparency_;
};

/**
 *  3 ֐IuWFNg AST m[h𐶐NX.
 */
template <typename T, typename Functor>
class Func3Creator : public Func3CreatorBase<T>
{
public:
  static RefPtr create(std::wstring const &name, Functor func, bool refTransparency)
    { return RefPtr(new Func3Creator<T, Functor>(name, func, refTransparency)); }
  virtual AstRefPtr createFunction(AstRefPtr refArg0, AstRefPtr refArg1, AstRefPtr refArg2)
    { return AstRefPtr(new Func3<T, Functor>(name_, func_, refArg0, refArg1, refArg2, refTransparency_)); }

private:
  Func3Creator(std::wstring const &name, Functor func, bool refTransparency)
    : name_(name), func_(func), refTransparency_(refTransparency) { }
  
private:
  std::wstring name_;
  Functor func_;
  bool refTransparency_;
};

/**
 *  4 ֐IuWFNg AST m[h𐶐NX.
 */
template <typename T, typename Functor>
class Func4Creator : public Func4CreatorBase<T>
{
public:
  static RefPtr create(std::wstring const &name, Functor func, bool refTransparency)
    { return RefPtr(new Func4Creator<T, Functor>(name, func, refTransparency)); }
  virtual AstRefPtr createFunction(AstRefPtr refArg0, AstRefPtr refArg1, AstRefPtr refArg2, AstRefPtr refArg3)
    { return AstRefPtr(new Func4<T, Functor>(name_, func_, refArg0, refArg1, refArg2, refArg3, refTransparency_)); }

private:
  Func4Creator(std::wstring const &name, Functor func, bool refTransparency)
    : name_(name), func_(func), refTransparency_(refTransparency) { }
  
private:
  std::wstring name_;
  Functor func_;
  bool refTransparency_;
};

//--------------------------------------------------------------------------------
/**
 *  PI ̖Ot̃e\m[h.
 */
template <typename T>
class NamedLiteral : public Literal<T>
{
public:
  static RefPtr create(std::wstring const &name, T const &val)  { return RefPtr(new NamedLiteral<T>(name, val)); }
  virtual std::wstring getString() const                        { return name_; }

private:
  NamedLiteral(std::wstring const &name, T const &val) : Literal<T>(val), name_(name) { }

private:
  std::wstring name_;
};

/**
 *  p[^\m[h.
 */
template <typename T>
class Parameter  : public Base<T>
{
public:
  static RefPtr create(std::wstring const &name, T const *pParameter)
    { return RefPtr(new Parameter<T>(name, pParameter)); }
  
  virtual T execute()                     { return *pParameter_; }
  virtual bool isLiteral() const          { return false; }
  virtual int getPriority() const         { return 0; }
  virtual std::wstring getString() const  { return name_; }
  virtual std::wstring getStringWithBrackets(int parentPriority) { return getString(); }
  virtual RefPtr cutDown()                { return create(name_, pParameter_); }

protected:
  explicit Parameter(std::wstring const &name, T const *pParameter)
    : name_(name), pParameter_(pParameter) { }

private:
  std::wstring name_;
  T const *pParameter_;
};

//--------------------------------------------------------------------------------
} //  namespace FormulaAst
