// M_PI ̒` FormulaStd.h ̒ŃCN[h math.h 
// wx/wx.h  2 ӏ. wx/wx.h ͊̕ɒ`Ƃ,
// Ē`Ȃ悤ɂȂĂ. FormulaStd.h ŏɃCN[h
// ƂŃG[.
#include "FormulaStd.h"
#include "FormulaCalculator.h"

#include <functional>

//--------------------------------------------------------------------------------
// V[hԂ.
static unsigned int GenerateRandomSeed()
{
  std::random_device rd;
  return rd();
}

//--------------------------------------------------------------------------------
FormulaCalculator::FormulaCalculator(GraphProject &project)
  : m_project(project),
    m_formula("0"),
    m_dataCount(1),
    m_x(0.0), m_i(0.0),
    m_randGen(GenerateRandomSeed()),
    m_randDist(0.0, 1.0)
{
}

void FormulaCalculator::SetFormula(wxString const &formula)
{
  m_formula = formula;
}

void FormulaCalculator::SetDataX(GraphDataX const &dataX)
{
  m_dataX = dataX;
}

void FormulaCalculator::SetDataCount(int count)
{
  m_dataCount = count;
}

void FormulaCalculator::SetLabel(wxString const &label)
{
  m_label = label;
}

wxString FormulaCalculator::GetFormula() const
{
  return m_formula;
}

GraphDataX FormulaCalculator::GetDataX() const
{
  return m_dataX;
}

int FormulaCalculator::GetDataCount() const
{
  return m_dataCount;
}

wxString FormulaCalculator::GetLabel() const
{
  return m_label;
}

bool FormulaCalculator::Parse()
{
  //  p[T[̐.
  FormulaParser<double> parser;
    
  //  p[T[ɕW֐E萔gݍ.
  installStandardFormulaFunc(parser);

  //  ̊֐𗘗pł悤ɂ.
  //  - y(n, x) n Ԗڂ̃Otf[^ X WŃANZX, QƓ
  //  - Y(n, i) n Ԗڂ̃Otf[^ɃCfbNXŃANZX, QƓ
  //  Ȃ, QƓ̏Ƃ, AST ؂̍œKs.
  using namespace std::placeholders;
  parser.setFunction2(L"y",  std::bind(&FormulaCalculator::GetSmallY, std::ref(*this),  _1, _2), true);
  parser.setFunction2(L"Y",  std::bind(&FormulaCalculator::GetLargeY, std::ref(*this),  _1, _2), true);

  //  KEXz̗֐𗘗pł悤ɂ. (QƓ͂Ȃ)
  parser.setFunction0(L"rand", std::bind(&FormulaCalculator::Rand, std::ref(*this)), false);

  //  p[T[ɕϐgݍ.
  parser.setParameter(L"x", &m_x);
  parser.setParameter(L"i", &m_i);
  
  //  p[Xs.
  m_refAst = parser.parse(m_formula.ToStdWstring());
  if (m_refAst)
  {
    // p[X.
    return true;
  }
  else
  {
    // p[Xs.
    int line = parser.getLine();
    int pos  = parser.getLinePosition();
    m_err.Printf(_("̃p[XɎs܂ (%d s, %d )!"), line, pos);
    return false;
  }
}

ResultCode FormulaCalculator::Calculate(int mipmapResolution, ProcContext *pPC)
{
  using namespace std::placeholders;

  // AST pČvZ. ʂ m_refData Ɋi[.
  ResultCode result = RunProcWithContext<ResultCode>
    (std::bind(&FormulaCalculator::CalculateInternal, this,  _1), pPC);
  if (result != ResultCode::SUCCESS)
  {
    m_refData.reset();
    return result;
  }

  // ~bv}bv𐶐.
  result = m_refData->CreateMipmapForResolution(mipmapResolution, pPC);
  if (result != ResultCode::SUCCESS)
  {
    m_refData.reset();
    m_err = _("Failed to allocate memory!");
    return result;
  }

  // vWFNgɒǉ.
  m_project.Append(m_refData);
  GraphDataRefPtr graph = m_project.GetItem(m_project.GetCount() - 1);
  graph->SetLabel(m_label);

  m_refData.reset();
  return result;
}

ResultCode FormulaCalculator::CalculateInternal(Progress *pProgress)
{
  GraphDataYRefPtr dataY;

  try
  {
    // Y f[^쐬
    dataY = GraphDataY::Create();
    dataY->raw.resize(m_dataCount);
  
    // ̌vZs.
    int updateInterval = m_dataCount / 500;
    if (0 == updateInterval) updateInterval = 1;
    if (pProgress != NULL)
    {
      wxString message;
      message.Printf(_("Calculating: %s ..."), m_formula);
      pProgress->UpdateMessage(message);
    }

    std::vector<double>::iterator iter = dataY->raw.begin();
    for (int i = 0; i < m_dataCount; ++i, ++iter)
    {
      m_i = static_cast<double>(i);
      m_x = m_dataX.begin + m_dataX.interval * m_i;
      *iter = m_refAst->execute();

      // vOXo[XV.
      if (0 == i % updateInterval && pProgress != NULL)
      {
        pProgress->UpdateProgress(i, m_dataCount);
      }

      // fmF
      if (0 == i % 10000 && pProgress != NULL && pProgress->IsAbortRequested())
      {
        return ResultCode::ABORT;
      }
    }

    if (pProgress != NULL)
    {
      pProgress->UpdateProgress(m_dataCount, m_dataCount);
    }

    m_refData = GraphData::Create(m_dataX, dataY);

    return ResultCode::SUCCESS;
  }
  catch (std::bad_alloc &)
  {
    // s
    m_err = _("Failed to allocate memory in the calculation process!");
    return ResultCode::MEMORY_ERROR;
  }
  catch (std::runtime_error &ex)
  {
    // AST ؏̗O std::runtime_error 𓊂ɂȂĂ. 
    m_err = ex.what();
    return ResultCode::ARITHMETIC_ERROR;
  }
  catch ( ... )
  {
    m_err = _("Unknown error!");
    return ResultCode::UNKNOWN_ERROR;
  }
}

/*ResultCode FormulaCalculator::CreateMipmapForResolution(int resolution, Progress *pProgress)
{
  wxASSERT(m_refData != NULL);
  return m_refData->CreateMipmapForResolution(resolution, pProgress);
}

void FormulaCalculator::AppendToProject()
{
  wxASSERT(m_refData != NULL);
  m_project.Append(m_refData);
  GraphDataRefPtr graph = m_project.GetItem(m_project.GetCount() - 1);
  graph->SetLabel(m_label);
}*/

wxString FormulaCalculator::GetLastErrorMessage() const
{
  return m_err;
}

double FormulaCalculator::GetSmallY(double n, double x)
{
  // n  1 ` (Ot)
  if (n >= 1 && n <= m_project.GetCount())
  {
    // GraphDataRefPtr gƎQƃJEg̑KvɂȂ, dȂ̂, |Cg擾.
    GraphData *graph = m_project.GetItem(n - 1).get();
    GraphDataX const &dataX = graph->GetDataX();
    double index = (x - dataX.begin) / dataX.interval;
    int leftIndex = static_cast<int>(std::floor(index));  // |Cg̃CfbNX
    int rightIndex = leftIndex + 1;                       // E|Cg̃CfbNX
    double t = index - leftIndex;                         // E|Cg̕

    if (leftIndex >= 0 && rightIndex < graph->GetCount())
    {
      GraphDataY *dataY = graph->GetDataY().get();
      double leftValue = dataY->raw[leftIndex];               // |Cg̒l
      double rightValue = dataY->raw[rightIndex];             // E|Cg̒l
      double value = leftValue * (1.0 - t) + rightValue * t;  // `
      return value;
    }
    else
    {
      return 0.0;
    }
  }
  else
  {
    return 0.0;
  }
}

double FormulaCalculator::GetLargeY(double n, double i)
{
  // n  1 ` (Ot)
  if (n >= 1 && n <= m_project.GetCount())
  {
    // GraphDataRefPtr gƎQƃJEg̑KvɂȂ, dȂ̂, |Cg擾.
    GraphData *graph = m_project.GetItem(n - 1).get();
    int index = static_cast<int>(i);
    if (index >= 0 && index < graph->GetCount())
    {
      GraphDataY *dataY = graph->GetDataY().get();
      return dataY->raw[index];
    }
    else
    {
      return 0.0;
    }
  }
  else
  {
    return 0.0;
  }
}

double FormulaCalculator::Rand()
{
  return m_randDist(m_randGen);
}

//--------------------------------------------------------------------------------
