#pragma once

#include <complex>
#include "Formula.h"

#define _USE_MATH_DEFINES
#include <math.h>

//--------------------------------------------------------------------------------

namespace FormulaUtil {
//--------------------------------------------------------------------------------
inline float round(float t)
{
  return t >= 0.0f ? std::floor(t + 0.5f) : std::ceil(t - 0.5f);
}

inline double round(double t)
{
  return t >= 0.0 ? std::floor(t + 0.5) : std::ceil(t - 0.5);
}

inline float wrap(float t)
{
  t = std::fmod(t, 2.0f * (float)M_PI);  //  t  (-2, 2) ƂȂ.
  
  if      (t < -(float)M_PI)  return t + 2.0f * (float)M_PI;
  else if (t >= (float)M_PI)  return t - 2.0f * (float)M_PI;
  else                        return t;
}

inline double wrap(double t)
{
  t = std::fmod(t, 2.0 * M_PI);  //  t  (-2, 2) ƂȂ.
  
  if      (t < -M_PI) return t + 2.0 * M_PI;
  else if (t >= M_PI) return t - 2.0 * M_PI;
  else                return t;
}

//--------------------------------------------------------------------------------
//  fZp֐

//  std::complex<float> p
inline std::complex<float> ceil(std::complex<float> const &x)
{
  return std::complex<float>(std::ceil(x.real()), std::ceil(x.imag()));
}

inline std::complex<float> floor(std::complex<float> const &x)
{
  return std::complex<float>(std::floor(x.real()), std::floor(x.imag()));
}

inline std::complex<float> round(std::complex<float> const &x)
{
  return std::complex<float>(round(x.real()), round(x.imag()));
}

inline std::complex<float> wrap(std::complex<float> const &x)
{
  return std::complex<float>(wrap(x.real()), wrap(x.imag()));
}

inline std::complex<float> asin(std::complex<float> const &x)
{
  if (x.imag() != 0) throw std::runtime_error("asin() is unavailable for complex number.");
  return std::complex<float>(std::asin(x.real()));
}

inline std::complex<float> acos(std::complex<float> const &x)
{
  if (x.imag() != 0) throw std::runtime_error("acos() is unavailable for complex number.");
  return std::complex<float>(std::acos(x.real()));
}

inline std::complex<float> atan(std::complex<float> const &x)
{
  if (x.imag() != 0) throw std::runtime_error("atan() is unavailable for complex number.");
  return std::complex<float>(std::atan(x.real()));
}

inline std::complex<float> atan2(std::complex<float> const &y, std::complex<float> const &x)
{
  if (x.imag() != 0 || y.imag() != 0) throw std::runtime_error("atan2() is unavailable for complex number.");
  return std::complex<float>(std::atan2(y.real(), x.real()));
}

//  std::complex<double> p
inline std::complex<double> ceil(std::complex<double> const &x)
{
  return std::complex<double>(std::ceil(x.real()), std::ceil(x.imag()));
}

inline std::complex<double> floor(std::complex<double> const &x)
{
  return std::complex<double>(std::floor(x.real()), std::floor(x.imag()));
}

inline std::complex<double> round(std::complex<double> const &x)
{
  return std::complex<double>(round(x.real()), round(x.imag()));
}

inline std::complex<double> wrap(std::complex<double> const &x)
{
  return std::complex<double>(wrap(x.real()), wrap(x.imag()));
}

inline std::complex<double> asin(std::complex<double> const &x)
{
  if (x.imag() != 0) throw std::runtime_error("asin() is unavailable for complex number.");
  return std::complex<double>(std::asin(x.real()));
}

inline std::complex<double> acos(std::complex<double> const &x)
{
  if (x.imag() != 0) throw std::runtime_error("acos() is unavailable for complex number.");
  return std::complex<double>(std::acos(x.real()));
}

inline std::complex<double> atan(std::complex<double> const &x)
{
  if (x.imag() != 0) throw std::runtime_error("atan() is unavailable for complex number.");
  return std::complex<double>(std::atan(x.real()));
}

inline std::complex<double> atan2(std::complex<double> const &y, std::complex<double> const &x)
{
  if (x.imag() != 0 || y.imag() != 0) throw std::runtime_error("atan2() is unavailable for complex number.");
  return std::complex<double>(std::atan2(y.real(), x.real()));
}

//--------------------------------------------------------------------------------
} //  namespace FormulaUtil

//--------------------------------------------------------------------------------
/**
 *  p[T[ɕW֐, 萔gݍ.
 */
template <typename T>
void installStandardFormulaFunc(FormulaParser<T> &parser)
{
}

template <>
void installStandardFormulaFunc(FormulaParser<float> &parser)
{
  //  gݍ݊֐o^.
  //  (I[o[[hꂽ֐̊֐|C^o, ֐̌^w肷.)
  typedef float (*Func1Ptr)(float);
  typedef float (*Func2Ptr)(float, float);
  
  //  ֐
  parser.setFunction1(L"abs",   (Func1Ptr)&std::abs,           true);
  parser.setFunction1(L"ceil",  (Func1Ptr)&std::ceil,          true);
  parser.setFunction1(L"floor", (Func1Ptr)&std::floor,         true);
  parser.setFunction1(L"round", (Func1Ptr)&FormulaUtil::round, true);
  parser.setFunction1(L"wrap",  (Func1Ptr)&FormulaUtil::wrap,  true);
  
  //  Op֐
  parser.setFunction1(L"acos",  (Func1Ptr)&std::acos,          true);
  parser.setFunction1(L"asin",  (Func1Ptr)&std::asin,          true);
  parser.setFunction1(L"atan",  (Func1Ptr)&std::atan,          true);
  parser.setFunction2(L"atan2", (Func2Ptr)&std::atan2,         true);
  parser.setFunction1(L"cos",   (Func1Ptr)&std::cos,           true);
  parser.setFunction1(L"cosh",  (Func1Ptr)&std::cosh,          true);
  parser.setFunction1(L"sin",   (Func1Ptr)&std::sin,           true);
  parser.setFunction1(L"sinh",  (Func1Ptr)&std::sinh,          true);
  parser.setFunction1(L"tan",   (Func1Ptr)&std::tan,           true);
  parser.setFunction1(L"tanh",  (Func1Ptr)&std::tanh,          true);
  
  //  wΐ֐
  parser.setFunction1(L"exp",   (Func1Ptr)&std::exp,           true);
  parser.setFunction1(L"log",   (Func1Ptr)&std::log,           true);
  parser.setFunction1(L"log10", (Func1Ptr)&std::log10,         true);
  parser.setFunction2(L"pow",   (Func2Ptr)&std::pow,           true);
  parser.setFunction1(L"sqrt",  (Func1Ptr)&std::sqrt,          true);

  //  gݍݒ萔o^.
  parser.setLiteral(L"PI", (float)M_PI);
  parser.setLiteral(L"E",  (float)M_E);
}

template <>
void installStandardFormulaFunc(FormulaParser<double> &parser)
{
  //  gݍ݊֐o^.
  //  (I[o[[hꂽ֐̊֐|C^o, ֐̌^w肷.)
  typedef double (*Func1Ptr)(double);
  typedef double (*Func2Ptr)(double, double);
  
  //  ֐
  parser.setFunction1(L"abs",   (Func1Ptr)&std::abs,           true);
  parser.setFunction1(L"ceil",  (Func1Ptr)&std::ceil,          true);
  parser.setFunction1(L"floor", (Func1Ptr)&std::floor,         true);
  parser.setFunction1(L"round", (Func1Ptr)&FormulaUtil::round, true);
  parser.setFunction1(L"wrap",  (Func1Ptr)&FormulaUtil::wrap,  true);
  
  //  Op֐
  parser.setFunction1(L"acos",  (Func1Ptr)&std::acos,          true);
  parser.setFunction1(L"asin",  (Func1Ptr)&std::asin,          true);
  parser.setFunction1(L"atan",  (Func1Ptr)&std::atan,          true);
  parser.setFunction2(L"atan2", (Func2Ptr)&std::atan2,         true);
  parser.setFunction1(L"cos",   (Func1Ptr)&std::cos,           true);
  parser.setFunction1(L"cosh",  (Func1Ptr)&std::cosh,          true);
  parser.setFunction1(L"sin",   (Func1Ptr)&std::sin,           true);
  parser.setFunction1(L"sinh",  (Func1Ptr)&std::sinh,          true);
  parser.setFunction1(L"tan",   (Func1Ptr)&std::tan,           true);
  parser.setFunction1(L"tanh",  (Func1Ptr)&std::tanh,          true);
  
  //  wΐ֐
  parser.setFunction1(L"exp",   (Func1Ptr)&std::exp,           true);
  parser.setFunction1(L"log",   (Func1Ptr)&std::log,           true);
  parser.setFunction1(L"log10", (Func1Ptr)&std::log10,         true);
  parser.setFunction2(L"pow",   (Func2Ptr)&std::pow,           true);
  parser.setFunction1(L"sqrt",  (Func1Ptr)&std::sqrt,          true);

  //  gݍݒ萔o^.
  parser.setLiteral(L"PI", (float)M_PI);
  parser.setLiteral(L"E",  (float)M_E);
}

#if 0
// VS 2013 ŃRpCƃG[ô, Ƃ肠RgAEg

template <>
void installStandardFormulaFunc(FormulaParser<std::complex<float> > &parser)
{
  typedef std::complex<float> (*Func1Ptr)(std::complex<float> const &);
  typedef std::complex<float> (*Func2Ptr)(std::complex<float> const &, std::complex<float> const &);
  
  //  f֐
  parser.setFunction1(L"abs",   &std::abs<float>,               true);
  parser.setFunction1(L"arg",   &std::arg<float>,               true);
  parser.setFunction1(L"conj",  &std::conj<float>,              true);
  parser.setFunction1(L"imag",  &std::imag<float>,              true);
  parser.setFunction1(L"norm",  &std::norm<float>,              true);
  parser.setFunction1(L"real",  &std::real<float>,              true);
  parser.setFunction1(L"ceil",  (Func1Ptr)&FormulaUtil::ceil,   true);
  parser.setFunction1(L"floor", (Func1Ptr)&FormulaUtil::floor,  true);
  parser.setFunction1(L"round", (Func1Ptr)&FormulaUtil::round,  true);
  parser.setFunction1(L"wrap",  (Func1Ptr)&FormulaUtil::wrap,   true);
  
  //  Op֐
  parser.setFunction1(L"acos",  (Func1Ptr)&FormulaUtil::acos,   true);
  parser.setFunction1(L"asin",  (Func1Ptr)&FormulaUtil::asin,   true);
  parser.setFunction1(L"atan",  (Func1Ptr)&FormulaUtil::atan,   true);
  parser.setFunction2(L"atan2", (Func2Ptr)&FormulaUtil::atan2,  true);
  parser.setFunction1(L"cos",   &std::cos<float>,               true);
  parser.setFunction1(L"cosh",  &std::cosh<float>,              true);
  parser.setFunction1(L"sin",   &std::sin<float>,               true);
  parser.setFunction1(L"sinh",  &std::sinh<float>,              true);
  parser.setFunction1(L"tan",   &std::tan<float>,               true);
  parser.setFunction1(L"tanh",  &std::tanh<float>,              true);
  
  //  wΐ֐
  parser.setFunction1(L"exp",   &std::exp<float>,               true);
  parser.setFunction1(L"log",   &std::log<float>,               true);
  parser.setFunction1(L"log10", &std::log10<float>,             true);
  parser.setFunction2(L"pow",   (Func2Ptr)&std::pow<float>,     true);
  parser.setFunction1(L"sqrt",  &std::sqrt<float>,              true);
    
  //  gݍݒ萔o^.
  parser.setLiteral(L"PI", (float)M_PI);
  parser.setLiteral(L"E",  (float)M_E);
}

template <>
void installStandardFormulaFunc(FormulaParser<std::complex<double> > &parser)
{
  typedef std::complex<double> (*Func1Ptr)(std::complex<double> const &);
  typedef std::complex<double> (*Func2Ptr)(std::complex<double> const &, std::complex<double> const &);
  
  //  f֐
  parser.setFunction1(L"abs",   &std::abs<double>,              true);
  parser.setFunction1(L"arg",   &std::arg<double>,              true);
  parser.setFunction1(L"conj",  &std::conj<double>,             true);
  parser.setFunction1(L"imag",  &std::imag<double>,             true);
  parser.setFunction1(L"norm",  &std::norm<double>,             true);
  parser.setFunction1(L"real",  &std::real<double>,             true);
  parser.setFunction1(L"ceil",  (Func1Ptr)&FormulaUtil::ceil,   true);
  parser.setFunction1(L"floor", (Func1Ptr)&FormulaUtil::floor,  true);
  parser.setFunction1(L"round", (Func1Ptr)&FormulaUtil::round,  true);
  parser.setFunction1(L"wrap",  (Func1Ptr)&FormulaUtil::wrap,   true);
  
  //  Op֐
  parser.setFunction1(L"acos",  (Func1Ptr)&FormulaUtil::acos,   true);
  parser.setFunction1(L"asin",  (Func1Ptr)&FormulaUtil::asin,   true);
  parser.setFunction1(L"atan",  (Func1Ptr)&FormulaUtil::atan,   true);
  parser.setFunction2(L"atan2", (Func2Ptr)&FormulaUtil::atan2,  true);
  parser.setFunction1(L"cos",   &std::cos<double>,              true);
  parser.setFunction1(L"cosh",  &std::cosh<double>,             true);
  parser.setFunction1(L"sin",   &std::sin<double>,              true);
  parser.setFunction1(L"sinh",  &std::sinh<double>,             true);
  parser.setFunction1(L"tan",   &std::tan<double>,              true);
  parser.setFunction1(L"tanh",  &std::tanh<double>,             true);
  
  //  wΐ֐
  parser.setFunction1(L"exp",   &std::exp<double>,              true);
  parser.setFunction1(L"log",   &std::log<double>,              true);
  parser.setFunction1(L"log10", &std::log10<double>,            true);
  parser.setFunction2(L"pow",   (Func2Ptr)&std::pow<double>,    true);
  parser.setFunction1(L"sqrt",  &std::sqrt<double>,             true);
    
  //  gݍݒ萔o^.
  parser.setLiteral(L"PI", (double)M_PI);
  parser.setLiteral(L"E",  (double)M_E);
}

#endif

//--------------------------------------------------------------------------------
