#pragma once

//--------------------------------------------------------------------------------
#include <sstream>
#include <cmath>
#include <complex>

//#define _USE_MATH_DEFINES
//#include <math.h>

namespace FormulaUtil {
//--------------------------------------------------------------------------------
//  g[N񂩂琔lւ̕ϊ.
template <typename T>
bool convTokenToValue(std::wstring const &token, T &value)
{
  //  lւ̕ϊ݂.
  std::wistringstream conv(token);
  conv >> value;
  return !conv.fail() && conv.eof();
}

template <typename T>
bool convComplexToValue(std::wstring const &token, std::complex<T> &value)
{
  bool imag = false;
  std::wstring tempToken;
  if (token[0] == 'i' || token[0] == 'j')
  {
    imag = true;
    tempToken = token.substr(1);
  }
  else if (token[token.size() - 1] == 'i' || token[token.size() - 1] == 'j')
  {
    imag = true;
    tempToken = token.substr(0, token.size() - 1);
  }
  else
  {
    tempToken = token;
  }

  T tempValue;
  if (convTokenToValue<T>(tempToken, tempValue))
  {
    if (imag) value = std::complex<T>(0, tempValue);
    else      value = std::complex<T>(tempValue);
    return true;
  }
  else
  {
    return false;
  }
}

template <>
inline bool convTokenToValue(std::wstring const &token, std::complex<float>  &value)
{
  return convComplexToValue(token, value);
}

template <>
inline bool convTokenToValue(std::wstring const &token, std::complex<double>  &value)
{
  return convComplexToValue(token, value);
}

//--------------------------------------------------------------------------------
//  lg[Nւ̕ϊ.
template <typename T>
std::wstring convValueToToken(T const &value)
{
  std::wostringstream conv;
  conv << value;
  return conv.str();
}

template <typename T>
std::wstring convValueToComplex(std::complex<T> const &value)
{
  std::wostringstream conv;
  if      (value.real() == 0) conv << value.imag() << L'i';
  else if (value.imag() == 0) conv << value.real();
  else                        conv << L'(' << value.real() << value.imag() << L"i)";
  return conv.str();
}

template <>
inline std::wstring convValueToToken(std::complex<float> const &value)
{
  return convValueToComplex(value);
}

template <>
inline std::wstring convValueToToken(std::complex<double> const &value)
{
  return convValueToComplex(value);
}

//--------------------------------------------------------------------------------
/**
 *  ]߂.
 */
template <typename T> inline T mod(T const &x, T const &y)                { return x % y; }
template <>           inline float mod(float const &x, float const &y)    { return std::fmod(x, y); }
template <>           inline double mod(double const &x, double const &y) { return std::fmod(x, y); }

template <>
inline std::complex<float> mod(std::complex<float> const &x, std::complex<float> const &y)
{
  if (x.imag() != 0 || y.imag() != 0) throw std::runtime_error("Operator \'%\' is unavailable for complex number.");
  return std::fmod(x.real(), y.real());
}

template <>
inline std::complex<double> mod(std::complex<double> const &x, std::complex<double> const &y)
{
  if (x.imag() != 0 || y.imag() != 0) throw std::runtime_error("Operator \'%\' is unavailable for complex number.");
  return std::fmod(x.real(), y.real());
}

//--------------------------------------------------------------------------------
} //  namespace FormulaUtil
