#include "GraphData.h"
#include <cmath>
#include <cfloat>
#include <functional>

//--------------------------------------------------------------------------------
GraphDataY::GraphDataY()
{
}

GraphDataYRefPtr GraphDataY::Create()
{
  return GraphDataYRefPtr(new GraphDataY());
}

ResultCode GraphDataY::CreateMipmap(int layerNumber, ProcContext *pPC)
{
  using namespace std::placeholders;

  ResultCode result = RunProcWithContext<ResultCode>
    (std::bind(&GraphDataY::CreateMipmapInternal, this, layerNumber, _1), pPC);

  return result;
}

ResultCode GraphDataY::CreateMipmapInternal(int layerNumber, Progress *pProgress)
{
  if (layerNumber == mipmap.size())
  {
    return ResultCode::SUCCESS;
  }

  // V~bv}bvi[邽߂̗̈𐶐.
  mipmap.clear();

  size_t size = raw.size();
  int progUpper = 0;
  try
  {
    for (int layer = 0; layer < layerNumber; ++layer)
    {
      size /= 2;
      progUpper += size;
      mipmap.push_back(MipmapLayer());
      mipmap.back().resize(size);
    }
  }
  catch (std::bad_alloc &)
  {
    // s
    mipmap.clear();
    return ResultCode::MEMORY_ERROR;
  }

  if (layerNumber <= 0) return ResultCode::SUCCESS;

  // vOX̊Jn.
  if (pProgress != NULL)
  {
    pProgress->UpdateMessage(_("Creating Mipmap..."));
  }
  int pos = 0;
  int progStep = progUpper / 100;
  if (progStep == 0) progStep = 1;

  //  1 Kw̃~bv}bv𐶐.
  {
    MipmapLayer &layer = mipmap.front();
    std::vector<double>::const_iterator iterRaw = raw.begin();

    for (MipmapLayer::iterator j = layer.begin(); j != layer.end(); ++j, ++pos)
    {
      double const &v1 = *(iterRaw++);
      double const &v2 = *(iterRaw++);
      if (v1 > v2)
      {
        j->upper = v1;
        j->lower = v2;
      }
      else
      {
        j->upper = v2;
        j->lower = v1;
      }

      if (0 == (pos % progStep) && pProgress != NULL)
      {
        pProgress->UpdateProgress(pos, progUpper);
        if (pProgress->IsAbortRequested())
        {
          // f
          mipmap.clear();
          return ResultCode::ABORT;
        }
      }
    }
  }

  if (layerNumber <= 1) return ResultCode::SUCCESS;

  //  2 Kwȍ~̃~bv}bv𐶐.
  for (Mipmap::iterator i = mipmap.begin() + 1; i != mipmap.end(); ++i)
  {
    MipmapLayer &layer1 = *(i - 1);
    MipmapLayer &layer2 = *i;
    MipmapLayer::const_iterator iterUp = layer1.begin();

    for (MipmapLayer::iterator j = layer2.begin(); j != layer2.end(); ++j, ++pos)
    {
      MipmapValue const &v1 = *(iterUp++);
      MipmapValue const &v2 = *(iterUp++);
      if (v1.upper > v2.upper)
      {
        j->upper = v1.upper;
      }
      else
      {
        j->upper = v2.upper;
      }
      
      if (v1.lower > v2.lower)
      {
        j->lower = v2.lower;
      }
      else
      {
        j->lower = v1.lower;
      }

      if (0 == (pos % progStep) && pProgress != NULL)
      {
        pProgress->UpdateProgress(pos, progUpper);
        if (pProgress->IsAbortRequested())
        {
          // f
          mipmap.clear();
          return ResultCode::ABORT;
        }
      }
    }
  }

  pProgress->UpdateProgress(progUpper, progUpper);
  return ResultCode::SUCCESS;
}

void GraphDataY::CalcMinMax(double &min, double &max) const
{
  if (mipmap.empty())
  {
    // ~bv}bvȂƂ͐f[^őƍŏ𒲂ׂ.
    min = DBL_MAX;
    max = -DBL_MAX;
    for (std::vector<double>::const_iterator i = raw.begin(); i != raw.end(); ++i)
    {
      if (*i < min) min = *i;
      if (*i > max) max = *i;
    }
  }
  else
  {
    // ԃf[^Ȃ~bv}bvKwőƍŏ𒲂ׂ.
    min = DBL_MAX;
    max = -DBL_MAX;
    MipmapLayer const &layer = mipmap.back();
    for (MipmapLayer::const_iterator i = layer.begin(); i != layer.end(); ++i)
    {
      if (i->lower < min) min = i->lower;
      if (i->upper > max) max = i->upper;
    }
  }
}

//--------------------------------------------------------------------------------
GraphData::GraphData(GraphDataX const &dataX, GraphDataYRefPtr dataY)
  : m_dataX(dataX),
    m_dataY(dataY),
    m_enabled(true),
    m_maxY(0.0), m_minY(0.0),
    m_minMaxYValid(false)
{
}

GraphData::~GraphData()
{
}

GraphDataRefPtr GraphData::Create(GraphDataX const &dataX, GraphDataYRefPtr dataY)
{
  return GraphDataRefPtr(new GraphData(dataX, dataY));
}

ResultCode GraphData::CreateMipmap(int layerNumber, ProcContext *pPC)
{
  return m_dataY->CreateMipmap(layerNumber, pPC);
  /*GraphDataUpdate param(GraphDataUpdate::MIPMAP_UPDATED);
  Notify(param);*/
}

ResultCode GraphData::CreateMipmapForResolution(int resolution, ProcContext *pPC)
{
  int size = m_dataY->raw.size();
  int layerNumber = static_cast<int>(std::ceil(std::log(static_cast<double>(size) / resolution) / std::log(2.0)));
  return m_dataY->CreateMipmap(layerNumber, pPC);
}

#if 0
void GraphData::RemoveRange(int begin, int last, Progress *pProgress)
{

}

void GraphData::TrimRange(int begin, int last, Progress *pProgress)
{

}

void GraphData::Downsample(int begin, int interval, Progress *pProgress)
{

}
#endif

GraphDataX GraphData::GetDataX() const
{
  return m_dataX;
}

GraphDataYRefPtr GraphData::GetDataY() const
{
  return m_dataY;
}

int GraphData::GetPlotData(PlotData &plotData, int begin, int last, int resolution) const
{
  //wxASSERT(begin >= 0 && begin < static_cast<int>(m_dataY->raw.size()));
  //wxASSERT(last >= 0 && last < static_cast<int>(m_dataY->raw.size()));
  if (begin < 0) begin = 0;
  if (begin >= static_cast<int>(m_dataY->raw.size())) begin = static_cast<int>(m_dataY->raw.size()) - 1;
  if (last < 0) last = 0;
  if (last >= static_cast<int>(m_dataY->raw.size())) last = static_cast<int>(m_dataY->raw.size()) - 1;

  // K؂ȃ~bv}bvKwZo.
  int size = last - begin + 1;
  if (size <= 0) return 0;
  int level = static_cast<int>(std::floor(std::log(static_cast<double>(size) / resolution) / std::log(2.0)));

  int maxLevel = m_dataY->mipmap.size();
  if (level > maxLevel) level = maxLevel;

  if (level <= 0)
  {
    // RAW f[^vbgׂƂ
    plotData.useRaw = true;
    plotData.dataX.begin = m_dataX.begin + m_dataX.interval * begin;
    plotData.dataX.interval = m_dataX.interval;
    plotData.beginRaw = m_dataY->raw.begin() + begin;
    plotData.endRaw = m_dataY->raw.begin() + last + 1;
    return size;
  }
  else
  {
    // ~bv}bvf[^vbgׂƂ
    int scale = (1 << level);

    //wxASSERT((begin / scale) < static_cast<int>(m_dataY->mipmap[level - 1].size()));
    //wxASSERT((last / scale) < static_cast<int>(m_dataY->mipmap[level - 1].size()));

    plotData.useRaw = false;
    plotData.dataX.begin = m_dataX.begin + m_dataX.interval * (begin / scale) * scale;
    plotData.dataX.interval = m_dataX.interval * scale;

    int mipmapSize = m_dataY->mipmap[level - 1].size();
    int beginIndex = begin / scale;
    if (beginIndex < 0) beginIndex = 0;
    if (beginIndex >= mipmapSize) beginIndex = mipmapSize - 1;
    int endIndex = last / scale + 1;
    if (endIndex < 0) endIndex = 0;
    if (endIndex >= mipmapSize) endIndex = mipmapSize - 1;

    plotData.beginMip = m_dataY->mipmap[level - 1].begin() + beginIndex;
    plotData.endMip = m_dataY->mipmap[level - 1].begin() + endIndex;
    
    //wxASSERT(plotData.beginMip >= m_dataY->mipmap[level - 1].begin());
    //wxASSERT(plotData.endMip <= m_dataY->mipmap[level - 1].end());

    return plotData.endMip - plotData.beginMip;
  }
}

int GraphData::GetPlotDataInRangeX(PlotData &plotData, double begin, double last, int resolution) const
{
  return GetPlotData(
    plotData,
    static_cast<int>(std::ceil((begin - m_dataX.begin) / m_dataX.interval)),
    static_cast<int>(std::floor((last - m_dataX.begin) / m_dataX.interval)),
    resolution);
}

double GraphData::GetMaxY() const
{
  if (!m_minMaxYValid)
  {
    m_dataY->CalcMinMax(m_minY, m_maxY);
    m_minMaxYValid = true;
  }
  return m_maxY;
}

double GraphData::GetMinY() const
{
  if (!m_minMaxYValid)
  {
    m_dataY->CalcMinMax(m_minY, m_maxY);
    m_minMaxYValid = true;
  }
  return m_minY;
}

double GraphData::GetMaxX() const
{
  return m_dataX.begin + m_dataX.interval * (m_dataY->raw.size() - 1);
}

double GraphData::GetMinX() const
{
  return m_dataX.begin;
}

int GraphData::GetCount() const
{
  return m_dataY->raw.size();
}
  
void GraphData::SetLabel(wxString const &label)
{
  if (m_label != label)
  {
    m_label = label;
    GraphDataUpdate param(GraphDataUpdate::LABEL_UPDATED);
    Notify(param);
  }
}

wxString GraphData::GetLabel() const
{
  return m_label;
}

void GraphData::Enable(bool enabled)
{
  if (m_enabled != enabled)
  {
    m_enabled = enabled;
    GraphDataUpdate param(GraphDataUpdate::ENABLED_UPDATED);
    Notify(param);
  }
}

bool GraphData::IsEnabled() const
{
  return m_enabled;
}

void GraphData::SetPlotProperty(GraphDataPlotProperty const &prop)
{
  /*if (m_prop != prop)
  {*/
    m_prop = prop;
    GraphDataUpdate param(GraphDataUpdate::PROP_UPDATED);
    Notify(param);
  /*}*/
}

GraphDataPlotProperty GraphData::GetPlotProperty() const
{
  return m_prop;
}

//--------------------------------------------------------------------------------
