#pragma once

#include <memory>
#include <vector>
#include <wx/wx.h>
#include "Base.h"
#include "Progress.h"
#include "Observer.h"

class GraphData;
typedef std::shared_ptr<GraphData> GraphDataRefPtr;

class GraphDataY;
typedef std::shared_ptr<GraphDataY> GraphDataYRefPtr;

//--------------------------------------------------------------------------------
/**
 * Otvf[^̒ʒmp[^.
 */
class GraphDataUpdate
{
public:
  enum
  {
    //DATA_X_UPDATED  = 0x00000001, ///<< X f[^̍XV
    //DATA_Y_UPDATED  = 0x00000002, ///<< Y f[^̍XV
    //MIPMAP_UPDATED  = 0x00000004, ///<< ~bv}bv̍XV
    LABEL_UPDATED   = 0x00000008, ///<< x̍XV
    ENABLED_UPDATED = 0x00000010, ///<< Cl[u̍XV
    PROP_UPDATED    = 0x00000020  ///<< vpeB̍XV
  };

  GraphDataUpdate() : m_content(0) { }
  explicit GraphDataUpdate(int content) : m_content(content) { }

  /**
   * XVeݒ肵܂.
   */
  void SetContents(int content) { m_content = content; }

  /**
   * XVe擾܂.
   */
  int Contents() const { return m_content; }

  /**
   * w̓eXVꂽǂmF܂.
   */
  bool IsUpdated(int content) { return (m_content & content) != 0; }

private:
  int m_content;  // XVe
};

//--------------------------------------------------------------------------------
/**
 * Ot X \f[^ (X ͓Ԋu).
 */
struct GraphDataX
{
  GraphDataX()
    : begin(0.0), interval(1.0) { }
  
  double begin;     // CfbNX 0 ̍W.
  double interval;  // W̊Ԋu.
};

inline bool operator == (GraphDataX const &lhs, GraphDataX const &rhs)
{
  return lhs.begin == rhs.begin && lhs.interval == rhs.interval;
}

inline bool operator != (GraphDataX const &lhs, GraphDataX const &rhs)
{
  return lhs.begin != rhs.begin || lhs.interval != rhs.interval;
}

//--------------------------------------------------------------------------------
/**
 * Ot̃~bv}bvf[^̗vf.
 */
struct MipmapValue
{
  MipmapValue()
    : lower(0.0), upper(0.0) { }

  double lower;
  double upper;
};

/**
 * Kw̃~bv}bvf[^.
 */
typedef std::vector<MipmapValue> MipmapLayer;

/**
 * ~bv}bvf[^.
 */
typedef std::vector<MipmapLayer> Mipmap;

//--------------------------------------------------------------------------------
/**
 * Ot̃vbgɕKvȃf[^.
 */
struct PlotData
{
  typedef std::vector<double>::const_iterator RawIter;
  typedef MipmapLayer::const_iterator MipmapIter;

  PlotData()
    : useRaw(false) { }

  bool useRaw;                  // RAW (Kw[) ̃f[^vbgׂƂ true, ~bv}bvvbgׂƂ false.
  GraphDataX dataX;             // x f[^.
  RawIter beginRaw, endRaw;     // RAW f[^̃Ce[^.
  MipmapIter beginMip, endMip;  // ~bv}bvf[^̃Ce[^.
};

//--------------------------------------------------------------------------------
/**
 * Ot Y \f[^.
 */
class GraphDataY
{
public:
  /**
   * Ot Y \f[^𐶐܂.
   * @param raw swap() \bhɂ, f[^ɂȂ邱Ƃɒ.
   */
  static GraphDataYRefPtr Create();

  /**
   * ~bv}bv𐶐܂.
   * @todo
   * svȊKw̍Đ͍sȂ悤ɂ.
   */
  ResultCode CreateMipmap(int layerNumber, ProcContext *pPC);

  /**
   * ŏlƍőlvZ܂.
   */
  void CalcMinMax(double &min, double &max) const;

  std::vector<double> raw;
  Mipmap mipmap;

private:
  GraphDataY();
  GraphDataY(GraphDataY const &);
  GraphDataY &operator = (GraphDataY const &);

  ResultCode CreateMipmapInternal(int layerNumber, Progress *pProgress);
};

//--------------------------------------------------------------------------------
/**
 * Otf[^̃vbgvpeB.
 */
class GraphDataPlotProperty
{
public:
  GraphDataPlotProperty()
    : m_lineW(1.0),
      m_lineColor(*wxBLUE)
  {
  }

  void SetLineWidth(double val) { m_lineW = val; }
  void SetLineColor(wxColour const &val) { m_lineColor = val; }

  double LineWidth() const { return m_lineW; }
  wxColour LineColor() const { return m_lineColor; }

private:
  double m_lineW;                     // Ot̃C
  wxColour m_lineColor;               // Ot̃CF
};

//--------------------------------------------------------------------------------
/**
 * Otf[^.
 */
class GraphData : public Observable
{
public:
  virtual ~GraphData();

  /**
   * Otf[^𐶐܂.
   * @param dataX Ot X \f[^.
   * @param dataY Ot Y \f[^.
   * @return Otf[^IuWFNg.
   */
  static GraphDataRefPtr Create(GraphDataX const &dataX, GraphDataYRefPtr dataY);

  /**
   * ~bv}bv𐶐܂.
   */
  ResultCode CreateMipmap(int layerNumber, ProcContext *pPC);

  /**
   * w̕\ (|Cg) ȉɂȂ܂ł̃~bv}bvKw𐶐܂.
   */
  ResultCode CreateMipmapForResolution(int resolution, ProcContext *pPC);

#if 0
  /**
   * w͈͂̃f[^폜܂.
   * IɃ~bv}bvĐ܂.
   */
  void RemoveRange(int begin, int last, ProcContext *pPC);

  /**
   * w͈͊Õf[^폜܂.
   * IɃ~bv}bvĐ܂.
   */
  void TrimRange(int begin, int last, ProcContext *pPC);

  /**
   * f[^Ԉ܂.
   * IɃ~bv}bvĐ܂.
   */
  void Downsample(int begin, int interval, ProcContext *pPC);
#endif

  /**
   * X f[^擾܂.
   */
  GraphDataX GetDataX() const;

  /**
   * Y f[^擾܂.
   */
  GraphDataYRefPtr GetDataY() const;

  /**
   * `ɕKvȃf[^擾܂.
   * @param plotData `f[^.
   * @param begin `Jn|Cg̃CfbNX.
   * @param last `I|Cg̃CfbNX.
   * @param resolution \ɗpf[^_̍ŏl.
   * @return `f[^_.
   */
  int GetPlotData(PlotData &plotData, int begin, int last, int resolution) const;

  /**
   * `ɕKvȃf[^擾܂.
   * @param plotData `f[^.
   * @param begin `Jn X l.
   * @param last `I X l.
   * @param resolution \ɗpf[^_̍ŏl.
   * @return `f[^_.
   */
  int GetPlotDataInRangeX(PlotData &plotData, double begin, double last, int resolution) const;

  /**
   * Y ̍ől擾܂.
   */
  double GetMaxY() const;

  /**
   * Y ̍ŏl擾܂.
   */
  double GetMinY() const;

  /**
   * X ̍ől擾܂.
   */
  double GetMaxX() const;

  /**
   * X ̍ŏl擾܂.
   */
  double GetMinX() const;

  /**
   * Otf[^̌Ԃ܂.
   * @return Otf[^̌.
   */
  int GetCount() const;
  
  /**
   * xݒ肵܂.
   */
  void SetLabel(wxString const &label);

  /**
   * x擾܂.
   */
  wxString GetLabel() const;

  /**
   * vbgLɂ܂.
   */
  void Enable(bool enabled);

  /**
   * vbgLł邩ǂ擾܂.
   */
  bool IsEnabled() const;

  /**
   * vbg̃vpeBݒ肵܂.
   */
  void SetPlotProperty(GraphDataPlotProperty const &prop);

  /**
   * vbg̃vbg擾܂.
   */
  GraphDataPlotProperty GetPlotProperty() const;

private:
  GraphData();
  GraphData(GraphDataX const &dataX, GraphDataYRefPtr dataY);
  GraphData(GraphData const &);
  GraphData &operator = (GraphData const &);

  GraphDataX m_dataX;
  GraphDataYRefPtr m_dataY;
  wxString m_label;
  bool m_enabled;
  GraphDataPlotProperty m_prop;

  mutable double m_maxY, m_minY;  // Y ̍ől, ŏl
  mutable bool m_minMaxYValid;    // Y ̍ől, ŏlL?
};

//--------------------------------------------------------------------------------
