#include "GraphPanel.h"

#include "Cairo.h"
#include "Config.h"

//--------------------------------------------------------------------------------
wxDEFINE_EVENT(wxEVT_COMMAND_NOTIFY_COORDINATES, wxCommandEvent);

//--------------------------------------------------------------------------------
GraphPanel::GraphPanel(wxWindow* parent, wxWindowID id,
    const wxPoint& pos, const wxSize& size, long style)
  : GraphPanelDesign(parent, id, pos, size, style),
    m_proj(NULL),
    m_renderingCompleted(false),
    m_zoomXEnabled(true), m_zoomYEnabled(true),
    //m_zoomRate(1.1),
    m_leftDown(false),
    m_pt0(0, 0),
    m_minX0(0.0), m_maxX0(1.0), m_minY0(0.0), m_maxY0(1.0),
    m_isInnerOfGraph(false),
    m_graphCoor(0.0, 0.0)
{
}

GraphPanel::~GraphPanel()
{
}

void GraphPanel::SetProject(GraphProject *project)
{
  if (m_proj != project)
  {
    if (m_proj != NULL)
    {
      m_proj->Detach(this);
    }
    m_proj = project;
    if (m_proj != NULL)
    {
      m_proj->Attach(this);
      m_graphWindow->Refresh();
    }
  }
}

void GraphPanel::GraphWindow_OnEraseBackground(wxEraseEvent &event)
{
}

void GraphPanel::GraphWindow_OnMouseEvents(wxMouseEvent &event)
{
  if (NULL == m_proj) return;

  //wxLogDebug("MainWindow_OnMouseEvents()");

  wxPoint pt = event.GetPosition();

  //wxLogDebug("x = %d, y = %d, delta = %d, rotation = %d\n", pt.x, pt.y, delta, rotation);

  if (wxEVT_LEFT_DOWN == event.GetEventType())
  {
    // hbOɂXN[Jn
    m_leftDown = true;
    m_pt0 = pt;
    m_proj->GetPlotRange(m_minX0, m_maxX0, m_minY0, m_maxY0);

    // IuU[o̒ʒme|[hɂ, vpeBplŃhbO
    // XVȂ悤ɂ.
    m_proj->EnableNotifyTemporal();

    //CaptureMouse();
  }
  else if (wxEVT_MOUSEWHEEL == event.GetEventType())
  {
    // }EXzC[ɂY[
    if (m_renderingCompleted)
    {
      //int delta = event.GetWheelDelta();
      int rotation = event.GetWheelRotation();
      double zoom = ReadConf(CONF_MOUSE_ZOOM);
      //double zoom = m_zoomRate;
      if (rotation > 0) zoom = 1.0 / zoom;

      wxPoint2DDouble center = m_layout.TransformClientToGraphPoint(pt);
      if (m_zoomXEnabled) m_proj->ZoomPlotRangeX(center.m_x, zoom);
      if (m_zoomYEnabled) m_proj->ZoomPlotRangeY(center.m_y, zoom);
    }
  }
  else if (wxEVT_LEFT_UP == event.GetEventType() || wxEVT_MOTION == event.GetEventType())
  {
    // Ot̃vbgWvZ.
    wxPoint2DDouble delta = m_layout.TransformClientToGraphDistance(pt - m_pt0);
    double minX = m_minX0 - delta.m_x;
    double maxX = m_maxX0 - delta.m_x;
    double minY = m_minY0 - delta.m_y;
    double maxY = m_maxY0 - delta.m_y;

    if (wxEVT_LEFT_UP == event.GetEventType())
    {
      // hbOɂXN[I
      m_leftDown = false;

      // IuU[o̒ʒm̃e|[h܂.
      m_proj->EnableNotifyTemporal(false);

      m_proj->SetPlotRange(minX, maxX, minY, maxY);
      wxLogDebug("GraphWindow_OnMouseEvents(): eventType = wxEVT_LEFT_UP");
      
      //ReleaseMouse();
    }
    else /* if (wxEVT_MOTION == event.GetEventType()) */
    {
      // hbOɂXN[
      if (m_renderingCompleted && m_leftDown)
      {
        // IuU[o̒ʒme|[hɂ, vpeBplŃhbO
        // XVȂ悤ɂ.
        m_proj->SetPlotRange(minX, maxX, minY, maxY);
      }

      // WCxgŒʒm.
      wxPoint2DDouble ptf(pt);
      bool isInnerOfGraph = m_renderingCompleted && m_layout.rectGraphFrame.Contains(ptf);
      wxPoint2DDouble coor = m_layout.TransformClientToGraphPoint(pt);

      int notifyState = -1;
      if (!m_isInnerOfGraph && isInnerOfGraph)
      {
        notifyState = NOTIFY_COORDINATES_ENTER;
        m_graphCoor = coor;
      }
      else if (m_isInnerOfGraph && !isInnerOfGraph)
      {
        notifyState = NOTIFY_COORDINATES_LEAVE;
        // Oẗ̊OɏoĂ̂ m_graphCoor ͍XVȂ.
      }
      else if (isInnerOfGraph && m_graphCoor != coor)
      {
        notifyState = NOTIFY_COORDINATES_MOVE;
        m_graphCoor = coor;
      }
      m_isInnerOfGraph = isInnerOfGraph;

      if (notifyState != -1)
      {
        wxCommandEvent e(wxEVT_COMMAND_NOTIFY_COORDINATES, GetId());
        e.SetEventObject(this);
        e.SetInt(notifyState);
        ProcessWindowEvent(e);  // Cxg𑗐M.
      }

      // }EXJ[\̕ύX
      if (m_renderingCompleted && m_graphWindow->HasFocus() && m_layout.rectGraphFrame.Contains(ptf))
      {
        m_graphWindow->SetCursor(*wxCROSS_CURSOR);
      }
      else
      {
        m_graphWindow->SetCursor(*wxSTANDARD_CURSOR);
      }
    }
  }

  // wxEVT_LEFT_DOWN Cxgnhł, event.Skip() ĂяoKv.
  // ̏Ȃꍇ, EBhE̓tH[JX󂯎邱ƂłȂ.
  // EBhEtH[JX󂯎邱ƂłȂꍇ, }EXzC[̃Cxg
  // 邱ƂłȂ.
  if (wxEVT_LEFT_DOWN == event.GetEventType())
  {
    event.Skip();
  }
}

void GraphPanel::GraphWindow_OnPaint(wxPaintEvent &event)
{
  if (NULL == m_proj) return;

  // Ql
  // https://code.google.com/p/wxcairo/wiki/IntegratingWithCairo#Rendering_on_Windows
  wxPaintDC dc(m_graphWindow);
  
  wxRect rect = m_graphWindow->GetClientRect();
  
  if (0 == rect.width || 0 == rect.height)
  {
    return;
  }

  // Get the handle of the wxWindow
  HWND hwnd = static_cast<HWND>(m_graphWindow->GetHandle());

  // Now get the native windows DC for this window
  HDC hdc = ::GetDC(hwnd);

  // Create a double buffer for blitting to
  // the screen to prevent screen flicker. Pass
  // the double buffer to cairo and blit it
  // in the paint routine.
  HDC dcbuffer = ::CreateCompatibleDC(hdc);
  HBITMAP hbuffer = ::CreateCompatibleBitmap(hdc, rect.width, rect.height);
  SelectObject(dcbuffer, hbuffer); 

  {
    // Create a cairo surface we can draw on directly
    Cairo cairo(dcbuffer);
    if (cairo.GetType() != Cairo::TYPE_INVALID)
    {
      // Cairo T[tF[X, ReLXg̐ɐƂ, _Os.

      // Render the output into the cairo image
      if (m_proj->GetCount() != 0)
      {
        int accuracy = 1;
        GraphRenderer renderer(cairo, *m_proj);
        renderer.SetArea(rect.width, rect.height);
        renderer.SetResolution(-1, accuracy);
        renderer.Render();
        renderer.GetLayout(m_layout);
        m_renderingCompleted = true;

        // tH[JXƂ, g`悷.
        if (m_graphWindow->HasFocus())
        {
          ::cairo_rectangle(cairo(), 0.0, 0.0, rect.width, rect.height);
          ::cairo_set_source_rgb(cairo(), 0.0, 0.0, 0.8);
          ::cairo_set_line_width(cairo(), 4.0);
          ::cairo_stroke(cairo());
        }

        // `̌, 肳ObhԊuvWFNg vpeBɔf.
        // ObhԊu̎ݒ肪Ƃɖɂ.
        GraphProjectPlotProperty prop = m_proj->GetPlotProperty();
        prop.SetGridIntervalX(m_layout.gridIntervalX);
        prop.SetGridIntervalY(m_layout.gridIntervalY);

        // ꎞIɃIuU[o̒ʒm𖳌ɂ, vpeBp[^ݒ肷.
        m_proj->EnableNotify(false);
        m_proj->SetPlotProperty(prop);
        m_proj->EnableNotify();
      }
      else
      {
        ::cairo_rectangle(cairo(), 0.0, 0.0, rect.width, rect.height);
        ::cairo_set_source_rgb(cairo(), 0.8, 0.8, 0.8);
        ::cairo_fill(cairo());
        m_renderingCompleted = false;
      }
    }

    // Now blit the object to the screen
    ::BitBlt(hdc, 0, 0, rect.width, rect.height, dcbuffer, 0, 0, SRCCOPY);

    // Tear down the cairo object in cairo object destructor,
    // now that we don't need it anymore .
  }

  ::DeleteDC(dcbuffer);
  ::DeleteObject(hbuffer);

  // Because we called ::GetDC make sure we release the handle
  // back to the system or we'll have a memory leak.
  ::ReleaseDC(hwnd, hdc);
}

void GraphPanel::GraphWindow_OnEnterWindow(wxMouseEvent &event)
{
  wxLogDebug("GraphWindow_OnEnterWindow()");
}

void GraphPanel::GraphWindow_OnLeaveWindow(wxMouseEvent &event)
{
  wxLogDebug("GraphWindow_OnLeaveWindow()");
}

void GraphPanel::GraphWindow_OnSetFocus(wxFocusEvent &event)
{
  wxLogDebug("GraphWindow_OnSetFocus()");
  m_graphWindow->Refresh();
}

void GraphPanel::GraphWindow_OnKillFocus(wxFocusEvent &event)
{
  wxLogDebug("GraphWindow_OnKillFocus()");
  m_graphWindow->Refresh();
}

void GraphPanel::Update(Observable *sender, ObserverParamBase const &param)
{
  GraphProjectUpdate update;
  if (GetUserParam(param, update))
  {
    // OtvWFNgXVƂ͍ĕ`.
    m_graphWindow->Refresh();
  }
}

#if 0
void MainFrame::SetMainWindowScrollbars()
{
  if (0 == m_proj->GetCount())
  {
    m_scrollBarVert->Disable();
    m_scrollBarHorz->Disable();
  }
  else
  {
    double minX, maxX, minY, maxY;
    double plotMinX, plotMaxX, plotMinY, plotMaxY;
    m_proj->GetRange(minX, maxX, minY, maxY);
    m_proj->GetPlotRange(plotMinX, plotMaxX, plotMinY, plotMaxY);

    int const rangeX = 100000, rangeY = 100000;
    int const pageX = rangeX / 10, pageY = rangeY / 10;
    int posX = static_cast<int>(rangeX * (plotMinX - minX) / (maxX - minX));
    int posY = static_cast<int>(rangeY * (plotMinY - minY) / (maxY - minY));
    int thumbX = static_cast<int>(rangeX * (plotMaxX - plotMinX) / (maxX - minX));
    int thumbY = static_cast<int>(rangeY * (plotMaxY - plotMinY) / (maxY - minY));

    m_scrollBarVert->SetScrollbar(posX, thumbX, rangeX, pageX);
    m_scrollBarHorz->SetScrollbar(posY, thumbY, rangeY, pageY);
    m_scrollBarVert->Enable();
    m_scrollBarHorz->Enable();
  }
}
#endif

//--------------------------------------------------------------------------------
