#pragma once

#include <memory>
#include <vector>
#include "GraphData.h"
#include "Observer.h"

class GraphProject;
typedef std::tr1::shared_ptr<GraphProject> GraphProjectRefPtr;

//--------------------------------------------------------------------------------
/**
 * OtvWFNg̒ʒmp[^.
 */
class GraphProjectUpdate
{
public:
  enum
  {
    TEMPORAL      = 0x80000000, ///<< ꎞIȍXVł邱ƂtO (XN[ɐݒ肳).
    DATA_APPENDED = 0x00000001, ///<< Otf[^̒ǉ.
    DATA_REMOVED  = 0x00000002, ///<< Otf[^̍폜.
    DATA_UPDATED  = 0x00000004, ///<< Otf[^̍XV.
    RANGE_UPDATED = 0x00000008, ///<< `惌W̍XV.
    PROP_UPDATED  = 0x00000010  ///<< vpeB̍XV.
  };

  GraphProjectUpdate() : m_content(0) { }

  explicit GraphProjectUpdate(int content) : m_content(content), m_data(NULL) { }

  /**
   * XVeݒ肵܂.
   */
  void SetContents(int content) { m_content = content; }

  /**
   * XVe TEMPORAL tOt܂.
   */
  void SetTemporal() { m_content |= TEMPORAL; }

  /**
   * XVe擾܂.
   */
  int Contents() const { return m_content; }

  /**
   * w̓eXVꂽǂmF܂.
   */
  bool IsUpdated(int content) const { return (m_content & content) != 0; }

  /**
   * TEMPORAL tOtĂ邩ǂmF܂.
   */
  bool IsTemporal() const { return (m_content & TEMPORAL) != 0; }

  /**
   * XVf[^ݒ肵܂.
   * XVe DATA_UPDATED łƂ̂ݗLł.
   */
  void SetData(GraphData *data) { m_data = data; }

  /**
   * XVf[^擾܂.
   * XVe DATA_UPDATED łƂ̂ݗLł.
   */
  GraphData *Data() const { return m_data; }

  /**
   * XVf[^̏ڍׂݒ肵܂.
   * XVe DATA_UPDATED łƂ̂ݗLł.
   */
  void SetDataUpdateDetail(GraphDataUpdate detail) { m_detail = detail; }
  
  /**
   * XVf[^̏ڍׂ擾܂.
   * XVe DATA_UPDATED łƂ̂ݗLł.
   */
  GraphDataUpdate DataUpdateDetail() const { return m_detail; }

private:
  int m_content;            // XVe.
  GraphData *m_data;        // f[^̃CfbNX.
  GraphDataUpdate m_detail; // f[^̍XVe.
};

//--------------------------------------------------------------------------------
/**
 * OtvWFNg̃f[^̃vbgvpeB.
 */
class GraphProjectPlotProperty
{
public:
  GraphProjectPlotProperty();

  // ̂߂ɗpOt vWFNg̏ݒ
  /*void SetGraphProject(GraphProject *proj) { m_proj = proj; }
  void SetGraphArea(double w, double h) { m_graphAreaW = w; m_graphAreaH = h; }*/

  // 擾\bh
  wxColour BgColor() const { return m_bgColor; }
  
  double MarginTop() const { return m_marginT; }
  double MarginBottom() const { return m_marginB; }
  double MarginLeft() const { return m_marginL; }
  double MarginRight() const { return m_marginR; }
  
  double GraphAreaFrameWidth() const { return m_graphAreaFrameW; }
  wxColour GraphAreaFrameColor() const { return m_graphAreaFrameColor; }

  bool GridXEnabled() const { return m_gridXEnabled; }
  bool GridYEnabled() const { return m_gridYEnabled; }
  bool GridIntervalXAuto() const { return m_gridIntervalXAuto; }
  bool GridIntervalYAuto() const { return m_gridIntervalYAuto; }
  double GridIntervalX() const { return m_gridIntervalX; }
  double GridIntervalY() const { return m_gridIntervalY; }
  double GridLineWidth() const { return m_gridLineW; }
  wxColour GridLineColor() const { return m_gridLineColor; }

  double CalcGridIntervalX(GraphProject const &proj, double graphAreaW) const; // ̂Ƃ̓ObhȂ
  double CalcGridIntervalY(GraphProject const &proj, double graphAreaH) const; // ̂Ƃ̓ObhȂ

  double TickLineLength() const { return m_tickLineLength; }
  double TickLineWidth() const { return m_tickLineW; }
  wxColour TickLineColor() const { return m_tickLineColor; }

  wxFont TickLabelFont() const { return m_tickLabelFont; }
  wxColour TickLabelFontColor() const { return m_tickLabelFontColor; }
  double TickLabelPosX() const { return m_tickLabelPosX; }
  double TickLabelPosY() const { return m_tickLabelPosY; }
  wxString TickLabelFormatX() const { return m_tickLabelFormatX; }
  wxString TickLabelFormatY() const { return m_tickLabelFormatY; }

  wxString AxisXLabel() const { return m_axisXLabel; }
  wxString AxisYLabel() const { return m_axisYLabel; }
  wxFont AxisLabelFont() const { return m_axisLabelFont; }
  wxColour AxisLabelFontColor() const { return m_axisLabelFontColor; }
  double AxisXLabelPosY() const { return m_axisXLabelPosY; }
  double AxisYLabelPosX() const { return m_axisYLabelPosX; }

  // ݒ胁\bh
  void SetBgColor(wxColour const &val) { m_bgColor = val; }

  void SetMarginTop(double val) { m_marginT = val; }
  void SetMarginBottom(double val) { m_marginB = val; }
  void SetMarginLeft(double val) { m_marginL = val; }
  void SetMarginRight(double val) { m_marginR = val; }

  void SetGraphAreaFrameWidth(double val) { m_graphAreaFrameW = val; }
  void SetGraphAreaFrameColor(wxColour const &val) { m_graphAreaFrameColor = val; }
    
  void SetGridXEnabled(bool val) { m_gridXEnabled = val; }
  void SetGridYEnabled(bool val) { m_gridYEnabled = val; }
  void SetGridIntervalXAuto(bool val) { m_gridIntervalXAuto = val; }
  void SetGridIntervalYAuto(bool val) { m_gridIntervalYAuto = val; }
  void SetGridIntervalX(double val) { m_gridIntervalX = val; }
  void SetGridIntervalY(double val) { m_gridIntervalY = val; }
  void SetGridLineWidth(double val) { m_gridLineW = val; }
  void SetGridLineColor(wxColour const &val) { m_gridLineColor = val; }

  void SetTickLineLength(double val) { m_tickLineLength = val; }
  void SetTickLineWidth(double val) { m_tickLineW = val; }
  void SetTickLineColor(wxColour const &val) { m_tickLineColor = val; }

  void SetTickLabelFont(wxFont const &val) { m_tickLabelFont = val; }
  void SetTickLabelFontColor(wxColour const &val) { m_tickLabelFontColor = val; }
  void SetTickLabelPosX(double val) { m_tickLabelPosX = val; }
  void SetTickLabelPosY(double val) { m_tickLabelPosY = val; }
  void SetTickLabelFormatX(wxString const &val) { m_tickLabelFormatX = val; }
  void SetTickLabelFormatY(wxString const &val) { m_tickLabelFormatY = val; }

  void SetAxisXLabel(wxString const &val) { m_axisXLabel = val; }
  void SetAxisYLabel(wxString const &val) { m_axisYLabel = val; }
  void SetAxisLabelFont(wxFont const &val) { m_axisLabelFont = val; }
  void SetAxisLabelFontColor(wxColour const &val) { m_axisLabelFontColor = val; }
  void SetAxisXLabelPosY(double val) { m_axisXLabelPosY = val; }
  void SetAxisYLabelPosX(double val) { m_axisYLabelPosX = val; }

  bool IsEqualTo(GraphProjectPlotProperty const &rhs) const;

private:
  wxColour m_bgColor;                                 // wiF

  double m_marginT, m_marginB, m_marginL, m_marginR;  // `GAƃOtGÃ}[W
  
  double m_graphAreaFrameW;                           // OtGÃC
  wxColour m_graphAreaFrameColor;                     // OtGÃCF

  bool m_gridXEnabled, m_gridYEnabled;                // Obh̗L
  bool m_gridIntervalXAuto, m_gridIntervalYAuto;      // ObhԊuݒ
  double m_gridIntervalX, m_gridIntervalY;            // ObhԊu
  double m_gridLineW;                                 // Obh
  wxColour m_gridLineColor;                           // ObhF

  double m_tickLineLength;                            // ڐ̒
  double m_tickLineW;                                 // ڐ̐
  wxColour m_tickLineColor;                           // ڐ̐F

  wxFont m_tickLabelFont;                             // ڐ胉x̃tHg
  wxColour m_tickLabelFontColor;                      // ڐ胉x̃tHg̐F
  double m_tickLabelPosX;                             // ڐ胉x X ʒu (OtGA[̑Έʒu, ō, ŉE)
  double m_tickLabelPosY;                             // ڐ胉x Y ʒu (OtGA[̑Έʒu, ŉ, ŏ)
  wxString m_tickLabelFormatX;                        // ڐ胉x X tH[}bg
  wxString m_tickLabelFormatY;                        // ڐ胉x Y tH[}bg

  wxString m_axisXLabel;                              // X x
  wxString m_axisYLabel;                              // Y x
  wxFont m_axisLabelFont;                             // x̃tHg
  wxColour m_axisLabelFontColor;                      // x̃tHg̐F
  double m_axisXLabelPosY;                            // X x Y ʒu (OtGA[̑Έʒu, ŉ, ŏ)
  double m_axisYLabelPosX;                            // Y x X ʒu (OtGA[̑Έʒu, ō, ŉE)
};

inline bool operator == (GraphProjectPlotProperty const &lhs, GraphProjectPlotProperty const &rhs)
{
  return lhs.IsEqualTo(rhs);
}

inline bool operator != (GraphProjectPlotProperty const &lhs, GraphProjectPlotProperty const &rhs)
{
  return !lhs.IsEqualTo(rhs);
}

//--------------------------------------------------------------------------------
/**
 * Ot vWFNg.
 */
class GraphProject : public Observable, public Observer
{
public:
  ~GraphProject();

  /**
   * Ot vWFNg𐶐܂.
   */
  static GraphProjectRefPtr Create();

  /**
   * Ot f[^ǉ܂.
   */
  void Append(GraphDataRefPtr refData);
  
  /**
   * Ot f[^폜܂.
   */
  void Remove(int index);

  /**
   * Ot f[^擾܂.
   */
  GraphDataRefPtr GetItem(int index) const;

 /**
   * w̕\ (|Cg) ȉɂȂ܂ł̃~bv}bvKw𐶐܂.
   */
  void CreateMipmapForResolution(int resolution, ProcContext *pPC);
  
  /**
   * Y ̍ől擾܂.
   * @param onlyEnabledData Lȃf[^Ɍ肷邩ǂw肵܂.
   */
  double GetMaxY(bool onlyEnabledData = false) const;

  /**
   * Y ̍ŏl擾܂.
   * @param onlyEnabledData Lȃf[^Ɍ肷邩ǂw肵܂.
   */
  double GetMinY(bool onlyEnabledData = false) const;

  /**
   * X ̍ől擾܂.
   * @param onlyEnabledData Lȃf[^Ɍ肷邩ǂw肵܂.
   */
  double GetMaxX(bool onlyEnabledData = false) const;

  /**
   * X ̍ŏl擾܂.
   * @param onlyEnabledData Lȃf[^Ɍ肷邩ǂw肵܂.
   */
  double GetMinX(bool onlyEnabledData = false) const;

  /**
   * ŏlƍől擾܂.
   * @param onlyEnabledData Lȃf[^Ɍ肷邩ǂw肵܂.
   */
  void GetRange(double &minX, double &maxX, double &minY, double &maxY, bool onlyEnabledData = false) const;

  /**
   * Ot f[^̐擾܂.
   */
  int GetCount() const;

  /**
   * vbg̃vpeBݒ肵܂.
   * @param prop      vpeBݒl.
   */
  void SetPlotProperty(GraphProjectPlotProperty const &prop);

  /**
   * vbg̃vbg擾܂.
   */
  GraphProjectPlotProperty GetPlotProperty() const;

  /**
   * ftHg̃vbg͈͂ݒ肵܂.
   * X , Y ̍ŏl, ől玩ݒ肵܂.
   * onlyEnabledData = true ƂƂɗL (Enable) ȃf[^ 1 ȂƂ,
   * ׂẴf[^vbg͈͂肵܂.
   * f[^ 1 ȂƂ, 0 ` 1 ͈̔͂vbg͈͂ɂ܂.
   * ŏlƍőlvƂ, ̒ĺ} 1 ͈̔͂vbg͈͂ɂ܂.
   * @param directionX      true ̂Ƃ, X ̃vbg͈͂ŏl, ől玩ݒ肵܂.
   * @param directionY      true ̂Ƃ, Y ̃vbg͈͂ŏl, ől玩ݒ肵܂.
   * @param onlyEnabledData Lȃf[^Ɍ肷邩ǂw肵܂.
   */
  void SetDefaultPlotRange(bool directionX = true, bool directionY = true, bool onlyEnabledData = false);

  /**
   * vbgW Y ̍ől擾܂.
   */
  double GetPlotMaxY() const;

  /**
   * vbgW Y ̍ŏl擾܂.
   */
  double GetPlotMinY() const;

  /**
   * vbgW X ̍ől擾܂.
   */
  double GetPlotMaxX() const;

  /**
   * vbgW X ̍ŏl擾܂.
   */
  double GetPlotMinX() const;

  /**
   * vbg͈͂擾܂.
   */
  void GetPlotRange(double &minX, double &maxX, double &minY, double &maxY) const;

  /**
   * vbg͈͂ݒ肵܂.
   */
  void SetPlotRange(double minX, double maxX, double minY, double maxY);

  /**
   * IuU[o̒ʒmݒ̗L/ݒ肵܂.
   * ɐݒ肵ꍇ, Lɂ܂ł̊Ԃ̕ύX͖,
   * ʒm邱Ƃ͂܂.
   * @param notifyEnabled   false ݒ肷邱Ƃ, IuU[oɂʒmIɖɂ܂.
   *                        ObhԊu̎ݒ (GridIntervalXAuto, GridIntervalYAuto) L
   *                        ĂƂ, ObhԊu (GridIntervalX, GridIntervalY) ̕ύXʒm
   *                        KvȂƂɗp܂. ftHg true ł.
   */
  void EnableNotify(bool notifyEnabled = true);

  /**
   * IuU[o̒ʒmɂ, e|[hݒ肵܂.
   * e|[hLɐݒ肳, ȍ~̒ʒmɂ TEMPORAL tO
   * t, p[^̕ύXꎞIł邱ƂӖ܂.
   * e|[hꂽ (e|[hL疳ɐ؂ւ)
   * Ƃ, e|[hɒʒme TEMPORAL tO̕tȂ, ēxʒm܂.
   * , Ēʒme RANGE_UPDATED  PROP_UPDATED Ɍ肳܂.
   */
  void EnableNotifyTemporal(bool notifyTemporal = true);

  /**
   * vbg͈݂͈͂̔͂瑊ΓIɃY[܂ (X ).
   * @param centerX Y[̃Z^[ʒu X.
   * @param zoomX   X g嗦.
   */
  void ZoomPlotRangeX(double centerX, double zoomX);

  /**
   * vbg͈݂͈͂̔͂瑊ΓIɃY[܂ (Y ).
   * @param centerY Y[̃Z^[ʒu Y.
   * @param zoomY   Y g嗦.
   */
  void ZoomPlotRangeY(double centerY, double zoomY);

  /**
   * vbg͈݂͈͂̔͂瑊ΓIɃY[܂ (X ).
   * @param zoomX   X g嗦.
   */
  void ZoomPlotRangeX(double zoomX);

  /**
   * vbg͈݂͈͂̔͂瑊ΓIɃY[܂ (Y ).
   * @param zoomY   Y g嗦.
   */
  void ZoomPlotRangeY(double zoomY);

protected:
  // ϑΏۂ̒ʒm̃R[obN֐.
  virtual void Update(Observable *sender, ObserverParamBase const &param);

private:
  GraphProject();
  GraphProject(GraphProject const &);
  GraphProject &operator = (GraphProject const &);

  void NotifyGraphProjectUpdate(GraphProjectUpdate const &userParam);

  std::vector<GraphDataRefPtr> m_data;    // f[^
  GraphProjectPlotProperty m_prop;        // vpeB
  double m_minX, m_maxX, m_minY, m_maxY;  // vbg͈

  bool m_notifyEnabled, m_notifyTemporal; // IuU[oʒmݒ
  int m_renotifyContent;                  // e|[h̍Ēʒme
};

//--------------------------------------------------------------------------------
