#include "MainFrame.h"

#include <functional>
#include <wx/event.h>
#include "GraphPanel.h"
#include "ProjectPanel.h"
#include "PropertyPanel.h"
#include "ManualPanel.h"
#include "CSVImporter.h"
#include "CSVImportOptionDlg.h"
#include "CSVExporter.h"
#include "CSVExportOptionDlg.h"
#include "SeparatorSettingDlg.h"
#include "CalculatorDlg.h"
#include "ProgressDlg.h"
#include "FormulaCalculator.h"
#include "ImageSizeSettingDlg.h"
#include "AboutDlg.h"
#include "Config.h"

//--------------------------------------------------------------------------------
// }`Xbhs
#define USE_MULTITHREAD_PROCESS

//--------------------------------------------------------------------------------
static int const STATUS_FIELDS_COUNT = 2;            // Xe[^Xo[̃tB[h̐
static int const STATUS_IDX_MSG = 0;                 // bZ[W̕\̈
static int const STATUS_IDX_COOR = 1;                // W̕\̈
static int const STATUS_WIDTH_ARRAY[] = { -1, 150 }; // tB[h̕

//--------------------------------------------------------------------------------
// RXgN^, fXgN^
MainFrame::MainFrame()
  : MainFrameDesign(NULL),
    m_modified(false),
    m_zoomXEnabled(true), m_zoomYEnabled(true),
    m_timerStatusMsg(NULL)
{
  wxLogDebug("MainFrame::MainFrame(): ThreadID = %d", wxThread::GetCurrentId());

#if !defined(NDEBUG)
  // fobOpj[ǉ
  m_menuHelp->AppendSeparator();

	m_menuItemDebugProgressDlg = new wxMenuItem(m_menuHelp, wxID_ANY, _T("Debug ProgressDlg"), wxEmptyString, wxITEM_NORMAL );
	m_menuHelp->Append(m_menuItemDebugProgressDlg);

  Connect(m_menuItemDebugProgressDlg->GetId(), wxEVT_COMMAND_MENU_SELECTED, wxCommandEventHandler(MainFrame::Command_OnDebugProgressDlg));

  wxLogDebug("Debug menus are appended.");
#endif

  // Xe[^Xo[.
  m_statusBar->SetFieldsCount(STATUS_FIELDS_COUNT, STATUS_WIDTH_ARRAY);
  
  // Xe[^XbZ[Wp^C}.
  m_timerStatusMsg = new wxTimer(this);
  Connect(m_timerStatusMsg->GetId(), wxEVT_TIMER, wxTimerEventHandler(MainFrame::TimerStatusMsg_OnTimeout));

  // Xe[^Xo[ɋNbZ[W\.
  SetStatusMsg(_("Thank you for using CSV Viewer!"));

  // AUI CAEgǂݍ.
  // AUI plʂ邽߂ wxFormBuilder  aui_name Kv邱Ƃɒӂ邱.
  /*wxString perspective = ReadConf(CONF_MAINFRAME_PERS);
  if (!perspective.IsEmpty())
  {
    m_mgr.LoadPerspective(perspective, false);
  }*/

  // Otpl̍WʒmCxgnhݒ肷.
  Connect(m_panelGraph->GetId(), wxEVT_COMMAND_NOTIFY_COORDINATES, wxCommandEventHandler(MainFrame::PanelGraph_OnNotifyCoordinates));

  // ACRݒ肷
  SetIcon(wxICON(APP_ICON));

  // vWFNg𐶐.
  CreateProject();
}

MainFrame::~MainFrame()
{
  Disconnect(m_panelGraph->GetId(), wxEVT_COMMAND_NOTIFY_COORDINATES, wxCommandEventHandler(MainFrame::PanelGraph_OnNotifyCoordinates));
  Disconnect(m_timerStatusMsg->GetId(), wxEVT_TIMER, wxTimerEventHandler(MainFrame::TimerStatusMsg_OnTimeout));

  if (m_timerStatusMsg) delete m_timerStatusMsg;
}

//--------------------------------------------------------------------------------
// R}h Cxgnh
void MainFrame::OnClose(wxCloseEvent& event)
{
  // AUI CAEgۑ.
  // ȂOĂ܂Ȃ̂ŕۗ.
  /*wxString perspective = m_mgr.SavePerspective();
  WriteConf(CONF_MAINFRAME_PERS, perspective);*/

  event.Skip();
}

void MainFrame::Command_OnNewProj(wxCommandEvent& event)
{
  if (m_modified)
  {
    int ans = wxMessageBox(_("This project is modified. Do you want to remove all?"), _("New Project"),
      wxICON_EXCLAMATION | wxYES_NO, this);
    if (ans != wxYES)
    {
      return; // abort
    }
  }

  CreateProject();

  SetStatusMsg(_("Creating new project is completed!"));
}

void MainFrame::Command_OnImport(wxCommandEvent& event)
{
  bool dataEmpty = (0 == m_proj->GetCount());
  
  // t@C[UɊmF.
  wxFileDialog fileDlg(this, _("Import"), _T(""), _T(""), _T("CSV files (*.csv)|*.csv|TXT files (*.txt)|*.txt"),
    wxFD_OPEN | wxFD_FILE_MUST_EXIST);

  if (fileDlg.ShowModal() == wxID_CANCEL)
  {
    return;
  }

  // eLXgt@ĈƂ̓Zp[^mF.
  wxUniChar separator(_T(','));
  wxFileName fileName(fileDlg.GetPath());
  if (_T("txt") == fileName.GetExt())
  {
    SeparatorSettingDlg separatorSettingDlg(this);
    separatorSettingDlg.SetSeparator(separator);
    if (separatorSettingDlg.ShowModal() == wxID_CANCEL)
    {
      return;
    }
    separator = separatorSettingDlg.GetSeparator();
  }

  // CSV t@CXL, s, 񐔂𒲂ׂ.
  CSVImporter importer(fileName.GetFullPath(), *m_proj);
  importer.SetSeparator(separator);

  ProgressDlg progressDlg(this);
#if defined(USE_MULTITHREAD_PROCESS)
  ProcContext *pPC = &progressDlg;
#else
  ProcContext *pPC = NULL;
#endif

  ResultCode result = importer.ScanInfo(pPC);
  if (ResultCode::ABORT == result)
  {
    return;
  }
  else if (result != ResultCode::SUCCESS)
  {
    wxMessageBox(importer.GetLastErrorMessage(), _("Error"), wxOK | wxICON_ERROR, this);
    return;
  }

  // CSV t@C̓ǂݍ݃IvV_CAOŐݒ肷.
  CSVImportOptionDlg optionDlg(this);

  CSVFileInfo fileInfo;
  importer.GetFileInfo(fileInfo);
  optionDlg.SetFileInfo(fileInfo);

  CSVImportOption option;
  importer.GetOption(option);
  optionDlg.SetImportOption(option);

#if !defined(NDEBUG)
  option.LogDebug();
#endif

  int retCode = optionDlg.ShowModal();
  if (wxID_CANCEL == retCode)
  {
    return;
  }

  optionDlg.GetImportOption(option);
  importer.SetOption(option);

#if !defined(NDEBUG)
  option.LogDebug();
#endif

  // CSV t@CC|[g.
  int resolution = ReadConf(CONF_MIPMAP_RESO);
  result = importer.Import(resolution, pPC);
  if (ResultCode::ABORT == result)
  {
    return;
  }
  else if (result != ResultCode::SUCCESS)
  {
    wxMessageBox(importer.GetLastErrorMessage(), _("Error"), wxOK | wxICON_ERROR, this);
    return;
  }

  // ̃vWFNgɃf[^C|[gꍇ, ftHg̃vbg͈͂ݒ肷.
  if (dataEmpty)
  {
    m_proj->SetDefaultPlotRange();
  }

  SetStatusMsg(_("Importing is completed!"));
}

void MainFrame::Command_OnExport(wxCommandEvent& event)
{
  // CSV t@C̏݃IvV[U[ɊmF.
  CSVExportOptionDlg optionDlg(this);

  CSVExportOption option;
  option.SetDefaultFromProject(*m_proj);
  optionDlg.SetGraphProject(m_proj.get());
  optionDlg.SetExportOption(option);

  if (optionDlg.ShowModal() == wxID_CANCEL)
  {
    return;
  }

  optionDlg.GetExportOption(option);

#if !defined(NDEBUG)
  option.LogDebug();
#endif
  
  // t@C[UɊmF.
  wxFileDialog fileDlg(this, _("Export"), _T(""), _T(""), _T("CSV files (*.csv)|*.csv|TXT files (*.txt)|*.txt"),
    wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
  
  if (fileDlg.ShowModal() == wxID_CANCEL)
  {
    return;
  }

  // eLXgt@ĈƂ̓Zp[^mF.
  wxUniChar separator(_T(','));
  wxFileName fileName(fileDlg.GetPath());
  if (_T("txt") == fileName.GetExt())
  {
    SeparatorSettingDlg separatorSettingDlg(this);
    separatorSettingDlg.SetSeparator(separator);
    if (separatorSettingDlg.ShowModal() == wxID_CANCEL)
    {
      return;
    }
    separator = separatorSettingDlg.GetSeparator();
  }

  // GNX|[g.
  ProgressDlg progressDlg(this);
#if defined(USE_MULTITHREAD_PROCESS)
  ProcContext *pPC = &progressDlg;
#else
  ProcContext *pPC = NULL;
#endif
  
  CSVExporter exporter(fileName.GetFullPath(), *m_proj);
  exporter.SetOption(option);
  exporter.SetSeparator(separator);
  
  ResultCode result = exporter.Export(pPC);
  if (ResultCode::ABORT == result)
  {
    return;
  }
  else if (result != ResultCode::SUCCESS)
  {
    wxMessageBox(exporter.GetLastErrorMessage(), _("Error"), wxOK | wxICON_ERROR, this);
    return;
  }

  SetStatusMsg(_("Exporting is completed!"));
}

void MainFrame::Command_OnExportImage(wxCommandEvent& event)
{
  // f[^Ȃ͏I.
  if (0 == m_proj->GetCount() || !m_panelGraph->IsRenderingCompleted())
  {
    wxMessageBox(_("No Data!"), _("Error"), wxOK | wxICON_ERROR, this);
    return;
  }

  // C[WTCY擾܂.
  GraphRenderingLayout layout = m_panelGraph->GetLastRenderingLayout();

  ImageSizeSettingDlg sizeDlg(this);
  sizeDlg.SetWindowImageSize(layout.clientWidth, layout.clientHeight);
  if (sizeDlg.ShowModal() == wxID_CANCEL) return;
  int width, height;
  sizeDlg.GetImageSize(width, height);

  // t@C[UɊmF.
  wxFileDialog fileDlg(this, _("Export Image"), _T(""), _T(""),
    _T("PNG files (*.png)|*.png|")
    _T("SVG files (*.svg)|*.svg|")
    _T("PDF files (*.pdf)|*.pdf|")
    _T("PS files (*.ps)|*.ps"),
    wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
  if (fileDlg.ShowModal() == wxID_CANCEL) return;

  // _O.
  int accuracy = 1;
  Cairo cairo(fileDlg.GetPath(), width, height);
  if (cairo.GetType() == Cairo::TYPE_INVALID)
  {
    wxMessageBox(_("Internal error!"), _("Error"), wxOK | wxICON_ERROR, this);
    return;
  }
  GraphRenderer renderer(cairo, *m_proj);
  renderer.SetArea(width, height);
  renderer.SetResolution(-1, accuracy);
  renderer.Render();

  SetStatusMsg(_("Exporting the image is completed!"));
}

void MainFrame::Command_OnCalculate(wxCommandEvent& event)
{
  bool dataEmpty = (0 == m_proj->GetCount());

  // vZIuWFNg𐶐
  FormulaCalculator calc(*m_proj);

  // ̓_CAO𐶐
  CalculatorDlg calcDlg(this);
  calcDlg.SetGraphProject(m_proj.get());
  calcDlg.SetFormula(ReadConf(CONF_CALCDLG_FORMULA));

  bool success = false;
  do
  {
    int retCode = calcDlg.ShowModal();
    if (wxID_CANCEL == retCode)
    {
      return;
    }

    // vZIuWFNgɃ_CAO̒l𔽉f.
    calc.SetFormula(calcDlg.GetFormula());
    calc.SetDataX(calcDlg.GetGraphDataX());
    calc.SetDataCount(calcDlg.GetGraphDataSize());
    calc.SetLabel(calcDlg.GetGraphLabel());

    // p[X, AST 𐶐.
    success = calc.Parse();
    if (!success)
    {
      wxMessageBox(calc.GetLastErrorMessage(), _("Error"), wxOK | wxICON_ERROR, this);
    }
  }
  while (!success);

  WriteConf(CONF_CALCDLG_FORMULA, calcDlg.GetFormula());

  // AST pČvZs, ~bv}bv, vWFNgɒǉ.
  ProgressDlg progressDlg(this);
#if defined(USE_MULTITHREAD_PROCESS)
  ProcContext *pPC = &progressDlg;
#else
  ProcContext *pPC = NULL;
#endif
  
  using namespace std::placeholders;

  int resolution = ReadConf<int>(CONF_MIPMAP_RESO);
  ResultCode result = calc.Calculate(resolution, pPC);
  
  if (ResultCode::ABORT == result)
  {
    return;
  }
  else if (result != ResultCode::SUCCESS)
  {
    wxMessageBox(calc.GetLastErrorMessage(), _("Error"), wxOK | wxICON_ERROR, this);
    return;
  }

  // ̃vWFNgɃf[^C|[gꍇ, ftHg̃vbg͈͂ݒ肷.
  if (dataEmpty)
  {
    m_proj->SetDefaultPlotRange();
  }

  SetStatusMsg(_("Calculating the image is completed!"));
}

void MainFrame::Command_OnDelete(wxCommandEvent& event)
{
  int index = m_panelProj->GetSelectedGraphIndex();
  if (wxNOT_FOUND == index)
  {
    wxMessageBox(_("Please select some graph data!"), _("Error"), wxOK | wxICON_ERROR, this);
    return;
  }

  m_proj->Remove(index);

  SetStatusMsg(_("Deleting selected data is completed!"));
}

void MainFrame::Command_OnExit(wxCommandEvent& event)
{
  Close(true);
}

void MainFrame::Command_OnProjPanel(wxCommandEvent& event)
{
  wxAuiPaneInfo &info = m_mgr.GetPane(m_panelProj);
  info.Show(!info.IsShown());
  m_mgr.Update(); // update AUI manager
}

void MainFrame::Command_OnProjPanelUpdatedUI(wxUpdateUIEvent& event)
{
  wxAuiPaneInfo &info = m_mgr.GetPane(m_panelProj);
  event.Check(info.IsShown());
}

void MainFrame::Command_OnPropPanel(wxCommandEvent& event)
{
  wxAuiPaneInfo &info = m_mgr.GetPane(m_panelProp);
  info.Show(!info.IsShown());
  m_mgr.Update(); // update AUI manager
}

void MainFrame::Command_OnPropPanelUpdatedUI(wxUpdateUIEvent& event)
{
  wxAuiPaneInfo &info = m_mgr.GetPane(m_panelProp);
  event.Check(info.IsShown());
}

void MainFrame::Command_OnFit(wxCommandEvent& event)
{
  m_proj->SetDefaultPlotRange(true, true, true);
}

void MainFrame::Command_OnFitHorz(wxCommandEvent& event)
{
  m_proj->SetDefaultPlotRange(true, false, true);
}

void MainFrame::Command_OnFitVert(wxCommandEvent& event)
{
  m_proj->SetDefaultPlotRange(false, true, true);
}

void MainFrame::Command_OnZoomIn(wxCommandEvent& event)
{
  double zoom = ReadConf(CONF_MOUSE_ZOOM);
  if (m_zoomXEnabled) m_proj->ZoomPlotRangeX(1.0 / zoom);
  if (m_zoomYEnabled) m_proj->ZoomPlotRangeY(1.0 / zoom);
}

void MainFrame::Command_OnZoomOut(wxCommandEvent& event)
{
  double zoom = ReadConf(CONF_MOUSE_ZOOM);
  if (m_zoomXEnabled) m_proj->ZoomPlotRangeX(zoom);
  if (m_zoomYEnabled) m_proj->ZoomPlotRangeY(zoom);
}

void MainFrame::Command_OnZoomHorz(wxCommandEvent& event)
{
  m_zoomXEnabled = !m_zoomXEnabled;
  if (!m_zoomXEnabled && !m_zoomYEnabled)
  {
    // , ̏ȂƂБ͊g傷悤ɂ
    m_zoomYEnabled = true;
  }
  m_panelGraph->EnableZoomX(m_zoomXEnabled);
  m_panelGraph->EnableZoomY(m_zoomYEnabled);
}

void MainFrame::Command_OnZoomHorzUpdateUI(wxUpdateUIEvent& event)
{
  event.Check(m_zoomXEnabled);
}

void MainFrame::Command_OnZoomVert(wxCommandEvent& event)
{
  m_zoomYEnabled = !m_zoomYEnabled;
  if (!m_zoomXEnabled && !m_zoomYEnabled)
  {
    // , ̏ȂƂБ͊g傷悤ɂ
    m_zoomXEnabled = true;
  }
  m_panelGraph->EnableZoomX(m_zoomXEnabled);
  m_panelGraph->EnableZoomY(m_zoomYEnabled);
}

void MainFrame::Command_OnZoomVertUpdateUI(wxUpdateUIEvent& event)
{
  event.Check(m_zoomYEnabled);
}

void MainFrame::Command_OnManualPanel(wxCommandEvent& event)
{
  wxAuiPaneInfo &info = m_mgr.GetPane(m_panelManual);
  bool shown = !info.IsShown();
  if (shown)
  {
    // }jAǂݍ݂܂.
    m_panelManual->LoadManual();
  }
  info.Show(shown);
  m_mgr.Update(); // update AUI manager
}

void MainFrame::Command_OnManualPanelUpdateUI(wxUpdateUIEvent& event)
{
  wxAuiPaneInfo &info = m_mgr.GetPane(m_panelManual);
  event.Check(info.IsShown());
}

void MainFrame::Command_OnLicensePanel(wxCommandEvent& event)
{
  wxAuiPaneInfo &info = m_mgr.GetPane(m_panelLicense);
  bool shown = !info.IsShown();
  if (shown)
  {
    // CZXǂݍ݂܂.
    m_panelLicense->LoadLicense();
  }
  info.Show(shown);
  m_mgr.Update(); // update AUI manager
}

void MainFrame::Command_OnLicensePanelUpdateUI(wxUpdateUIEvent& event)
{
  wxAuiPaneInfo &info = m_mgr.GetPane(m_panelLicense);
  event.Check(info.IsShown());
}

void MainFrame::Command_OnAbout(wxCommandEvent& event)
{
  AboutDlg dlg(this);
  dlg.ShowModal();
}

void MainFrame::PanelGraph_OnNotifyCoordinates(wxCommandEvent& event)
{
  switch (event.GetInt())
  {
  case NOTIFY_COORDINATES_ENTER:
    //wxLogDebug("GraphPanel_OnNotifyCoordinates(): state = NOTIFY_COORDINATES_ENTER");
    break;
  case NOTIFY_COORDINATES_LEAVE:
    //wxLogDebug("GraphPanel_OnNotifyCoordinates(): state = NOTIFY_COORDINATES_LEAVE");
    m_statusBar->SetStatusText(_T(""), STATUS_IDX_COOR);
    break;
  case NOTIFY_COORDINATES_MOVE:
    //wxLogDebug("GraphPanel_OnNotifyCoordinates(): state = NOTIFY_COORDINATES_MOVE");
    {
      wxPoint2DDouble coor = m_panelGraph->GetGraphCoordinates();
      wxString text;
      text.Printf(_T("%f, %f"), coor.m_x, coor.m_y);
      m_statusBar->SetStatusText(text, STATUS_IDX_COOR);
    }
    break;
  }
}

void MainFrame::TimerStatusMsg_OnTimeout(wxTimerEvent& event)
{
  m_statusBar->SetStatusText(_T(""), STATUS_IDX_MSG);
}

void MainFrame::Update(Observable *sender, ObserverParamBase const &param)
{
  GraphProjectUpdate update;
  if (GetUserParam(param, update))
  {
    // vWFNgvWFNg̒ʒm
    m_modified = true;
  }
}

//--------------------------------------------------------------------------------
void MainFrame::CreateProject()
{
  // ݂̃vWFNg폜.
  m_panelGraph->SetProject(NULL);
  m_panelProj->SetProject(NULL);
  m_panelProp->SetProject(NULL);
  m_proj.reset();

  // VvWFNg𐶐.
  m_proj = GraphProject::Create();
  m_proj->Attach(this);

  // eGUIIuWFNgɐVvWFNgݒ肷.
  m_panelGraph->SetProject(m_proj.get());
  m_panelProj->SetProject(m_proj.get());
  m_panelProp->SetProject(m_proj.get());
  
  m_modified = false;
}

void MainFrame::SetStatusMsg(wxString const &msg)
{
  m_statusBar->SetStatusText(msg, STATUS_IDX_MSG);
  m_timerStatusMsg->StartOnce(5000);
}

//--------------------------------------------------------------------------------
#if !defined(NDEBUG)

void MainFrame::Command_OnDebugProgressDlg(wxCommandEvent &event)
{
  wxLogDebug("Start debugging ProgressDlg class...");

#if 0
  DummyProgress progress;
  DebugProgressFunc(&progress);
#else
  ProgressDlg progressDlg(this);
  bool ret = DebugProgress(&progressDlg);
  wxLogDebug("Ret = %d", ret);
#endif

  wxLogDebug("Debugging is completed!");
}

#endif

//--------------------------------------------------------------------------------
