#pragma once

#include <list>
#include <wx/thread.h>

class Observer;
class Observable;

typedef std::list<Observer *> ObserverList;
typedef std::list<Observable *> ObservableList;

//--------------------------------------------------------------------------------
/**
 *  IuU[o p[^̃x[X NX.
 */
class ObserverParamBase
{
public:
  virtual ~ObserverParamBase() { }
};

/**
 *  IuU[o p[^.
 */
template <typename T>
class ObserverParam : public ObserverParamBase
{
public:
  ObserverParam() { }
  explicit ObserverParam(T const &v) : value(v) { }
  T value;
};

template <typename T>
bool GetUserParam(ObserverParamBase const &param, T &userParam)
{
  ObserverParam<T> const *ptr = dynamic_cast<ObserverParam<T> const *>(&param);
  if (ptr != NULL)
  {
    // [U[p[^oƂłƂ
    userParam = ptr->value;
    return true;
  }
  else
  {
    return false;
  }
}

//--------------------------------------------------------------------------------
/**
 *  IuU[o (ϑ).
 */
class Observer
{
public:
  Observer();

  /**
   * fXgN^.
   * ׂẴIuU[ouɑ΂ Detach() Ăяo܂.
   */
  virtual ~Observer();

  /**
   * ڑĂ邷ׂĂ Observable 擾܂.
   */
  ObservableList const &GetConnectedObservables() const;

protected:
  /**
   * ʒmꂽɎsR[obNo֐.
   * Ƃ, ʒmɑΉ鏈Lq܂.
   * @param sender     ĂяõIuWFNg.
   * @param param      p[^.
   *
   * @note
   * ͎̒ʂł.
   * @code
   *  void Update(Observable *sender, ObserverParamBase const &param)
   *  {
   *    UserParamType userParam;            // Observable  Notify ꂽ[U[p[^.
   *    
   *    if (GetUserParam(param, userParam)) // [U[p[^擾݂܂.
   *    {
   *      // [U[p[^擾łƂ,
   *      // userParam gɏ܂.
   *    }
   *  }
   * @endcode
   */
  virtual void Update(Observable *sender, ObserverParamBase const &param);
  
private:
  // Rs[֎~
  Observer(Observer const &);
  Observer &operator = (Observer const &);
  
  // Observable ɐڑꂽɌĂяo郁o֐.
  // @param observable    ڑ Observable IuWFNg.
  void OnAttached(Observable *observable);

  // Observable ؒfꂽɌĂяo郁o֐.
  // @param observable    ؒf Observable IuWFNg.
  void OnDetached(Observable *observable);
  
  // ڑĂϑΏۂ̈ꗗ.
  // zQƂh, QƃJEg𑝂₳Ȃʂ̃|C^g.
  ObservableList m_observables;
  
  // Observable  OnAttached, OnDetached Ăяo悤ɂ邽.
  friend class Observable;
};

//--------------------------------------------------------------------------------
/**
 * IuU[ou (ϑΏ).
 * }`Xbhɂ͑Ή܂.  Notify() ֐, CXbhŌĂяo
 * [Ƃ܂. Notify() ֐CXbhƂ͈قȂXbhĂяoꂽ
 * ꍇ̓AT[g܂.
 */
class Observable
{
public:
  Observable();
  virtual ~Observable();

  /**
   * IuU[oڑ܂.
   * @param observer  ڑIuU[o[.
   * @return          Ɏw肵IuU[oڑĂƂ̓G[ƂȂ,
   *                  false Ԃ܂. Ƃ true Ԃ܂.
   *
   * @note
   * IuU[o̎QƃJEg𑝂₳, IuU[oXgɒǉ܂.
   * ȂȂ, observer ̔hNXł observable ̔hNXW񂷂
   * 邱Ƃ邽, X}[g|C^̏zQƂh߂ł.
   */  
  bool Attach(Observer *observer);

  /**
   * IuU[oؒf܂.
   * @param observer  ؒfIuU[o.
   * @return          w肵IuU[oIuU[oXgɊ܂܂ȂƂ̓G[ƂȂ,
   *                  false Ԃ܂. Ƃ true Ԃ܂.
   */
  bool Detach(Observer *observer);

  /**
   * ׂẴIuU[oؒf܂.
   */
  void DetachAll();

  /**
   * IuU[oڑĂ邩ǂׂ܂.
   * @param observer  ڑĂ邩ǂmFIuU[o.
   * return           ڑĂꍇ true, ڑĂȂꍇ false Ԃ܂.
   */
  bool IsAttached(Observer *observer) const;

  /**
   * ڑĂ邷ׂẴIuU[oɑ΂ʒm܂.
   * ̓Iɂ͐ڑĂ邷ׂẴIuU[o Update o֐Ăяo܂.
   * @param userParam p[^. Cӂ̌^̃p[^wł܂.
   */
  template <typename T>
  void Notify(T const &userParam)
  {
    ObserverParam<T> param(userParam);
    NotifyInternal(param);
  }

  /**
   * ڑĂ邷ׂĂ Observable 擾܂.
   */
  ObserverList const &GetConnectedObservers() const;

private:
  // Rs[֎~
  Observable(Observable const &);
  Observable &operator = (Observable const &);

  void NotifyInternal(ObserverParamBase const &param);

  // ڑĂIuU[o[̈ꗗ.
  // zQƂh, QƃJEg𑝂₳Ȃʂ̃|C^g.
  ObserverList m_observers;
};

//--------------------------------------------------------------------------------
