#pragma once

#include <memory>
#include <wx/string.h>

//--------------------------------------------------------------------------------
/**
 * vOX̃x[XNX.
 */
class Progress
{
public:
  virtual ~Progress() { }

  /**
   * vOX̐iAbvf[g܂.
   * @param pos   i pos / range Ŏw肵܂.
   * @param range i pos / range Ŏw肵܂.
   */
  virtual void UpdateProgress(int pos, int range) = 0;

  /**
   * bZ[WAbvf[g܂.
   * @param message bZ[W.
   */
  virtual void UpdateMessage(wxString const &message) = 0;

  /**
   * vOX\ĂRg[ŏ̒fvꂽǂ擾܂.
   */
  virtual bool IsAbortRequested() = 0;
};

//--------------------------------------------------------------------------------
/**
 * vOX̃_~[NX.
 * vOX̏SsȂ.
 */
class DummyProgress : public Progress
{
public:
  virtual ~DummyProgress() { }
  virtual void UpdateProgress(int pos, int range) { }
  virtual void UpdateMessage(wxString const &message) { }
  virtual bool IsAbortRequested() { return false; }
};

//--------------------------------------------------------------------------------
/**
 * vOX̎sNX̃x[XNX.
 */
class ProgressProcBase
{
public:
  virtual ~ProgressProcBase() { }
  virtual void Execute(Progress *progress) = 0;
};

/**
 * vOX̎sNX (߂l).
 */
template <typename RetValue, typename ProcFunc>
class ProgressProc : public ProgressProcBase
{
public:
  explicit ProgressProc(ProcFunc func) : m_func(func) { }
  virtual ~ProgressProc() { }
  virtual void Execute(Progress *progress) { m_ret = m_func(progress); }
  RetValue GetRetValue() const { return m_ret; }

private:
  // ftHgRXgN^֎~.
  ProgressProc();

  // Rs[֎~.
  ProgressProc(ProgressProc const &);
  ProgressProc &operator = (ProgressProc const &);

  ProcFunc m_func;
  RetValue m_ret;
};

/**
 * vOX̎sNX (߂lȂ).
 */
template <typename ProcFunc>
class ProgressProc<void, ProcFunc> : public ProgressProcBase
{
public:
  explicit ProgressProc(ProcFunc func) : m_func(func) { }
  virtual ~ProgressProc() { }
  virtual void Execute(Progress *progress) { m_func(progress); }

private:
  // ftHgRXgN^֎~.
  ProgressProc();

  // Rs[֎~.
  ProgressProc(ProgressProc const &);
  ProgressProc &operator = (ProgressProc const &);

  ProcFunc m_func;
};

//--------------------------------------------------------------------------------
/**
 * vV[W ReLXg̃x[XNX.
 * ׂۂ̃ReLXg\܂.
 *
 * \code
 *
 * \endcode
 */
class ProcContext
{
public:
  virtual void Run(std::shared_ptr<ProgressProcBase> proc) = 0;
};

//--------------------------------------------------------------------------------
/**
 * vV[WReLXgpĊ֐IuWFNgĂяo܂.
 * @param func  ֐IuWFNg.
 * @param pPC   vV[W ReLXg.
 *              NULL nƒڌďo܂. ̍ۃvOXIuWFNgƂ
 *              DummyProgress gp܂.
 * @return      ֐IuWFNg̖߂l.
 *
 * \code
 * // ProcContext ̗
 * class MyClass
 * {
 * public:
 *   short Method(int param, ProcContext *pPC);
 * 
 * private:
 *   short HeavyProc(int param, Progress *pProg);
 * };
 * 
 * short MyClass::Method(int param, ProcContext *pPC)
 * {
 *   ...
 * 
 *   // ev[g, vV[WReLXgŏ֐̖߂ľ^w肷.
 *   //  1 , std::bind() p Progress * ^̊֐IuWFNgn.
 *   //  2 , vV[WReLXgn.
 *   short result = RunProcWithContext<short>(std::bind(&MyClass::HeavyProc, this, param, _1), pPC);
 * 
 *   ...
 * 
 *   return result;
 * }
 * 
 * short MyClass::HeavyProc(int param, Progress *pProg)
 * {
 * 
 *   return 1;
 * }
 * \endcode
 */
template <typename RetValue, typename ProcFunc>
RetValue RunProcWithContext(ProcFunc func, ProcContext *pPC)
{
  if (NULL == pPC)
  {
    DummyProgress dummy;
    return func(&dummy);
  }
  else
  {
    typedef std::shared_ptr<ProgressProc<RetValue, ProcFunc> > ProgressProcRefPtr;
    ProgressProcRefPtr proc = ProgressProcRefPtr(new ProgressProc<RetValue, ProcFunc>(func));
    pPC->Run(proc);
    return proc->GetRetValue();
  }
}

template <typename ProcFunc>
void RunProcWithContext(ProcFunc func, ProcContext *pPC)
{
  if (NULL == pPC)
  {
    DummyProgress dummy;
    func(&dummy);
  }
  else
  {
    typedef std::shared_ptr<ProgressProc<void, ProcFunc> > ProgressProcRefPtr;
    ProgressProcRefPtr proc = ProgressProcRefPtr(new ProgressProc<RetValue, ProcFunc>(func));
    pPC->Run(proc);
  }
}

//--------------------------------------------------------------------------------
