#include "ProgressDlg.h"

#include <wx/log.h>

//--------------------------------------------------------------------------------
wxDEFINE_EVENT(wxEVT_COMMAND_PROGRESS_COMPLETED, wxThreadEvent);
wxDEFINE_EVENT(wxEVT_COMMAND_PROGRESS_UPDATE_POS, wxThreadEvent);
wxDEFINE_EVENT(wxEVT_COMMAND_PROGRESS_UPDATE_MSG, wxThreadEvent);

//--------------------------------------------------------------------------------
ProgressThread::ProgressThread(ProgressDlg *pDlg, std::tr1::shared_ptr<ProgressProcBase> proc)
  : wxThread(wxTHREAD_DETACHED),
    m_pDlg(pDlg),
    m_proc(proc)
{
}

ProgressThread::~ProgressThread()
{
  wxLogDebug("ProgressThread::~ProgressThread()");
  
  wxCriticalSectionLocker enter(m_pDlg->m_pThreadCS);
  // XbhjƂ, _CAO m_pThread |C^ȃ|C^
  // ȂȂ悤 NULL Ƃ.
  m_pDlg->m_pThread = NULL; 
}

void ProgressThread::UpdateProgress(int pos, int range)
{
  ProgressUpdateInfo info;
  info.pos = pos;
  info.range = range;

  // wxEVT_COMMAND_PROGRESS_UPDATE_POS CxgCXbh
  // _CAO ProgressDlg ɒʒm.
  wxThreadEvent *event = new wxThreadEvent(wxEVT_COMMAND_PROGRESS_UPDATE_POS);
  event->SetPayload(info);
  wxQueueEvent(m_pDlg, event);
}

void ProgressThread::UpdateMessage(wxString const &message)
{
  // wxEVT_COMMAND_PROGRESS_UPDATE_MSG CxgCXbh
  // _CAO ProgressDlg ɒʒm.
  wxThreadEvent *event = new wxThreadEvent(wxEVT_COMMAND_PROGRESS_UPDATE_MSG);
  event->SetString(message);
  wxQueueEvent(m_pDlg, event);
}

bool ProgressThread::IsAbortRequested()
{
  // Xbh̒~vĂꍇ true, łȂƂ false Ԃ.
  return TestDestroy();
}

wxThread::ExitCode ProgressThread::Entry()
{
  wxLogDebug("ProgressThread::Entry(): starting process");

  m_proc->Execute(this);

  wxLogDebug("ProgressThread::Entry(): stopping process");

  // ̃Xbhj邱Ƃʒm.
  // , pDlg ̃|C^ւ̃ANZX͈SłƉ肵Ă.
  wxQueueEvent(m_pDlg, new wxThreadEvent(wxEVT_COMMAND_PROGRESS_COMPLETED));
  return static_cast<wxThread::ExitCode>(0);     // success
}

//--------------------------------------------------------------------------------
ProgressDlg::ProgressDlg(wxWindow* parent)
  : ProgressDlgDesign(parent),
    m_pThread(NULL)
{
  Connect(wxID_ANY, wxEVT_COMMAND_PROGRESS_COMPLETED, wxThreadEventHandler(ProgressDlg::Thread_OnProgressCompleted));
  Connect(wxID_ANY, wxEVT_COMMAND_PROGRESS_UPDATE_POS, wxThreadEventHandler(ProgressDlg::Thread_OnProgressUpdatePos));
  Connect(wxID_ANY, wxEVT_COMMAND_PROGRESS_UPDATE_MSG, wxThreadEventHandler(ProgressDlg::Thread_OnProgressUpdateMsg));
}

ProgressDlg::~ProgressDlg()
{
  // fXgN^ł͂Ȃ, OnClose() CxgnhŃXbh̃N[ibv
  // ׂł. , ̗Rł.
  // Xbh̏IɃCxg (wxEVT_COMMAND_PROGRESS_COMPLETED) 𑗏oȂ, 
  // fXgN^ĂяoꂽƂ, ͂gbvxEBhẼCxg[v
  // ANeBuł͂Ȃ߂ł.
  // OnClose() XbhIȂ, Cxgnhŉ邱
  // łȂ.
  Disconnect(wxID_ANY, wxEVT_COMMAND_PROGRESS_COMPLETED, wxThreadEventHandler(ProgressDlg::Thread_OnProgressCompleted));
  Disconnect(wxID_ANY, wxEVT_COMMAND_PROGRESS_UPDATE_POS, wxThreadEventHandler(ProgressDlg::Thread_OnProgressUpdatePos));
  Disconnect(wxID_ANY, wxEVT_COMMAND_PROGRESS_UPDATE_MSG, wxThreadEventHandler(ProgressDlg::Thread_OnProgressUpdateMsg));
}

void ProgressDlg::Run(std::shared_ptr<ProgressProcBase> proc)
{
  m_proc = proc;

  // _CAO[_ŕ\, XbhN܂.
  /*int result =*/ ProgressDlgDesign::ShowModal();
}

void ProgressDlg::OnClose(wxCloseEvent &event)
{
  {
    wxCriticalSectionLocker enter(m_pThreadCS);
    if (m_pThread)         // does the thread still exist?
    {
      wxLogDebug("ProgressDlg::OnClose(): deleting thread");
      if (m_pThread->Delete() != wxTHREAD_NO_ERROR )
      {
        wxLogError("Can't delete the thread!");
      }
    }
  }
  // NeBJZNV甲邱Ƃɂ, Xbh̃fXgN^
  // NeBJZNVɓ悤ɂ.
  // (Xbh̃fXgN^ m_pThreadCS ŃK[hĂ!)
  
  while (1)
  {
    {
      // ~ProgressThread() ֐͎sꂽ?
      // sꂽȂ, m_pThread  NULL ɂȂĂ͂.
      wxCriticalSectionLocker enter(m_pThreadCS);
      if (!m_pThread) break;
    }
    
    // XbhI܂őҋ@.
    wxThread::This()->Sleep(1);
  }

  // ShowModal() ŕ\_CAO͕K EndModal() ŕȂ΂ȂȂ.
  // Destroy() p, _CAÕfXgN^őd delete ̃G[.
  EndModal(IDCANCEL);
}

void ProgressDlg::OnIdle(wxIdleEvent &event)
{
  // XbhI (m_pThread  NULL) mF, IĂ,
  // _CAO.
  bool threadStopped = false;
  {
    wxCriticalSectionLocker enter(m_pThreadCS);
    threadStopped = !m_pThread;
  }

  if (threadStopped && IsShown())
  {
    wxLogDebug("ProgressDlg::OnIdle(): call Close()");
    Close();
  }
}

void ProgressDlg::OnInitDialog(wxInitDialogEvent &event)
{
  DoStartThread();
}

void ProgressDlg::OnCancel(wxCommandEvent &event)
{
  Close();
  // OnClose() CxgĂяo, OnClose() CxgnhŃXbh̏I
  // s.
}

void ProgressDlg::Thread_OnProgressCompleted(wxThreadEvent &event)
{
  wxLogDebug("ProgressDlg::Thread_OnProgressCompleted");
}

void ProgressDlg::Thread_OnProgressUpdatePos(wxThreadEvent &event)
{
  wxLogDebug("ProgressDlg::Thread_OnProgressUpdatePos");

  ProgressUpdateInfo info = event.GetPayload<ProgressUpdateInfo>();  
  m_gaugeProgress->SetRange(info.range);
  m_gaugeProgress->SetValue(info.pos);
}

void ProgressDlg::Thread_OnProgressUpdateMsg(wxThreadEvent &event)
{
  wxLogDebug("ProgressDlg::Thread_OnProgressUpdateMsg");

  wxString msg = event.GetString();
  m_staticTextMsg->SetLabel(msg);
}

void ProgressDlg::DoStartThread()
{
  wxASSERT(NULL == m_pThread);
  wxASSERT(m_proc != NULL);

  // XbhN܂.
  wxLogDebug("ProgressDlg::DoStartThread(): starting thread");

  m_pThread = new ProgressThread(this, m_proc);
  if (m_pThread->Run() != wxTHREAD_NO_ERROR)
  {
    wxLogError("Can't create the thread!");
    delete m_pThread;
    m_pThread = NULL;
    wxASSERT(false);  // f, O𓊂ǂȂ.
  }

  // wxThread::Run() Ăяo, łXbh͏I\邽,
  // m_pThread |C^ւ̃ANZX͈Sł͂Ȃ.
}

//--------------------------------------------------------------------------------
#if !defined(NDEBUG)

// fobOp֐ 0.5 bɃOo͂.
bool DebugProgressFunc(Progress *pProgress)
{
  static int const REPEAT = 20;
  wxString msg;

  msg = _T("Start!");
  pProgress->UpdateProgress(0, REPEAT);
  
  for (int i = 1; i <= REPEAT; ++i)
  {
    wxMilliSleep(500);

    if (i < REPEAT / 2)
    {
      msg = _T("First Half!");
    }
    else
    {
      msg = _T("Last Half!");
    }

    pProgress->UpdateProgress(i, REPEAT);
    pProgress->UpdateMessage(msg);

    if (pProgress->IsAbortRequested())
    {
      // f
      return false;
    }

    wxLogDebug("DebugProgressFunc() : %d/%d", i, REPEAT);
  }

  // 
  return true;
}

bool DebugProgress(ProcContext *pPC)
{
  wxLogDebug("DebugProgress(): Begin");

  bool ret = RunProcWithContext<bool>(DebugProgressFunc, pPC);

  wxLogDebug("DebugProgress(): End");

  return ret;
}

#endif

//--------------------------------------------------------------------------------
