#include "ProjectPanel.h"

//--------------------------------------------------------------------------------
ProjectPanel::ProjectPanel(wxWindow *parent, wxWindowID id,
  const wxPoint& pos, const wxSize& size, long style)
  : ProjectPanelDesign(parent, id, pos, size, style),
    m_proj(NULL)
{
}

void ProjectPanel::SetProject(GraphProject *project)
{
  if (m_proj != project)
  {
    if (m_proj != NULL)
    {
      m_proj->Detach(this);
    }
    m_proj = project;
    if (m_proj != NULL)
    {
      m_proj->Attach(this);
      UpdateCheckListProj();
    }
  }
}

int ProjectPanel::GetSelectedGraphIndex() const
{
  int index = m_checkListProj->GetSelection();
  wxASSERT(index >= -1 && index < m_proj->GetCount());
  return index;
}

void ProjectPanel::Update(Observable *sender, ObserverParamBase const &param)
{
  GraphProjectUpdate update;
  if (GetUserParam(param, update))
  {
    if (update.IsUpdated(GraphProjectUpdate::DATA_APPENDED) ||
        update.IsUpdated(GraphProjectUpdate::DATA_REMOVED) ||
        (update.IsUpdated(GraphProjectUpdate::DATA_UPDATED) &&
          (update.DataUpdateDetail().IsUpdated(GraphDataUpdate::LABEL_UPDATED) ||
           update.DataUpdateDetail().IsUpdated(GraphDataUpdate::ENABLED_UPDATED))))
    {
      UpdateCheckListProj();
    }
  }
}

void ProjectPanel::CheckListProj_OnToggled(wxCommandEvent &event)
{
  int index = event.GetInt();
  wxASSERT(index >= 0 && index < m_proj->GetCount());
  m_proj->GetItem(index)->Enable(m_checkListProj->IsChecked(index));
}

void ProjectPanel::UpdateCheckListProj()
{
  wxASSERT(m_proj != NULL);

  // ACe̐Otf[^̐ƈv
  while (m_proj->GetCount() > static_cast<int>(m_checkListProj->GetCount()))
  {
    m_checkListProj->Append(_T(""));
  }
  while (m_proj->GetCount() < static_cast<int>(m_checkListProj->GetCount()))
  {
    m_checkListProj->Delete(m_checkListProj->GetCount() - 1);
  }

  // eLXg, `FbNݒ肷
  for (int i = 0; i < m_proj->GetCount(); ++i)
  {
    if (m_proj->GetItem(i)->GetLabel() != m_checkListProj->GetString(i))
    {
      wxString str;
      str.Printf(_T("[%02d]%s"), i + 1, m_proj->GetItem(i)->GetLabel());
      m_checkListProj->SetString(i, str);
    }
    if (m_proj->GetItem(i)->IsEnabled() != m_checkListProj->IsChecked(i))
    {
      m_checkListProj->Check(i, m_proj->GetItem(i)->IsEnabled());
    }
  }
}

//--------------------------------------------------------------------------------
