#include "PropertyPanel.h"

#include <wx/propgrid/advprops.h>
#include <wx/valnum.h>
#include <wx/log.h>

//--------------------------------------------------------------------------------
struct PropMsgItem
{
  wxChar const *szName;
  wxChar const *szDesc;
};

static PropMsgItem const projPropMsgs [] =
  { { _T("Information"),      _("Display the information about this project.") },
    { _T("DataRange"),        _("Display the range of all data.") },
    { _T("DataRange.MinX"),   _("Display the minimal X-axis value of all data.") },
    { _T("DataRange.MaxX"),   _("Display the maximum X-axis value of all data.") },
    { _T("DataRange.MinY"),   _("Display the minimal Y-axis value of all data.") },
    { _T("DataRange.MaxY"),   _("Display the maximum Y-axis value of all data.") },
    { _T("Plot"),             _("Please input the graph plotting properties.") },
    { _T("Axis"),             _("Please input the axes settings.") },
    { _T("Axis.LabelX"),      _("Please input the label text of X-axis.") },
    { _T("Axis.LabelY"),      _("Please input the label text of Y-axis.") },
    { _T("Axis.Font"),        _("Please set the font of the axes.") },
    { _T("Axis.FontColor"),   _("Please set the font color of the axes.") },
    { _T("Axis.PositionX"),   _("Please input the X-position of the Y-axis label against the graph frame.") },
    { _T("Axis.PositionY"),   _("Please input the Y-position of the X-axis label against the graph frame.") },
    { _T("BGColor"),          _("Please set the background color of graph frame.") },
    { _T("Frame"),            _("Please set the graph frame properties.") },
    { _T("Frame.LineWidth"),  _("Please input the line width of the graph frame.") },
    { _T("Frame.LineColor"),  _("Please set the line color of the graph frame.") },
    { _T("Grid"),             _("Please set the grid properties.") },
    { _T("Grid.EnabledX"),    _("Please set whether or not to show the X-axis grid.") },
    { _T("Grid.EnabledY"),    _("Please set whether or not to show the Y-axis grid.") },
    { _T("Grid.AutoX"),       _("Please set whether or not to decide the interval of X-axis grid automatically.") },
    { _T("Grid.AutoY"),       _("Please set whether or not to decide the interval of Y-axis grid automatically.") },
    { _T("Grid.IntervalX"),   _("Please input the interval of X-axis grid.") },
    { _T("Grid.IntervalY"),   _("Please input the interval of Y-axis grid.") },
    { _T("Grid.LineWidth"),   _("Please input the line width of the grid.") },
    { _T("Grid.LineColor"),   _("Please set the line color of the grid.") },
    { _T("Margin"),           _("Please set the margin of the graph frame.") },
    { _T("Margin.Top"),       _("Please input the top space between the graph frame and the display area.") },
    { _T("Margin.Bottom"),    _("Please input the bottom space between the graph frame and the display area.") },
    { _T("Margin.Left"),      _("Please input the left space between the graph frame and the display area.") },
    { _T("Margin.Right"),     _("Please input the right space between the graph frame and the display area.") },
    { _T("PlotRange"),        _("Please set the range of the graph.") },
    { _T("PlotRange.MinX"),   _("Please set the minimal X-axis value of plotted data.") },
    { _T("PlotRange.MaxX"),   _("Please set the maximum X-axis value of plotted data.") },
    { _T("PlotRange.MinY"),   _("Please set the minimal Y-axis value of plotted data.") },
    { _T("PlotRange.MaxY"),   _("Please set the maximum Y-axis value of plotted data.") },
    { _T("Tick"),             _("Please set the ticks properties.") },
    { _T("Tick.LineLength"),  _("Please input the line length of the ticks.") },
    { _T("Tick.LineWidth"),   _("Please input the line width of the ticks.") },
    { _T("Tick.LineColor"),   _("Please set the line color of the ticks.") },
    { _T("Tick.Font"),        _("Please set the font of the ticks.") },
    { _T("Tick.FontColor"),   _("Please set the font color of the ticks.") },
    { _T("Tick.PositionX"),   _("Please input the X-position of the Y-axis tick against the graph frame.") },
    { _T("Tick.PositionY"),   _("Please input the Y-position of the X-axis tick against the graph frame.") },
    { _T("Tick.FormatX"),     _("Please input the format of X-axis tick value like the printf() function of C language.") },
    { _T("Tick.FormatY"),     _("Please input the format of Y-axis tick value like the printf() function of C language.") } };

static PropMsgItem const graphPropMsgs [] =
  { { _T("Information"),      _("Display the information about this graph data.") },
    { _T("DataRange"),        _("Display the range of this graph  data.") },
    { _T("DataRange.MinX"),   _("Display the minimal X-axis value of this graph data.") },
    { _T("DataRange.MaxX"),   _("Display the maximum X-axis value of this graph data.") },
    { _T("DataRange.MinY"),   _("Display the minimal Y-axis value of this graph data.") },
    { _T("DataRange.MaxY"),   _("Display the maximum Y-axis value of this graph data.") },
    { _T("Label"),            _("Please input the label text.") },
    { _T("Points"),           _("Display the number of the data points.") },
    { _T("Plot"),             _("Please input the graph plotting properties.") },
    { _T("Curve"),            _("Please input the curve plotting properties.") },
    { _T("Curve.LineWidth"),  _("Please input the line width of the graph data.") },
    { _T("Curve.LineColor"),  _("Please set the line color of the graph data.") } };

//--------------------------------------------------------------------------------
PropertyPanel::PropertyPanel(wxWindow *parent, wxWindowID id,
  const wxPoint& pos, const wxSize& size, long style)
  : PropertyPanelDesign(parent, id, pos, size, style),
    m_proj(NULL),
    m_propGridUserAction(false)
{
  ClearPropertyList();
  m_propGrid->Connect(wxEVT_PG_CHANGED, wxPropertyGridEventHandler(PropertyPanel::PropGrid_OnChanged), NULL, this);
  m_propGrid->Connect(wxEVT_PG_SELECTED, wxPropertyGridEventHandler(PropertyPanel::PropGrid_OnSelected), NULL, this);
}

PropertyPanel::~PropertyPanel()
{
  m_propGrid->Disconnect(wxEVT_PG_CHANGED, wxPropertyGridEventHandler(PropertyPanel::PropGrid_OnChanged), NULL, this);
  m_propGrid->Disconnect(wxEVT_PG_SELECTED, wxPropertyGridEventHandler(PropertyPanel::PropGrid_OnSelected), NULL, this);
}

void PropertyPanel::SetProject(GraphProject *project)
{
  if (m_proj != project)
  {
    if (m_proj != NULL)
    {
      m_proj->Detach(this);
      UpdateTargetList();
    }
    m_proj = project;
    if (m_proj != NULL)
    {
      m_proj->Attach(this);
      UpdateTargetList();
    }
  }
}

void PropertyPanel::ComboBoxTarget_OnComboBox(wxCommandEvent& event)
{
  int index = m_comboBoxTarget->GetSelection();
  wxLogDebug("PropertyPanel::ComboBoxTarget_OnComboBox(): index = %d\n", index);
  UpdatePropertyList(UPDATE_ALL);
}

void PropertyPanel::PropGrid_OnChanged(wxPropertyGridEvent &event)
{
  wxASSERT(m_proj != NULL);

  wxPGProperty *property = event.GetProperty();

  // It may be NULL
  if (!property) return;

  // Get name of changed property
  wxString strName = property->GetName();

  // Get resulting value
  wxString strVal = property->GetValueAsString();
  
  // ȍ~, [U[ɂăvpeBObhꂽƂ.
  // ŃvWFNgIuWFNg̃vpeBݒ肳ꂽƂ, IuU[o[@\ɂ
  // Update() o֐Ăяo邪, Ƀ[U[ɂăvpeBObh
  // lɂȂĂ邽, l̍XVsȂ悤ɂ.
  m_propGridUserAction = true;

  int sel = m_comboBoxTarget->GetSelection();
  if (0 == sel)
  {
    GraphProjectPlotProperty prop = m_proj->GetPlotProperty();
    
    if (_T("PlotRange") == strName)
    {
      double minX = GetPropValue<double>(m_pidPlotRangeMinX);
      double maxX = GetPropValue<double>(m_pidPlotRangeMaxX);
      double minY = GetPropValue<double>(m_pidPlotRangeMinY);
      double maxY = GetPropValue<double>(m_pidPlotRangeMaxY);
      if (minX >= maxX)
      {
        // ͒l͈͊Oł邽, vbg͈͂ݒ肹Ɍɖ߂.
        UpdatePropertyList(UPDATE_PLOT_RANGE);
      }
      else if (minY >= maxY)
      {
        // ͒l͈͊Oł邽, vbg͈͂ݒ肹Ɍɖ߂.
        UpdatePropertyList(UPDATE_PLOT_RANGE);
      }
      else
      {
        m_proj->SetPlotRange(minX, maxX, minY, maxY);
      }
    }
    else if (_T("BGColor") == strName)
    {
      wxColour col = GetPropValue<wxColour>(m_pidBGColor);

      prop.SetBgColor(col);
      m_proj->SetPlotProperty(prop);
    }
    else if (_T("Margin") == strName)
    {
      double top    = GetPropValue<double>(m_pidMarginTop);
      double bottom = GetPropValue<double>(m_pidMarginBottom);
      double left   = GetPropValue<double>(m_pidMarginLeft);
      double right  = GetPropValue<double>(m_pidMarginRight);

      prop.SetMarginTop(top);
      prop.SetMarginBottom(bottom);
      prop.SetMarginLeft(left);
      prop.SetMarginRight(right);
      m_proj->SetPlotProperty(prop);
    }
    else if (_T("Frame") == strName)
    {
      double width = GetPropValue<double>(m_pidFrameLineWidth);
      wxColour color = GetPropValue<wxColour>(m_pidFrameLineColor);

      prop.SetGraphAreaFrameWidth(width);
      prop.SetGraphAreaFrameColor(color);
      m_proj->SetPlotProperty(prop);
    }
    else if (_T("Grid") == strName)
    {
      bool prevAutoX = prop.GridIntervalXAuto();
      bool prevAutoY = prop.GridIntervalYAuto();

      bool enabledX = GetPropValue<bool>(m_pidGridEnabledX);
      bool enabledY = GetPropValue<bool>(m_pidGridEnabledY);
      bool autoX = GetPropValue<bool>(m_pidGridAutoX);
      bool autoY = GetPropValue<bool>(m_pidGridAutoY);
      double intervalX = GetPropValue<double>(m_pidGridIntervalX);
      double intervalY = GetPropValue<double>(m_pidGridIntervalY);
      double width = GetPropValue<double>(m_pidGridLineWidth);
      wxColour color = GetPropValue<wxColour>(m_pidGridLineColor);
      
      prop.SetGridXEnabled(enabledX);
      prop.SetGridYEnabled(enabledY);
      prop.SetGridIntervalXAuto(autoX);
      prop.SetGridIntervalYAuto(autoY);
      if (!prevAutoX) prop.SetGridIntervalX(intervalX);
      if (!prevAutoY) prop.SetGridIntervalY(intervalY);
      prop.SetGridLineWidth(width);
      prop.SetGridLineColor(color);
      m_proj->SetPlotProperty(prop);

      if (m_pidGridIntervalX != NULL)
      {
        if (prevAutoX && !autoX)
        {
          // X̃ObhԊu蓮ɕωƂ
          m_propGrid->SetPropertyValue(m_pidGridIntervalX, prop.GridIntervalX());
        }
        else if (!prevAutoX && autoX && m_pidGridIntervalX != NULL)
        {
          // X̃ObhԊu蓮玩ɕωƂ
          m_propGrid->SetPropertyValue(m_pidGridIntervalX, 0.0);
        }
        m_propGrid->EnableProperty(m_pidGridIntervalX, !autoX);
      }

      if (m_pidGridIntervalY != NULL)
      {
        if (prevAutoY && !autoY)
        {
          // X̃ObhԊu蓮ɕωƂ
          m_propGrid->SetPropertyValue(m_pidGridIntervalY, prop.GridIntervalY());
        }
        else if (!prevAutoY && autoY && m_pidGridIntervalY != NULL)
        {
          // X̃ObhԊu蓮玩ɕωƂ
          m_propGrid->SetPropertyValue(m_pidGridIntervalY, 0.0);
        }
        m_propGrid->EnableProperty(m_pidGridIntervalY, !autoY);
      }
    }
    else if (_T("Tick") == strName)
    {
      double length = GetPropValue<double>(m_pidTickLineLength);
      double width = GetPropValue<double>(m_pidTickLineWidth);
      wxColour lineColor = GetPropValue<wxColour>(m_pidTickLineColor);
      wxFont font = GetPropValue<wxFont>(m_pidTickFont);
      wxColour fontColor = GetPropValue<wxColour>(m_pidTickFontColor);
      double positionX = GetPropValue<double>(m_pidTickPositionX);
      double positionY = GetPropValue<double>(m_pidTickPositionY);
      wxString formatX = GetPropValue<wxString>(m_pidTickFormatX);
      wxString formatY = GetPropValue<wxString>(m_pidTickFormatY);

      prop.SetTickLineLength(length);
      prop.SetTickLineWidth(width);
      prop.SetTickLineColor(lineColor);
      prop.SetTickLabelFont(font);
      prop.SetTickLabelFontColor(fontColor);
      prop.SetTickLabelPosX(positionX);
      prop.SetTickLabelPosY(positionY);
      prop.SetTickLabelFormatX(formatX);
      prop.SetTickLabelFormatY(formatY);
      m_proj->SetPlotProperty(prop);
    }
    else if (_T("Axis") == strName)
    {
      wxString labelX = GetPropValue<wxString>(m_pidAxisLabelX);
      wxString labelY = GetPropValue<wxString>(m_pidAxisLabelY);
      wxFont font = GetPropValue<wxFont>(m_pidAxisFont);
      wxColour color = GetPropValue<wxColour>(m_pidAxisFontColor);
      double positionX = GetPropValue<double>(m_pidAxisPositionX);
      double positionY = GetPropValue<double>(m_pidAxisPositionY);

      prop.SetAxisXLabel(labelX);
      prop.SetAxisYLabel(labelY);
      prop.SetAxisLabelFont(font);
      prop.SetAxisLabelFontColor(color);
      prop.SetAxisYLabelPosX(positionX);
      prop.SetAxisXLabelPosY(positionY);
      m_proj->SetPlotProperty(prop);
    }
  }
  else
  {
    wxASSERT(sel <= m_proj->GetCount());
    GraphDataRefPtr graph = m_proj->GetItem(sel - 1);
    GraphDataPlotProperty prop = graph->GetPlotProperty();
    
    if (_T("Label") == strName)
    {
      // R{{bNXKv邽, Update() Ăяo悤ɂ.
      m_propGridUserAction = false;

      wxString label = GetPropValue<wxString>(m_pidLabel);
      graph->SetLabel(label);
    }
    else if (_T("Curve") == strName)
    {
      double width = GetPropValue<double>(m_pidCurveLineWidth);
      wxColour color = GetPropValue<wxColour>(m_pidCurveLineColor);

      prop.SetLineWidth(width);
      prop.SetLineColor(color);
      graph->SetPlotProperty(prop);
    }
  }

  m_propGridUserAction = false;
  wxLogDebug("PropertyPanel::PropGrid_OnChanged(): name = %s, val = %s", strName, strVal);
}

void PropertyPanel::PropGrid_OnSelected(wxPropertyGridEvent &event)
{
  wxPGProperty *property = event.GetProperty();

  // It may be NULL
  if (!property) return;

  // Get name of changed property
  wxString strName = property->GetName();

  int sel = m_comboBoxTarget->GetSelection();
  
  static PropMsgItem const *items;
  int count;
  if (0 == sel)
  {
    items = projPropMsgs;
    count = sizeof(projPropMsgs) / sizeof(projPropMsgs[0]);
  }
  else
  {
    wxASSERT(sel <= m_proj->GetCount());
    items = graphPropMsgs;
    count = sizeof(graphPropMsgs) / sizeof(graphPropMsgs[0]);
  }

  m_textCtrlDesc->Clear();
  m_textCtrlDesc->AppendText(strName);
  m_textCtrlDesc->AppendText(_T("\n\n"));

  for (int i = 0; i < count; ++i)
  {
    if (strName == items[i].szName)
    {
      m_textCtrlDesc->AppendText(items[i].szDesc);
      break;
    }
  }

  wxLogDebug("PropGrid_OnSelected(): name = %s", strName);
}

void PropertyPanel::Update(Observable *sender, ObserverParamBase const &param)
{
  // [U[ɂăvpeBObhꂽƂ, XV̕Kv͂Ȃ.
  if (m_propGridUserAction) return;

  GraphProjectUpdate update;
  if (GetUserParam(param, update))
  {
    // OtvWFNgAbvf[gƂ
    if (update.IsUpdated(GraphProjectUpdate::DATA_APPENDED) ||
        update.IsUpdated(GraphProjectUpdate::DATA_REMOVED) ||
        (update.IsUpdated(GraphProjectUpdate::DATA_UPDATED) &&
          update.DataUpdateDetail().IsUpdated(GraphDataUpdate::LABEL_UPDATED)))
    {
      UpdateTargetList();
      wxLogDebug("PropertyPanel::Update: call UpdateTargetList()");
    }
    else if (update.IsUpdated(GraphProjectUpdate::PROP_UPDATED))
    {
      UpdatePropertyList(UPDATE_PROJ_PROP);
      wxLogDebug("PropertyPanel::Update: call UpdatePropertyList(UPDATE_PROJ_PROP)");
    }
    else if (update.IsUpdated(GraphProjectUpdate::RANGE_UPDATED) && !update.IsTemporal())
    {
      // XN[̈ꎞIł͂ȂƂ, vbgWXV.
      UpdatePropertyList(UPDATE_PLOT_RANGE);
      wxLogDebug("PropertyPanel::Update: call UpdatePropertyList(UPDATE_PLOT_RANGE)");
    }
    else if (update.IsUpdated(GraphProjectUpdate::DATA_UPDATED) &&
             update.DataUpdateDetail().IsUpdated(GraphDataUpdate::PROP_UPDATED))
    {
      // ł, Otf[^̂ǂꂪAbvf[gmF, vpeBXgɔfׂ
      UpdatePropertyList(UPDATE_GRAPH_PROP);
      wxLogDebug("PropertyPanel::Update: call UpdatePropertyList(UPDATE_GRAPH_PROP)");
    }
  }
}

void PropertyPanel::UpdateTargetList()
{
  wxASSERT(m_proj != NULL);

  int sel = m_comboBoxTarget->GetSelection();
  m_comboBoxTarget->Clear();

  if (NULL == m_proj) return;
  m_comboBoxTarget->AppendString(_("Project"));

  for (int i = 0; i < m_proj->GetCount(); ++i)
  {
    GraphDataRefPtr graph = m_proj->GetItem(i);
    wxString str;
    str.Printf(_T("[%02d]%s"), i + 1, graph->GetLabel());
    m_comboBoxTarget->AppendString(str);
  }

  // IvfĐݒ肷.
  if (-1 == sel) sel = 0;
  if (sel >= static_cast<int>(m_comboBoxTarget->GetCount()))
  {
    sel = m_comboBoxTarget->GetCount() - 1;
  }
  m_comboBoxTarget->SetSelection(sel);

  // vpeBAbvf[g.
  UpdatePropertyList(UPDATE_ALL);
}

void PropertyPanel::UpdatePropertyList(UpdatePropContent content)
{
  wxASSERT(m_proj != NULL);

  int sel = m_comboBoxTarget->GetSelection();

  switch (content)
  {
  case UPDATE_ALL:
    ClearPropertyList();
    
    if (0 == sel)
    {
      AppendProjectProperty();
    }
    else
    {
      AppendGraphProperty(sel - 1);
    }
    break;
  case UPDATE_PLOT_RANGE:
    if (0 == sel)
    {
      if (m_pidPlotRangeMinX != NULL) m_propGrid->SetPropertyValue(m_pidPlotRangeMinX, m_proj->GetPlotMinX());
      if (m_pidPlotRangeMaxX != NULL) m_propGrid->SetPropertyValue(m_pidPlotRangeMaxX, m_proj->GetPlotMaxX());
      if (m_pidPlotRangeMinY != NULL) m_propGrid->SetPropertyValue(m_pidPlotRangeMinY, m_proj->GetPlotMinY());
      if (m_pidPlotRangeMaxY != NULL) m_propGrid->SetPropertyValue(m_pidPlotRangeMaxY, m_proj->GetPlotMaxY());
    }    
    break;
  case UPDATE_PROJ_PROP:    // ͈CvWFNgvpeBXV̂ۂׂ͂čXV.
    if (0 == sel)
    {
      ClearPropertyList();
      AppendProjectProperty();
    }
    break;
  case UPDATE_GRAPH_PROP:   // ͈COtvpeBXV̂ۂׂ͂čXV.
    if (sel != 0)
    {
      ClearPropertyList();
      AppendGraphProperty(sel - 1);
    }
    break;
  }
}

void PropertyPanel::ClearPropertyList()
{
  m_pidPlotRangeMinX = m_pidPlotRangeMaxX = NULL;
  m_pidPlotRangeMinY = m_pidPlotRangeMaxY = NULL;
  m_pidBGColor = NULL;
  m_pidMarginTop = m_pidMarginBottom = m_pidMarginLeft = m_pidMarginRight = NULL;
  m_pidFrameLineWidth = m_pidFrameLineColor = NULL;
  m_pidGridEnabledX = m_pidGridEnabledY = NULL;
  m_pidGridAutoX = m_pidGridAutoY = NULL;
  m_pidGridIntervalX = m_pidGridIntervalY = NULL;
  m_pidGridLineWidth = m_pidGridLineColor = NULL;
  m_pidTickLineLength = m_pidTickLineWidth = m_pidTickLineColor = NULL;
  m_pidTickFont = m_pidTickFontColor = NULL;
  m_pidTickPositionX = m_pidTickPositionY = NULL;
  m_pidTickFormatX = m_pidTickFormatY = NULL;
  m_pidAxisLabelX = m_pidAxisLabelY = NULL;
  m_pidAxisFont = m_pidAxisFontColor = NULL;
  m_pidAxisPositionX = m_pidAxisPositionY = NULL;

  m_pidCurveLineWidth = m_pidCurveLineColor = NULL;

  m_propGrid->Clear();
}

void PropertyPanel::AppendProjectProperty()
{
  wxPGProperty *pid1;

  GraphProjectPlotProperty prop = m_proj->GetPlotProperty();

  wxFloatingPointValidator<double> validMargin;
  validMargin.SetRange(0.0, 200.0);

  wxFloatingPointValidator<double> validLineWidth;
  validLineWidth.SetRange(0.1, 5.0);

  wxFloatingPointValidator<double> validInterval;
  validInterval.SetRange(0.0, DBL_MAX);

  wxFloatingPointValidator<double> validTickAxisPos;
  validTickAxisPos.SetRange(-200.0, 200.0);

  // 
  m_propGrid->Append(new wxPropertyCategory(_T("Information")));

  pid1 = m_propGrid->Append(new wxStringProperty(_T("DataRange"), wxPG_LABEL, _T("<composed>")));
  m_propGrid->AppendIn(pid1, new wxFloatProperty(_("MinX"), wxPG_LABEL, m_proj->GetMinX()));
  m_propGrid->AppendIn(pid1, new wxFloatProperty(_("MaxX"), wxPG_LABEL, m_proj->GetMaxX()));
  m_propGrid->AppendIn(pid1, new wxFloatProperty(_("MinY"), wxPG_LABEL, m_proj->GetMinY()));
  m_propGrid->AppendIn(pid1, new wxFloatProperty(_("MaxY"), wxPG_LABEL, m_proj->GetMaxY()));
  //m_propGrid->SetPropertyReadOnly(pid1);
  m_propGrid->EnableProperty(pid1, false);
  
  // vbgݒ
  m_propGrid->Append(new wxPropertyCategory(_T("Plot")));

  pid1 = m_propGrid->Append(new wxStringProperty(_T("PlotRange"), wxPG_LABEL, _T("<composed>")));
  m_pidPlotRangeMinX = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("MinX"), wxPG_LABEL, m_proj->GetPlotMinX()));
  m_pidPlotRangeMaxX = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("MaxX"), wxPG_LABEL, m_proj->GetPlotMaxX()));
  m_pidPlotRangeMinY = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("MinY"), wxPG_LABEL, m_proj->GetPlotMinY()));
  m_pidPlotRangeMaxY = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("MaxY"), wxPG_LABEL, m_proj->GetPlotMaxY()));

  m_pidBGColor = m_propGrid->Append(new wxColourProperty(_T("BGColor"), wxPG_LABEL, prop.BgColor()));

  pid1 = m_propGrid->Append(new wxStringProperty(_T("Margin"), wxPG_LABEL, _T("<composed>")));
  m_pidMarginTop    = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("Top"),    wxPG_LABEL, prop.MarginTop()));
  m_pidMarginBottom = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("Bottom"), wxPG_LABEL, prop.MarginBottom()));
  m_pidMarginLeft   = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("Left"),   wxPG_LABEL, prop.MarginLeft()));
  m_pidMarginRight  = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("Right"),  wxPG_LABEL, prop.MarginRight()));
  m_pidMarginTop->SetValidator(validMargin);
  m_pidMarginBottom->SetValidator(validMargin);
  m_pidMarginLeft->SetValidator(validMargin);
  m_pidMarginRight->SetValidator(validMargin);
  
  pid1 = m_propGrid->Append(new wxStringProperty(_T("Frame"), wxPG_LABEL, _T("<composed>")));
  m_pidFrameLineWidth = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("LineWidth"),  wxPG_LABEL, prop.GraphAreaFrameWidth()));
  m_pidFrameLineColor = m_propGrid->AppendIn(pid1, new wxColourProperty(_T("LineColor"), wxPG_LABEL, prop.GraphAreaFrameColor()));
  m_pidFrameLineWidth->SetValidator(validLineWidth);

  pid1 = m_propGrid->Append(new wxStringProperty(_T("Grid"), wxPG_LABEL, _T("<composed>")));
  m_pidGridEnabledX  = m_propGrid->AppendIn(pid1, new wxBoolProperty (_T("EnabledX"),   wxPG_LABEL, prop.GridXEnabled()));
  m_pidGridEnabledY  = m_propGrid->AppendIn(pid1, new wxBoolProperty (_T("EnabledY"),   wxPG_LABEL, prop.GridYEnabled()));

  bool autoX = prop.GridIntervalXAuto();
  bool autoY = prop.GridIntervalYAuto();
  m_pidGridAutoX     = m_propGrid->AppendIn(pid1, new wxBoolProperty (_T("AutoX"),      wxPG_LABEL, autoX));
  m_pidGridAutoY     = m_propGrid->AppendIn(pid1, new wxBoolProperty (_T("AutoY"),      wxPG_LABEL, autoY));
  
  double intervalX = autoX ? 0.0 : prop.GridIntervalX();
  double intervalY = autoY ? 0.0 : prop.GridIntervalY();
  m_pidGridIntervalX = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("IntervalX"),  wxPG_LABEL, intervalX));
  m_pidGridIntervalY = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("IntervalY"),  wxPG_LABEL, intervalY));
  m_pidGridIntervalX->SetValidator(validInterval);
  m_pidGridIntervalY->SetValidator(validInterval);
  m_propGrid->EnableProperty(m_pidGridIntervalX, !autoX);
  m_propGrid->EnableProperty(m_pidGridIntervalY, !autoY);

  m_pidGridLineWidth = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("LineWidth"),  wxPG_LABEL, prop.GridLineWidth()));
  m_pidGridLineColor = m_propGrid->AppendIn(pid1, new wxColourProperty(_T("LineColor"), wxPG_LABEL, prop.GridLineColor()));
  m_pidGridLineWidth->SetValidator(validLineWidth);

  pid1 = m_propGrid->Append(new wxStringProperty(_T("Tick"), wxPG_LABEL, _T("<composed>")));
  m_pidTickLineLength = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("LineLength"), wxPG_LABEL, prop.TickLineLength()));
  m_pidTickLineWidth  = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("LineWidth"),  wxPG_LABEL, prop.TickLineWidth()));
  m_pidTickLineColor  = m_propGrid->AppendIn(pid1, new wxColourProperty(_T("LineColor"), wxPG_LABEL, prop.TickLineColor()));
  m_pidTickFont       = m_propGrid->AppendIn(pid1, new wxFontProperty (_T("Font"),       wxPG_LABEL, prop.TickLabelFont()));
  m_pidTickFontColor  = m_propGrid->AppendIn(pid1, new wxColourProperty(_T("FontColor"), wxPG_LABEL, prop.TickLabelFontColor()));
  m_pidTickPositionX  = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("PositionX"),  wxPG_LABEL, prop.TickLabelPosX()));
  m_pidTickPositionY  = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("PositionY"),  wxPG_LABEL, prop.TickLabelPosY()));
  m_pidTickFormatX    = m_propGrid->AppendIn(pid1, new wxStringProperty(_T("FormatX"),   wxPG_LABEL, prop.TickLabelFormatX())); 
  m_pidTickFormatY    = m_propGrid->AppendIn(pid1, new wxStringProperty(_T("FormatY"),   wxPG_LABEL, prop.TickLabelFormatY()));
  m_pidTickLineWidth->SetValidator(validLineWidth);
  m_pidTickPositionX->SetValidator(validTickAxisPos);
  m_pidTickPositionY->SetValidator(validTickAxisPos);

  pid1 = m_propGrid->Append(new wxStringProperty(_T("Axis"), wxPG_LABEL, _T("<composed>")));
  m_pidAxisLabelX    = m_propGrid->AppendIn(pid1, new wxStringProperty(_T("LabelX"),    wxPG_LABEL, prop.AxisXLabel()));
  m_pidAxisLabelY    = m_propGrid->AppendIn(pid1, new wxStringProperty(_T("LabelY"),    wxPG_LABEL, prop.AxisYLabel()));
  m_pidAxisFont      = m_propGrid->AppendIn(pid1, new wxFontProperty(_T("Font"),        wxPG_LABEL, prop.AxisLabelFont()));
  m_pidAxisFontColor = m_propGrid->AppendIn(pid1, new wxColourProperty(_T("FontColor"), wxPG_LABEL, prop.AxisLabelFontColor()));
  m_pidAxisPositionX = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("PositionX"),  wxPG_LABEL, prop.AxisYLabelPosX()));
  m_pidAxisPositionY = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("PositionY"),  wxPG_LABEL, prop.AxisXLabelPosY()));
  m_pidAxisPositionX->SetValidator(validTickAxisPos);
  m_pidAxisPositionY->SetValidator(validTickAxisPos);
}

void PropertyPanel::AppendGraphProperty(int index)
{
  wxPGProperty *pid1;

  wxASSERT(index < m_proj->GetCount());
  GraphDataRefPtr graph = m_proj->GetItem(index);
  GraphDataPlotProperty prop = graph->GetPlotProperty();

  // 
  m_propGrid->Append(new wxPropertyCategory(_T("Information")));

  m_pidLabel = m_propGrid->Append(new wxStringProperty(_T("Label"), wxPG_LABEL, graph->GetLabel()));

  pid1 = m_propGrid->Append(new wxIntProperty(_T("Points"), wxPG_LABEL, graph->GetCount()));
  //m_propGrid->SetPropertyReadOnly(pid1);
  m_propGrid->EnableProperty(pid1, false);

  pid1 = m_propGrid->Append(new wxStringProperty(_T("DataRange"), wxPG_LABEL, _T("<composed>")));
  m_propGrid->AppendIn(pid1, new wxFloatProperty(_("MinX"), wxPG_LABEL, graph->GetMinX()));
  m_propGrid->AppendIn(pid1, new wxFloatProperty(_("MaxX"), wxPG_LABEL, graph->GetMaxX()));
  m_propGrid->AppendIn(pid1, new wxFloatProperty(_("MinY"), wxPG_LABEL, graph->GetMinY()));
  m_propGrid->AppendIn(pid1, new wxFloatProperty(_("MaxY"), wxPG_LABEL, graph->GetMaxY()));
  //m_propGrid->SetPropertyReadOnly(pid1);
  m_propGrid->EnableProperty(pid1, false);

  // vbgݒ
  m_propGrid->Append(new wxPropertyCategory(_T("Plot")));

  pid1 = m_propGrid->Append(new wxStringProperty(_T("Curve"), wxPG_LABEL, _T("<composed>")));
  m_pidCurveLineWidth = m_propGrid->AppendIn(pid1, new wxFloatProperty(_T("LineWidth"),  wxPG_LABEL, prop.LineWidth()));
  m_pidCurveLineColor = m_propGrid->AppendIn(pid1, new wxColourProperty(_T("LineColor"), wxPG_LABEL, prop.LineColor()));
}

template <typename Value>
Value PropertyPanel::GetPropValue(wxPGProperty *pid)
{
  wxASSERT(pid != NULL);
  wxVariant var = m_propGrid->GetPropertyValue(pid);
  Value value;
  value << var;
  return value;
}

template <>
bool PropertyPanel::GetPropValue<bool>(wxPGProperty *pid)
{
  wxASSERT(pid != NULL);
  wxVariant var = m_propGrid->GetPropertyValue(pid);
  return var.GetBool();
}

template <>
double PropertyPanel::GetPropValue<double>(wxPGProperty *pid)
{
  wxASSERT(pid != NULL);
  wxVariant var = m_propGrid->GetPropertyValue(pid);
  return var.GetDouble();
}

template <>
wxString PropertyPanel::GetPropValue<wxString>(wxPGProperty *pid)
{
  wxASSERT(pid != NULL);
  wxVariant var = m_propGrid->GetPropertyValue(pid);
  return var.GetString();
}

//--------------------------------------------------------------------------------
