#include "SeparatorSettingDlg.h"

//--------------------------------------------------------------------------------
enum
{
  SEPARATOR_TAB_INDEX = 0,
  SEPARATOR_SEMICOLON_INDEX = 1,
  SEPARATOR_COMMA_INDEX = 2,
  SEPARATOR_SPACE_INDEX = 3,
  SEPARATOR_COUNT = 4,
};

struct SeparatorItem
{
  wxString caption;
  wxUniChar separator;
};

static SeparatorItem const separators[SEPARATOR_COUNT] =
  { { _T("Tab"), _T('\t') },
    { _T("Semicolon"), _T(';') },
    { _T("Comma"), _T(',') },
    { _T("Space"), _T(' ') } };

//--------------------------------------------------------------------------------
SeparatorSettingDlg::SeparatorSettingDlg(wxWindow* parent)
  : SeparatorSettingDlgDesign(parent),
    m_separator(_T(','))
{
  for (int i = 0; i < SEPARATOR_COUNT; ++i)
  {
    m_comboBoxSeparator->Append(separators[i].caption);
  }
  UpdateComboBox();
}

SeparatorSettingDlg::~SeparatorSettingDlg()
{
}

void SeparatorSettingDlg::SetSeparator(wxUniChar separator)
{
  wxASSERT(_T('\t') == separator || _T(';') == separator ||
           _T(',') == separator || _T(' ') == separator);
  m_separator = separator;
  UpdateComboBox();
}

wxUniChar SeparatorSettingDlg::GetSeparator() const
{
  return m_separator;
}

void SeparatorSettingDlg::ComboBoxSeparator_OnComboBox(wxCommandEvent &event)
{
  int index = m_comboBoxSeparator->GetSelection();
  wxASSERT(index >= 0 && index < SEPARATOR_COUNT);
  m_separator = separators[index].separator;
}

void SeparatorSettingDlg::UpdateComboBox()
{
  int index = 0;
  if      (_T('\t') == m_separator) index = SEPARATOR_TAB_INDEX;
  else if (_T(';') == m_separator)  index = SEPARATOR_SEMICOLON_INDEX;
  else if (_T(',') == m_separator)  index = SEPARATOR_COMMA_INDEX;
  else if (_T(' ') == m_separator)  index = SEPARATOR_SPACE_INDEX;
  else                              wxASSERT(0);

  m_comboBoxSeparator->SetSelection(index);
}

//--------------------------------------------------------------------------------
